/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.3 $
 * $Date: 2004/05/27 19:52:06 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.*;
import java.awt.geom.*;

import javax.swing.event.*;

import org.concord.graph.event.*;

/**
 * Coordinate system 
 * implemented as an affine transform matrix
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid C Moncada<p>
 *
 */

public class AffineTransformCoordinateSystem
	implements CoordinateSystem
{
    private double xScale = 20.0;	// Number of pixels in each world unit on the x axis 
    private double yScale = 20.0;	// Number of pixels in each world unit on the y axis 

    private double originX = 0.0;	// Position of the origin in world units
    private double originY = 0.0;

	private boolean fixOriginPixels = true;	/** Indicates if the origin will be fixed in the screen */

    private AffineTransform matrix;
    private AffineTransform inverse;

    private int screenHeight = 0;
    private boolean invertY = true;

	private Vector changeListeners;		//Listeners of changes in the coordinate system
	private boolean changing;

	public AffineTransformCoordinateSystem() 
	{
		changeListeners = new Vector();
		matrix = new AffineTransform();
		reset();
	}

	public void reset() 
	{	
		matrix.setToIdentity();
		if (invertY) {
			matrix.translate (0.0,(double) screenHeight);
			matrix.scale(xScale, -yScale);
			matrix.translate (originX, originY);
		} else {
			matrix.scale(xScale, yScale);
			matrix.translate(originX, originY);
		}
		try {
			inverse = matrix.createInverse();
		} catch (Exception e) {
			System.err.println("Can't create inverse matrix");
		}
		
		notifyChange();
		
//		checksum = (long)(xScale * 21477.7 + yScale * 7633.3 + 
//		           originX * 813.9 + originY); 
	}

	public AffineTransform getMatrix() 
	{
		return matrix;
	}

	public void setScale( double x, double y) 
	{
		if (x > 0.00001 && y > 0.00001){
			if (fixOriginPixels){
				double originXPx = xScale * originX;
				double originYPx = yScale * originY;
	
				originX = (double) originXPx / x;
				originY = (double) originYPx / y;
			}
			
			xScale = x;
			yScale = y;
			
			reset();
		}
	}

	public void setOrigin( double x, double y ) 
	{
		originX = x;
		originY = y;
		reset();
	}

	public AffineTransform getInverse() 
	{
		return inverse;
	}

	public void setOriginPixels(int x, int y) 
	{
		originX = (double) x / xScale;
		originY = (double) y / yScale;
		
		reset();
	}

	public void setFixOriginScreen(boolean b)
	{
		fixOriginPixels = b;
		
		reset();
	}

	public void setScreenHeight(int height) 
	{
		screenHeight = height;
		
		reset();
	}

    /** Returns the value of xScale */
	public double getXScale() 
	{
		return xScale;
	}

    /** Returns the value of yScale */
	public double getYScale() 
	{
		return yScale;
	}

	/** Returns the value of originX */
	public double getOriginX() 
	{
		return originX;
	}
	
	/** Returns the value of originY */
	public double getOriginY() 
	{
		return originY;
	}
	
	/** Returns the value of originX in pixels */
	public double getOriginXPixels() 
	{
		return originX * xScale;
	}

	/** Returns the value of originY in pixels */
	public double getOriginYPixels() 
	{
		return originY * yScale;
	}

	///////////////////////////////////////////////
	//Methods for Coordinate System interface
	
	/*
	 * Transform From display coordinates To world coordinates 
	 */
	public Point2D transformToWorld(Point2D d)
	{
		return transformToWorld(d, null);
	}
	
	/*
	 * Transform From display coordinates To world coordinates 
	 */
	public Point2D transformToWorld(Point2D d, Point2D w)
	{
		return getInverse().transform(d, w);
	}
	
	/*
	 * Transform From world coordinates To display coordinates 
	 */
	public Point2D transformToDisplay(Point2D w)
	{
		return transformToDisplay(w, null);
	}
	
	/*
	 * Transform From world coordinates To display coordinates 
	 */
	public Point2D transformToDisplay(Point2D w, Point2D d)
	{
		return getMatrix().transform(w, d);
	}
	
	/*
	 * Sets the scale (# of display coordinates in each world coordinate) 
	 */
	public void setScale(Point2D p)
	{
		setScale(p.getX(), p.getY());
	}

	/*
	 * Returns the scale (# of display coordinates in each world coordinate) 
	 */
	public Point2D getScale()
	{
		return new Point2D.Double(xScale, yScale);
	}

	/*
	 * Sets the origin in display coordinates 
	 */
	public void setOriginOffsetDisplay(Point2D p)
	{
	//The origin in display coordinates should be relative to the upper left corner
		if (screenHeight == 0){
			setOriginPixels((int)p.getX(), (int)(p.getY()));
		}
		else{
			setOriginPixels((int)p.getX(), (int)(screenHeight - p.getY()));
		}
	}

	/*
	 * Returns the origin in display coordinates 
	 */
	public Point2D getOriginOffsetDisplay()
	{
	//The origin in display coordinates should be relative to the upper left corner
		if (screenHeight == 0){
			return new Point2D.Double(getOriginXPixels(), getOriginYPixels());
		}
		else{
			return new Point2D.Double(getOriginXPixels(), screenHeight - getOriginYPixels());
		}
	}

	/**
	 * Add a listener to any change in the area of in its coordinate system
	 */
	public void addChangeListener(ChangeListener l)
	{
		changeListeners.add(l);
	}

	/**
	 * Removes a listener to changes in the area of in its coordinate system
	 */
	public void removeChangeListener(ChangeListener l)
	{
		changeListeners.remove(l);
	}

	protected void notifyChange() 
	{
		//For now, it notifies as it EVERYTHING changed 
		CoordinateSystemChangeEvent e = new CoordinateSystemChangeEvent(this, isChanging());
	
		notifyChange(e);
	}

	protected void notifyChange(CoordinateSystemChangeEvent e) 
	{
		ChangeListener l;
		Enumeration lists = changeListeners.elements();
	
		while (lists.hasMoreElements()) {
			l = (ChangeListener)lists.nextElement();
			l.stateChanged(e);
		}
	}

	/**
	 * Returns if the coordinate system is currently being changed 
	 */
	public boolean isChanging()
	{
		return changing;
	}

	/**
	 * Sets that the coordinate system is currently being changed or not 
	 */
	public void setChanging(boolean b)
	{
		changing = b;
	}
	///////////////////////////////////////////////
	
	public CoordinateSystem getCopy()
	{
		AffineTransformCoordinateSystem c = new AffineTransformCoordinateSystem();
		c.setScale(new Point2D.Double(xScale, yScale));
		c.setOrigin(originX, originY);
		c.setScreenHeight(screenHeight);
		return c;
	}
}
    
