/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.5 $
 * $Date: 2004/07/19 16:01:04 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.*;
import java.awt.geom.Point2D;

import javax.swing.event.*;

import org.concord.graph.event.*;

/**
 * DefaultCoordinateSystem2D
 * Orthogonal coordinate system that has independent x and y axis 
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid C Moncada<p>
 *
 */
public class DefaultCoordinateSystem2D
	implements CoordinateSystem
{
	private SingleAxis xAxis;
	private SingleAxis yAxis;
	
	private boolean changing = false;
	
	private Vector listeners;
	
	public DefaultCoordinateSystem2D()
	{
		listeners = new Vector();
		
		xAxis = new SingleAxis(0.0, 20.0);
		yAxis = new SingleAxis(0.0, 20.0);
		
		//By default the y axis is inverted 
		adjustScaleYAxis();
	}
	
	private void adjustScaleYAxis()
	{
		//By default the y axis is inverted 
		yAxis.setScale(-1 * yAxis.getScale());
	}
	
	/*
	 * Transform From display coordinates To world coordinates 
	 */
	public Point2D transformToWorld(Point2D d)
	{
		return transformToWorld(d, new Point2D.Double());
	}
	
	/*
	 * Transform From display coordinates To world coordinates 
	 */
	public Point2D transformToWorld(Point2D d, Point2D w)
	{
		double x = xAxis.transformDisplayToWorld(d.getX());
		double y = yAxis.transformDisplayToWorld(d.getY());
		w.setLocation(x,y);
		return w;
	}
	
	/*
	 * Transform From world coordinates To display coordinates 
	 */
	public Point2D transformToDisplay(Point2D w)
	{
		return transformToDisplay(w, new Point2D.Double());
	}
	
	/*
	 * Transform From world coordinates To display coordinates 
	 */
	public Point2D transformToDisplay(Point2D w, Point2D d)
	{
		double x = xAxis.transformWorldToDisplay(w.getX());
		double y = yAxis.transformWorldToDisplay(w.getY());
		d.setLocation(x,y);
		return d;
	}

	/*
	 * Sets the scale (# of display coordinates in each world coordinate) 
	 */
	public void setScale(Point2D p)
	{
		boolean xChange = false;
		boolean yChange = false;
		double newx;
		double newy;

		Point2D oldScale = new Point2D.Double(xAxis.getScale(), yAxis.getScale());
		
		newx = p.getX();
		newy = p.getY();

		if (newx > 0.0) 
			xAxis.setScale(newx);


		if (newy > 0.0) {
			yAxis.setScale(newy);

			//By default the y axis is inverted 
			adjustScaleYAxis();
		}
		
		if (oldScale.getX() != xAxis.getScale()){
			xChange = true;
		}

		if (oldScale.getY() != yAxis.getScale()){
			yChange = true;
		}
		
		if (xChange || yChange){
			notifyChange(true, false, xChange, yChange);
		}
	}

	/*
	 * Returns the scale (# of display coordinates in each world coordinate) 
	 */
	public Point2D getScale()
	{
		//By default the y axis is inverted 
		return new Point2D.Double(xAxis.getScale(), - yAxis.getScale());
	}

	/*
	 * Sets the origin in display coordinates 
	 */
	public void setOriginOffsetDisplay(Point2D p)
	{
		boolean xChange = false;
		boolean yChange = false;

		if (xAxis.getOrigin() != p.getX()){
			xAxis.setOrigin(p.getX());
			xChange = true;
		}
		if (yAxis.getOrigin() != p.getY()){
			yAxis.setOrigin(p.getY());
			yChange = true;
		}
		
		notifyChange(false, true, xChange, yChange);
	}

	/*
	 * Returns the origin in display coordinates 
	 */
	public Point2D getOriginOffsetDisplay()
	{
		return new Point2D.Double(xAxis.getOrigin(), yAxis.getOrigin());
	}

	/**
	 * Add a listener to any change in the coordinate system
	 */
	public void addChangeListener(ChangeListener l)
	{
		listeners.add(l);
	}

	/**
	 * Removes a listener to changes in the coordinate system
	 */
	public void removeChangeListener(ChangeListener l)
	{
		listeners.remove(l);
	}
	
	protected void notifyChange(boolean scale, boolean origin, boolean xChange, boolean yChange)
	{
		ChangeListener l;	
		CoordinateSystemChangeEvent e = new CoordinateSystemChangeEvent(this, scale, origin, xChange, yChange, isChanging());
		int i;
		
		for (i=0; i < listeners.size(); i++){
			l = (ChangeListener)listeners.elementAt(i);
			l.stateChanged(e);
		}
	}

	/**
	 * Returns if the coordinate system is currently being changed 
	 */
	public boolean isChanging()
	{
		return changing;
	}

	/**
	 * Sets that the coordinate system is currently being changed or not 
	 */
	public void setChanging(boolean b)
	{
		changing = b;
	}

	/**
	 * Returns a copy of itself
	 */
	public CoordinateSystem getCopy()
	{
		DefaultCoordinateSystem2D copy = new DefaultCoordinateSystem2D();
		copy.setScale(this.getScale());
		copy.setOriginOffsetDisplay(this.getOriginOffsetDisplay());
		return copy;
	}
	
}
