/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.1 $
 * $Date: 2004/05/24 20:16:30 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.Vector;
import java.util.Enumeration;
import java.util.EventObject;

import javax.swing.event.*;

import org.concord.graph.event.*;

/**
 * An interface for objects that are graphed or drawn in a graph window.
 * A Graphable is an object that can draw itself in a Graphics2D context 
 * It can also keep a buffered version of itself for speed purposes when drawing. 
 * It can be removed, be visible or invisible, and it can return a copy (clone) of itself.
 * Objects that want to be displayed in a Graph Window should implement this interface.
 * A Graphable also fires events when it has been changed or removed.
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid C Moncada<p>
 *
 */

public abstract class DefaultGraphable 
	implements Graphable, ChangeListener
{
	//////////////////////////////////////////////////////
	//Properties for Graphable interface 
	protected Vector graphableListeners;			//graphable listeners	

	private boolean visible = true;		//is the graphable currently visible

	protected GraphArea graphArea;
	//////////////////////////////////////////////////////
	
	public DefaultGraphable()
	{
		graphableListeners = new Vector();
	}
	
	/*
	 * Adds a GraphableListener to listen for events 
	 */
	public void addGraphableListener(GraphableListener gl) 
	{
		graphableListeners.add(gl);
	}
	
	/*
	 * Removes a GraphableListener.
	 */
	public void removeGraphableListener(GraphableListener gl) 
	{
		if (graphableListeners.contains(gl)) {
			graphableListeners.remove(gl);
		}
	}

	/** 
	 * Doesn't support buffering (cache)
	 */
	public void update()
	{
	}
	
	/** 
	 * Removes the object from any enclosing view.
	 * This also fires an objectRemoved(obj) event.
	 */
	public void remove () 
	{
		notifyRemove();
	}
	
	/** 
	 * Sets whether it is visible in a graph.
	 * This also fires an objectChanged(obj) event.
	 */
	public void setVisible(boolean visible) 
	{
		this.visible = visible;
		notifyChange();
	}
	
	/** 
	 * Returns true if it is visible in a graph 
	 */
	public boolean isVisible() 
	{
		return visible;
	}

	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		if (graphArea != null){
			graphArea.removeChangeListener(this);
		}
		this.graphArea = area;
		if (graphArea != null){
			graphArea.addChangeListener(this);
		}
	}

	/** 
	 * Returns the graph area where the graphable will be drawn 
	 */
	public GraphArea getGraphArea()
	{
		return this.graphArea;
	}

	/**
	 * Handler of the changes in the graph area
	 */
	public void stateChanged(ChangeEvent e)
	{
		update();
		notifyChange();
	}

    protected void notifyChange() 
	{		
		EventObject e = new EventObject(this);
		Enumeration lists = graphableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			gl.graphableChanged(e);
		}
    }

	protected void notifyRemove() 
	{
		EventObject e = new EventObject(this);
		Enumeration lists = graphableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			gl.graphableRemoved(e);
		}
	}
}
    
