/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.8 $
 * $Date: 2004/06/18 15:05:24 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.*;
import java.awt.*;
import java.awt.geom.Point2D;

import javax.swing.event.*;

import org.concord.graph.event.*;

/**
 * Graph Area that manages an area in a coordinate system
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid C Moncada<p>
 *
 */

public class GraphArea
	implements ChangeListener, GraphWindowListener
{
	private Point2D position;	//position of the graph area in display coordinates
								//relative to the LOWER LEFT corner of the screen
	private Dimension size;		//size of the graph area in display coordinates
	
	private CoordinateSystem coord;	//Coordinate system
	
	private Vector changeListeners;	//Listeners to any change in the graph area or its coordinate system

	//All this info is to handle automatic resizing/positioning of CS
    /** Turns on or off automatic centering (on resize) */
    private boolean centeredX = false;
    private boolean centeredY = false;	
	

	/** Constructs a GraphArea and Initializes with a default Coordinate
		system.*/
	public GraphArea()
	{
		CoordinateSystem cs = new DefaultCoordinateSystem2D();

		position = new Point2D.Double();
		size = new Dimension();
		setCoordinateSystem(cs);
		changeListeners = new Vector();
	}
	
	public GraphArea(CoordinateSystem c)
	{
		position = new Point2D.Double();
		size = new Dimension();
		setCoordinateSystem(c);
		changeListeners = new Vector();
	}
	
	public CoordinateSystem getCoordinateSystem()
	{
		return coord;
	}
	
	private final void setCoordinateSystem(CoordinateSystem c)
	{
		coord = c;
		coord.addChangeListener(this);
	}
	
	/**
	 * 
	 * FIXME - in some cases this should be world coordinates
	 */
	public void setSize(Dimension size)
	{
		this.size = size;
		notifyChange(true, false);
	}

	/**
	 * 
	 * FIXME - in some cases this should be world coordinates
	 */
	public Dimension getSize()
	{
		return this.size;
	}

	public void setPosition(Point2D position)
	{
		this.position = position;
		notifyChange(false, true);
	}

	public Point2D getPosition()
	{
		return this.position;
	}
	
	/**
	 * Returns if the given point in display coordinates
	 * is inside of the graph area or not
	 */
	public boolean isDisplayInside(Point2D p)
	{
		if ( p.getX() >= position.getX() && p.getX()<= position.getX() + size.getWidth() &&
		     p.getY() >= position.getY() && p.getY()<= position.getY() + size.getHeight() ){
			return true;
		}
		return false;
	}

	/**
	 * Returns if the given point in world coordinates
	 * is inside of the graph area or not
	 */
	public boolean isWorldInside(Point2D p)
	{
		Point2D wp = coord.transformToDisplay(p);
		return isDisplayInside(wp);
	}
	
	/**
	 * Handler of the change events in the coordinate system
	 */
	public void stateChanged(ChangeEvent e)
	{
		//Notify the changes
		notifyChange(e);
	}
	 
	/**
	 * Add a listener to any change in the area of in its coordinate system
	 */
	public void addChangeListener(ChangeListener l)
	{
		if (l == null) return;
		changeListeners.add(l);
	}

	/**
	 * Removes a listener to changes in the area of in its coordinate system
	 */
	public void removeChangeListener(ChangeListener l)
	{
		changeListeners.remove(l);
	}
	
	/**
	 * Notifies changes in the graph area
	 */
	public void notifyChange(boolean sizeChanged, boolean positionChanged)
	{
		ChangeEvent evt = new GraphAreaChangeEvent(this, sizeChanged, positionChanged);
		notifyChange(evt);
	}

	/**
	 * Notifies changes in the graph area
	 */
	public void notifyChange(ChangeEvent evt)
	{
		int i;
		ChangeListener l;
		for (i=0;i<changeListeners.size();i++){
			l = (ChangeListener)changeListeners.elementAt(i);
			if (l!=null){
				l.stateChanged(evt);
			}
		}
	}

	/**
	 * Handler of the resizing events of the GraphWindow
	 */
	public void windowResized(GraphWindowResizeEvent e)
	{
		Dimension d = e.getNewSize();
	
//temp hack of casting
		if (coord instanceof AffineTransformCoordinateSystem){
			((AffineTransformCoordinateSystem)coord).setScreenHeight((int)d.getHeight());
		}
		
//temp. the graph area is as big as the graph window
		this.size = d;
//temp. the position of the graph area is lower left corner
		this.position = new Point2D.Double(0,d.getHeight());  // -1 ????

		adjustCoordinateSystem();

		notifyChange(true, true);
	}

	public void windowChanged(EventObject e)
	{
	}

	public Point2D getLowerLeftCornerDisplay()
	{
		return position;
	}

	public Point2D getUpperRightCornerDisplay()
	{
		return new Point2D.Double(position.getX() + size.getWidth(),
			position.getY() - size.getHeight());
	}

	public Point2D getLowerLeftCornerWorld()
	{
		return coord.transformToWorld(getLowerLeftCornerDisplay());
	}

	public Point2D getUpperRightCornerWorld()
	{
		return coord.transformToWorld(getUpperRightCornerDisplay());
	}

	public void setXCentered(boolean b)
	{
		if (centeredX != b){
			centeredX = b;
			adjustCoordinateSystem();
		}
	}

	public void setYCentered(boolean b)
	{
		if (centeredY != b){
			centeredY = b;
			adjustCoordinateSystem();
		}
	}

	public void setOriginCentered(boolean b)
	{
		if (centeredX != b || centeredY != b){
			centeredX = b;
			centeredY = b;
			adjustCoordinateSystem();
		}
	}

    public boolean isXCentered()
	{
		return centeredX;
	}

    public boolean isYCentered()
	{
		return centeredY;
	}
	
    public boolean isOriginCentered()
	{
		if (centeredX && centeredY) return true;
		return false;
	}

	private boolean adjustCoordinateSystem()
	{
		//Modify the default graph area
		CoordinateSystem cs = this.getCoordinateSystem();
	
		if (isOriginCentered()) {
			cs.setOriginOffsetDisplay(new Point2D.Double(getSize().getWidth()/2, getSize().getHeight()/2));
			return true;
		}
		else if (isXCentered()) {
			cs.setOriginOffsetDisplay(new Point2D.Double(getSize().getWidth()/2, cs.getOriginOffsetDisplay().getY()));
			return true;
		}
		else if (isYCentered()) {
			cs.setOriginOffsetDisplay(new Point2D.Double(cs.getOriginOffsetDisplay().getX(), getSize().getHeight()/2));
			return true;
		}
		return false;
	}
}
    
