/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.1 $
 * $Date: 2004/05/24 20:16:30 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.*;

import org.concord.graph.event.*;

/**
 * The GraphableList class is a Vector that contains Graphables 
 * and fires events when objects have been added, removed or changed.
 *
 * @author Ingrid Moncada<p>
 *
*/
public class GraphableList extends Vector
	implements GraphableListener
{
	protected Vector listeners = new Vector();

	/////////////////////
	// Constructors (dont get inherited automatically)
	/////////////////////

	/**
     * Constructs an empty vector so that its internal 
	 * data array has size 10 and its standard capacity increment is zero.
	*/
	public GraphableList()
	{
		super();
	}

	/**
	Constructs a vector containing the elements of the specified collection, in the order they are returned by the collection's iterator.
	*/
	public GraphableList(Collection c)
	{
		super(c);
	}  

	/**
          Constructs an empty vector with the specified initial capacity and with its capacity increment equal to zero. 
	*/
	public GraphableList(int initialCapacity)
	{
		super(initialCapacity);
	}
	
	/**
	  Constructs an empty vector with the specified initial capacity and capacity increment.
	*/
	public GraphableList(int initialCapacity, int capacityIncrement)
	{
		super(initialCapacity, capacityIncrement);
	}  
	/////////////////////
	
	/////////////////////
	// Add methods
	// They add a listener if the object added is a Graphable
	// They also fire a listGraphableAdded(obj) event.

	/**
	 * Adds the specified element to the list in the index
	 * Fires a listGraphableAdded(obj) event
	 */
	public void add(int index, Object element)
	{
		super.add(index,element);
		if (element instanceof Graphable){
			addListenerToObject((Graphable)element);
		}
		notifyAdd(element);
	}

	/**
	 * Adds the specified object to the list
	 * Fires a listGraphableAdded(obj) event
	 */
	public boolean add(Object o)
	{
		if (super.add(o)){
			if (o instanceof Graphable){
				addListenerToObject((Graphable)o);
			}
			notifyAdd(o);
			return true;
		}
		return false;
	}
	
	/**
	 * Adds the specified object to the list
	 * Fires a listGraphableAdded(obj) event
	 */
	public void addElement(Object obj)
	{
		super.addElement(obj);
		if (obj instanceof Graphable){
			addListenerToObject((Graphable)obj);
		}
		notifyAdd(obj);
	}

	private void addListenerToObject(Graphable obj)
	{
		obj.addGraphableListener(this);
	}

	/////////////////////
	// Remove methods
	
	//NOOOOOOOOOOOOOOOOOOOO
	//Modified !!!!
	//It does NOT call the remove method
	
	// They call the remove() method on the Graphable object.
	// They do NOT delete the object from the Vector, 
	// because since its listening to the events on the Graphable, 
	// it will receive the graphableRemoved() event and then 
	// it will update the list internal state 
	// (deleting the object from its list).
	// For the same reason, it does NOT remove the listener to the
	// object, and it does NOT fire the removed event.

	/**
	 * Calls the remove() method on the Selectable located at the specified index
	 * If the object is not Selectable, it removes it and fires the removed event
	 */
	public Object remove(int index)
	{
		Object obj = null;
		if (index>=0 && index<this.size()){
			obj = this.elementAt(index);
			if (obj==null) return null;
			removeInternal(obj);
		}
		
		return obj;
	}
	
	/**
	 * Calls the remove() method on the specified Selectable 
	 * If the object is not Selectable, it removes it and fires the removed event
	 */
	public boolean remove(Object obj)
	{
		if (obj==null) return false;
		if (this.contains(obj)){
			removeInternal(obj);
		}
		return false;
	}
  
  	protected void removeInternal(Object obj)
	{
		if (obj instanceof Graphable){
			removeListenerToObject((Graphable)obj);
		}
		super.remove(obj);
		notifyRemove(obj);
	}
	
	/**
	 * It does NOT call the remove() method on the Selectables 
	 * It removes all objects from the list and fires the removed event
	 */
  	public void removeAllElements()
	{
		removeAllElements(false);
	}

	/**
	 * Calls the remove() method on the all Selectable, if indicated by the parameter 
	 * If any object is not Selectable, it removes it and fires the removed event
	 */
  	public void removeAllElements(boolean removeGraphables)
	{
		int i;
		int t = this.size();
		for (i=0;i<t;i++){
			//Always take element 0 because I'm deleting all objects
			Object obj = this.elementAt(0);
			if (obj instanceof Graphable){
				if (removeGraphables){
				//The object is a Graphable, so it removes "himself"
				//(this class is listening to its removed event)
					((Graphable)obj).remove();
				}
				else{
					removeInternal(obj);
				}
			}
			else{
				if (super.remove(obj)){
					notifyRemove(obj);
				}
			}
		}
	}

	private void removeListenerToObject(Graphable obj)
	{
		obj.removeGraphableListener(this);
	}
	/////////////////////

	/**
	 * Registers an object as a listener to the ObjectListEvents.
	 */
	public void addGraphableListListener(GraphableListListener l)
	{
		if (!listeners.contains(l)){
			listeners.add(l);
		}
	}
 
 	/**
	 * Removes an object from the listeners to the ObjectListEvents.
	 */
	public void removeGraphableListListener(GraphableListListener l)
	{
		listeners.remove(l);
	}
	
	/////////////////////
	// Listener to the Graphable objects
	// (Graphable events handler)

	/*
	 * Handles the graphable change event from the Graphable
	 * objects in its list. When it receives this event, 
	 * it fires a listGraphableChanged(obj) event.
	 */
	public void graphableChanged(EventObject e)
	{
		notifyChange(e.getSource());
	} 

	/**
	 * Handles the graphable removed event from the Graphable
	 * objects in its list. When it receives this event, 
	 * it removes the listener to the object, 
	 * removes the object from the list, 
	 * and fires a listGraphableRemoved(obj) event.
	 */
	public void graphableRemoved(EventObject e)
	{ 
		removeInternal(e.getSource());
	}

	/////////////////////

	/**
	 * Fires a listGraphableAdded(obj) event.
	 */
	protected void notifyAdd(Object obj)
	{
		if (obj==null) return;
		GraphableListListener l;
		EventObject e = new EventObject(obj); 
		for (int i=0;i<listeners.size();i++){
			l = (GraphableListListener)listeners.elementAt(i);
			l.listGraphableAdded(e);
		}
	}

	/**
	 * Fires a listGraphableChanged(obj) event.
	 */
	protected void notifyChange(Object obj)
	{
		if (obj==null) return;
		GraphableListListener l;
		EventObject e = new EventObject(obj); 
		for (int i=0;i<listeners.size();i++){
			l = (GraphableListListener)listeners.elementAt(i);
			l.listGraphableChanged(e);
		}
	}
	
	/**
	 * Fires a listGraphableRemoved(obj) event.
	 */
	protected void notifyRemove(Object obj)
	{
		if (obj==null) return;
		GraphableListListener l;
		EventObject e = new EventObject(obj); 
		for (int i=0;i<listeners.size();i++){
			l = (GraphableListListener)listeners.elementAt(i);
			l.listGraphableRemoved(e);
		}
	}
}
