/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.6 $
 * $Date: 2004/07/14 17:34:57 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.engine;

import java.util.*;

import org.concord.graph.event.*;

/**
 * The SelectableList class is an GraphableList that contains Selectable objects.
 * It also provides the capability of select objects and maintains a list of selected objects. 
 * Its selection model currently includes single selection and multiple selection. 
 * In single selection model, only one object can be selected at a time, so when an object is 
 * selected, all the other objects are automatically deselected. In the multiple selection model, 
 * multiple objects can be selected at a time. There is also a selection mode that doesn't allow 
 * any selection. 
 * It works with the capability of Selectable objects of being selected, and notify about selection. 
 * The SelectableList listens to each Selectable object event (selected, deselected, changed, removed) 
 * and notifies the events back with GraphableList events. It also calls the methods remove, select, 
 * deselect in the Selectable objects whenever an object is selected, deselected and removed through the list. 
 * Besides the events fired by the GraphableList class (added, removed, changed), 
 * the SelectableList also fires events when an object has been selected and deselected.
 * 
 * Date created: 04/13/2004
 *
 * @author Ingrid Moncada<p>
 *
*/
public class SelectableList extends GraphableList
	implements Selectable, SelectableListener
{
	/**
	 * This mode does not allow selection
	 */
    public static final int SELECTIONMODE_NONE = 0;

	/**
	 * In single selection mode, only one object can be selected at a time, so when an object is 
	 * selected, all the other objects are automatically deselected
	 */
    public static final int SELECTIONMODE_SINGLE = 2;

	/*
	 * In the multiple selection model, 
	 * multiple objects can be selected at a time.
	 */
	public static final int SELECTIONMODE_MULTIPLE = 3;

	/**
	 * This mode does not allow selection
	 */
    public static final int MOUSEMODE_NONE = 0;

	/**
	 * In single mouse mode, the selected objects
	 * of the list will receive the mouse events,
	 * but if one object is a list itself, its 
	 * elements will not be notified.
	 */
    public static final int MOUSEMODE_SINGLE = 2;

	/*
	 * In multiple mouse mode, the selected objects
	 * of the list will receive the mouse events,
	 * AND if one object is a list itself, its 
	 * elements will ALSO be notified.
	 */
	public static final int MOUSEMODE_MULTIPLE = 3;

	private Vector selectableListeners; 
	private Vector selectedObjects;
	private Vector lastSelectedObjects;
	private boolean selectable = true;
		
	private int selectionMode = SELECTIONMODE_SINGLE;
	private int mousePropagationMode = MOUSEMODE_SINGLE;

	public SelectableList()
	{
		selectableListeners = new Vector();
		selectedObjects = new Vector();
	}
	
	/////////////////////
	// Add methods
	// They add a listener if the object added is a Selectable

	/**
	 * Adds the specified element to the list in the index
	 * Fires a listGraphableAdded(obj) event
	 */
	public void add(int index, Object element)
	{
		super.add(index,element);
		if (element instanceof Selectable){
			addListenerToObject((Selectable)element);
		}
	}

	/**
	 * Adds the specified object to the list
	 * Fires a listGraphableAdded(obj) event
	 */
	public boolean add(Object o)
	{
		if (super.add(o)){
			if (o instanceof Selectable){
				addListenerToObject((Selectable)o);
			}
			return true;
		}
		return false;
	}
	
	/**
	 * Adds the specified object to the list
	 * Fires a listGraphableAdded(obj) event
	 */
	public void addElement(Object obj)
	{
		super.addElement(obj);
		if (obj instanceof Selectable){
			addListenerToObject((Selectable)obj);
		}
	}

	private void addListenerToObject(Selectable obj)
	{
		obj.addSelectableListener(this);
	}

	/////////////////////
	// Remove methods
	
	private void removeListenerToObject(Selectable obj)
	{
		obj.removeSelectableListener(this);
	}
	/////////////////////

	/**
	 * Returns a Vector of ALL the current selected objects only.
	 * (INCLUDING SUB LISTS)
	 * If there are no objects in the ObjectList, or if there are no objects 
	 * currently selected, it returns an empty vector.
	 */
	public Vector getMousePropagationObjects()
	{
//		return getSelectedObjects(true);
		//Construct a vector
		Vector v = new Vector();
		Enumeration e = elements();
		while (e.hasMoreElements()){
			Object obj = e.nextElement();
			if (mousePropagationMode == MOUSEMODE_MULTIPLE){
				if (obj instanceof SelectableList){
					v.addAll(((SelectableList)obj).getSelectedObjects());
					continue;
				}
			}
			if (obj instanceof Selectable){
				if (((Selectable)obj).isSelected()){
					v.add(obj);
				}
			}
		}
		return v;
	}

	/**
	 * Returns a Vector of ALL the current selected objects only.
	 * (INCLUDING OBJECTS FROM THE SUB LISTS)
	 * If there are no objects in the ObjectList, or if there are no objects 
	 * currently selected, it returns an empty vector.
	 */
	public Vector getSelectedObjects()
	{
		return getSelectedObjects(true);
	}

	/**
	 * Returns a Vector of the current selected objects only,
	 * if recursive is true then the selected objects in the sublists
	 *   are included.
	 * if recursive is false then the selected sub lists themself are included.
	 * If there are no objects in the ObjectList, or if there are no objects 
	 * currently selected, it returns an empty vector.
	 */
	public Vector getSelectedObjects(boolean recursive)
	{
		//Construct a vector
		Vector v = new Vector();
		Enumeration e = elements();
		while (e.hasMoreElements()){
			Object obj = e.nextElement();
			if (recursive){
				if (obj instanceof SelectableList){
					v.addAll(((SelectableList)obj).getSelectedObjects());
					continue;
				}
			}
			if (obj instanceof Selectable){
				if (((Selectable)obj).isSelected()){
					v.add(obj);
				}
			}
		}
		return v;
	}

	/**
	 * Returns the first object in the list of selected objects. 
	 * This object should coincide with the object that was first 
	 * selected within all the current selected objects. 
	 * If there are no objects in the ObjectList, or if there are no objects 
	 * currently selected, it returns null.
	 */
	public Object getFirstSelectedObject()
	{
		if (selectedObjects.size()>0){
			return selectedObjects.elementAt(0);
		}
		return null;
	}
	
	/**
	 * Returns the last object in the list of selected objects. 
	 * This object should coincide with the object that was 
	 * selected LAST, within all the current selected objects. 
	 * If there are no objects in the ObjectList, or if there are no objects 
	 * currently selected, it returns null.
	 */
	public Object getLastSelectedObject()
	{
		if (selectedObjects.size()>0){
			return selectedObjects.elementAt(selectedObjects.size()-1);
		}
		return null;
	}

	/** 
	 * Returns the index of the first object in the list of selected objects. 
	 * This object should coincide with the object that was first selected 
	 * within all the current selected objects. If there are no objects in 
	 * the ObjectList, or if there are no objects currently selected, it returns -1.
	 */
	public int getFirstSelectedIndex()
	{
		return indexOf(getFirstSelectedObject());
	}

	/**
	 * Changes the selection mode. 
	 * Currently, the selection modes are SELECTIONMODE_NONE, SELECTIONMODE_SINGLE 
	 * and SELECTIONMODE_MULTIPLE. The selection model by default is SELECTIONMODE_SINGLE.
	 */	
	public void setSelectionMode(int mode)
	{
		selectionMode = mode;
	}

	/**
	 * Returns the current selection mode.
	 */
	public int getSelectionMode()
	{
		return selectionMode;
	}

	/**
	 * Returns the mouse propagation mode
	 * @return
	 */
	public int getMousePropagationMode()
	{
		return mousePropagationMode;
	}
	
	/**
	 * Sets the mouse propagation mode
	 * @param mousePropagationMode MOUSEMODE_NONE, 
	 * MOUSEMODE_SINGLE, MOUSEMODE_MULTIPLE
	 */
	public void setMousePropagationMode(int mousePropagationMode)
	{
		this.mousePropagationMode = mousePropagationMode;
	}

	/////////////////////
	// Selection methods
	// They call the select() method on the Selectable object.
	// They do NOT keep track of the object as being selected yet, 
	// because since it's listening to the events on the selectable, 
	// it will receive the event and then it will update the list internal state 
	// (keeping track of the object as being selected).
	// For the same reason, they do NOT fire the selected list event.

	/**
	 * Calls the select() method on the Selectable located at the specified index
	 * If the object is not Selectable, it keeps track of it
	 * as selected and fires the selection event
	 */
	public void select(int index)
	{
		//Don't allow selection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		
		if (index>=0 && index<this.size()){
			Object obj = elementAt(index);
			if (obj==null) return;
			if (obj instanceof Selectable){
			//The object is a Selectable, so it selects "himself"
			//(this class is listening to its selected event)
				Selectable sel = (Selectable)obj;
				if (!sel.isSelected()){
					sel.select();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as selected, and fire the event
				selectInternal(obj);
			}
		}
	}

	/**
	 * Calls the select() method on the specified Selectable
	 * If the object is not Selectable, it keeps track of it
	 * as selected and fires the selection event
	 */
	public void select(Object obj)
	{
		if (obj==null) return;
		//Don't allow selection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		if (this.contains(obj)){
			if (obj instanceof Selectable){
			//The object is a Selectable, so it selects "himself"
			//(this class is listening to its selected event)
				Selectable sel = (Selectable)obj;
				if (!sel.isSelected()){
					sel.select();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as selected, and fire the event
				selectInternal(obj);
			}
		}
	}

	/**
	 * Calls the deselect() method on the Selectable located at the specified index
	 * If the object is not Selectable, it keeps track of it
	 * as deselected and fires the deselection event
	 */
	public void deselect(int index)
	{
		//Don't allow selection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		
		if (index>=0 && index<this.size()){
			Object obj = elementAt(index);
			if (obj==null) return;
			if (obj instanceof Selectable){
			//The object is a Selectable, so it deselects "himself"
			//(this class is listening to its deselected event)
				Selectable sel = (Selectable)obj;
				if (sel.isSelected()){
					sel.deselect();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as deselected, and fire the event
				deselectInternal(obj);
			}
		}
	}

	/**
	 * Calls the deselect() method on the specified Selectable
	 * If the object is not Selectable, it keeps track of it
	 * as deselected and fires the deselection event
	 */
	public void deselect(Object obj)
	{
		if (obj==null) return;
		//Don't allow selection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		if (this.contains(obj)){
			if (obj instanceof Selectable){
			//The object is a Selectable, so it deselects "himself"
			//(this class is listening to its deselected event)
				Selectable sel = (Selectable)obj;
				if (sel.isSelected()){
					sel.deselect();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as deselected, and fire the event
				deselectInternal(obj);
			}
		}
	}

	/**
	 * Calls the select() method on all Selectables
	 * If any object is not Selectable, it keeps track of it
	 * as selected and fires the selection event
	 */
	public void selectAll()
	{
		int i;
		//Don't allow selection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		for (i=0;i<this.size();i++){
			Object obj = this.elementAt(i);
			if (obj instanceof Selectable){
			//The object is a Selectable, so it selects "himself"
			//(this class is listening to its selected event)
				Selectable sel = (Selectable)obj;
				if (!sel.isSelected()){
					sel.select();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as selected, and fire the event
				selectInternal(obj);
			}
		}
	}

	/**
	 * Calls the deselect() method on all Selectables
	 * If any object is not Selectable, it keeps track of it
	 * as deselected and fires the deselection event
	 */
	public void deselectAll()
	{
		int i;
		int t;
		//Don't allow deselection with mode SELECTIONMODE_NONE
		if (selectionMode == SELECTIONMODE_NONE) return;
		t = selectedObjects.size();
		for (i=0;i<t;i++){
			//Always take element 0 because I'm removing the objects from the selected list
			Object obj = selectedObjects.elementAt(0);
			if (obj instanceof Selectable){
			//The object is a Selectable, so it selects "himself"
			//(this class is listening to its selected event)
				Selectable sel = (Selectable)obj;
				if (sel.isSelected()){
					sel.deselect();
				}
			}
			else{
			//The object is not a Selectable, so it has to 
			//physcally keep track of the object as selected, and fire the event
				deselectInternal(obj);
			}
		}
	}
	
	/** 
	 * Keeps track of the object as a selected object
	 * It also fires a listGraphableSelected(obj) event
	 */
	protected void selectInternal(Object obj)
	{
		//Don't add the selection if it was already done (???)
		if (selectedObjects.contains(obj)) return;
		selectedObjects.add(obj);
		notifySelect(obj);
		if (selectionMode == SELECTIONMODE_SINGLE){
			adjustSelected(1);
		}
		notifySelect();
	}

	/** 
	 * Keeps track of the object as a deselected object
	 * It also fires a listGraphableDeselected(obj) event
	 */
	protected void deselectInternal(Object obj)
	{
		selectedObjects.remove(obj);
		notifyDeselect(obj);
	}

	/** 
	 * Adjusts the selection to a maximum of objects selected
	 */
	protected void adjustSelected(int maxNumberSelected)
	{
		int t;
		t = selectedObjects.size();
		while (t>0 && t>maxNumberSelected){
			//Always take element 0 because I'm removing the objects from the selected list
			deselect(selectedObjects.elementAt(0));
			t = selectedObjects.size();
		}
	}
	
	/**
	 * Removes all the selected elements from the list
	 */
	public void removeAllSelected()
	{
		int i;
		int t;
		Graphable g;
		t = selectedObjects.size();
		for (i=0;i<t;i++){
			//Always take element 0 because I'm removing the objects from the selected list
			g = (Graphable)selectedObjects.elementAt(0);
			//The object is a Graphable, so it removes "himself"
			//(this class is listening to its removed event)
			g.remove();
		}
	}
	
	/**
	 * Overrides the removeInternal event handling 
	 * from Graphable List.
	 * When removed, it also 
	 * removes the object from the list of selected
	 */
  	protected void removeInternal(Object obj)
	{ 
		//If it is selected, deselect it first
		if (obj instanceof Selectable){
			((Selectable)obj).deselect();
			removeListenerToObject((Selectable)obj);
		}
		
		super.removeInternal(obj);
		
		//This is not necessary anymore because the object is being deselected first
		////Delete also from the selected objects
		//selectedObjects.remove(obj);
	}

	/////////////////////

	/////////////////////
	// Listener to the Graphable objects
	// (Graphable events handler)

	/**
	 * When it receives this event, it keeps track internally of the object as selected. 
	 * It also fires a listGraphableSelected(obj) event (in the parent class).
	 * <TODO> The behavior of these listeners regarding selection behavior is not clear
	 * i.e. what to do when received a selection event but the
	 * selection mode is SELECTIONMODE_NONE </TODO>
	 */
	public void graphableSelected(EventObject e) 
	{
		selectInternal(e.getSource());	//This also notifies selection
	}

	/* When it receives this event, it keeps track internally of the object as NOT selected.
	 * It also fires a listGraphableDeselected (obj) event (in the parent class).
	 */
	public void graphableDeselected(EventObject e)
	{
		deselectInternal(e.getSource());	//This also notifies deselection
	}
	/////////////////////

	/**
	 * Fires a listGraphableRemoved(obj) event.
	 */
	protected void notifySelect(Object obj)
	{
		GraphableListListener l;
		EventObject e = new EventObject(obj); 

		for (int i=0;i<listeners.size();i++){
			l = (GraphableListListener)listeners.elementAt(i);
			if (l instanceof SelectableListListener){
				((SelectableListListener)l).listGraphableSelected(e);
			}
		}
	}

	protected void notifyDeselect(Object obj)
	{
		GraphableListListener l;
		EventObject e = new EventObject(obj); 
		for (int i=0;i<listeners.size();i++){
			l = (GraphableListListener)listeners.elementAt(i);
			if (l instanceof SelectableListListener){
				((SelectableListListener)l).listGraphableDeselected(e);
			}
		}
	}
	
	
	////////////////////////////////////
	// Methods from interface Selectable
	////////////////////////////////////
	/**
	 * Selects this object. 
	 * This also fires an objectSelected(obj) event.
	 */
	public void select()
	{
		if (lastSelectedObjects == null) return;
		
		int i;
		int j;
		for (i=0; i < lastSelectedObjects.size(); i++){
			j = this.indexOf(lastSelectedObjects.elementAt(i));
			if (j != -1){
				this.select(j);
			}
		}
		notifySelect();
	}
	
	/** Deselects this object
	 * This also fires an objectDeselected(obj) event.
	 */
	public void deselect()
	{
		lastSelectedObjects = (Vector)selectedObjects.clone();
		this.deselectAll();
		notifyDeselect();
	}
	
	/**
	 * Returns true if at least ONE obejct is selected 
	 */
	public boolean isSelected()
	{
		Enumeration e = elements();
		while (e.hasMoreElements()){
			Object obj = e.nextElement();
			if (obj instanceof Selectable){
				if (((Selectable)obj).isSelected()){
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Sets if the object can be selected or not 
	 */
	public void setSelectionEnabled(boolean b)
	{
		selectable = b;
	}

	/**
	 * Returns if the object can be selected or not 
	 */
	public boolean isSelectionEnabled()
	{
		return selectable;
	}

	/*
	 * Adds a SelectableListener to listen for selection events 
	 */
	public void addSelectableListener(SelectableListener l)
	{
		selectableListeners.add(l);
	}
	
	/*
	 * Removes a SelectableListener.
	 */
	public void removeSelectableListener(SelectableListener l)
	{
		selectableListeners.remove(l);
	}

	protected void notifySelect() 
	{		
		EventObject e = new EventObject(this);
		Enumeration lists = selectableListeners.elements();
		while (lists.hasMoreElements()) {			
			GraphableListener gl = (GraphableListener)lists.nextElement();
			if (gl instanceof SelectableListener) {
				((SelectableListener)gl).graphableSelected(e);
			}
		}
	}

	protected void notifyDeselect() 
	{
		EventObject e = new EventObject(this);
		Enumeration lists = selectableListeners.elements();
		while (lists.hasMoreElements()) {			
			GraphableListener gl = (GraphableListener)lists.nextElement();
			if (gl instanceof SelectableListener) {
				((SelectableListener)gl).graphableDeselected(e);
			}
		}
	}
	///////////////////////////////////////
	
}
