/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.4 $
 * $Date: 2004/06/19 02:59:44 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.examples;

import java.awt.*;
import java.awt.geom.*;
import java.util.EventObject;
import java.util.Vector;

import org.concord.graph.engine.*;
import org.concord.graph.event.CursorListener;

/**
 * Box drawing class.  It can be used for selection
 * boxes or zoom boxes
 * 
 * @author Scott Cytacki<p>
 *
*/
public class DashedBox extends DefaultControllable
	implements Cursorable
{
	private GeneralPath path = null;
	
	private Color lineColor = Color.BLACK;
	private float lineWidth = 2;

	public static final int DRAGMODE_NONE = 0; 
	public static final int DRAGMODE_TRANSLATE = 1; 
	public static final int DRAGMODE_RESIZE_N = 2; 
	public static final int DRAGMODE_RESIZE_S = 3; 
	public static final int DRAGMODE_RESIZE_E = 4; 
	public static final int DRAGMODE_RESIZE_W = 5; 
	public static final int DRAGMODE_RESIZE_NE = 6; 
	public static final int DRAGMODE_RESIZE_NW = 7; 
	public static final int DRAGMODE_RESIZE_SE = 8; 
	public static final int DRAGMODE_RESIZE_SW = 9; 
	
	private int dragMode = DRAGMODE_NONE;
	private boolean newBox = false;
	
	private Rectangle2D.Float bounds = new Rectangle2D.Float();

	protected Rectangle2D originalBounds;		//Bounds of the object when the mouse was clicked in world coordinates

	private Cursor currentCursor;
	private Vector cursorListeners;

	public DashedBox()
	{
		cursorListeners = new Vector();
	}
	
	public DashedBox(boolean newBox)
	{
		this();
		this.newBox = newBox;
	}

	public void setBounds(float x, float y, float width, float height)
	{
		//Maintain width and height always positive 
		if (height < 0){
			y = y + height;
			height = -height;
		}
		if (width < 0){
			x = x + width;
			width = -width;
		}
		bounds = new Rectangle2D.Float(x,y,width,height);
		location = new Point2D.Double((double)x, (double)y);
		notifyChange();
	}

	public Rectangle2D.Float getBounds()
	{
		return bounds;
	}

	public void setColor(int r, int g, int b)
	{
		lineColor = new Color(r,g,b);
	}

	public void setLineWidth(float width)
	{
		lineWidth = width;
	}

	public void setLocation(Point2D p)
	{
		super.setLocation(p);
		setBounds((float)p.getX(), (float)p.getY(), (float)bounds.getWidth(), (float)bounds.getHeight());
	}
	
    /** Draws this object on Graphics g */
    public void draw(Graphics2D g)
	{
		update();

		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		
		g.setColor(lineColor);
		float dash[] = {10.0f};
		
		if (isSelected()){
			g.setStroke(new BasicStroke(lineWidth+1, BasicStroke.CAP_BUTT,
										BasicStroke.JOIN_MITER, 5.0f, dash, 0.0f));
		}
		else{
			g.setStroke(new BasicStroke(lineWidth, BasicStroke.CAP_BUTT,
										BasicStroke.JOIN_MITER, 5.0f, dash, 0.0f));
		}			
		g.draw(path);
		
		g.setColor(oldColor);
		g.setStroke(oldStroke);
	}

    /** Forces classes to redraw any internal image buffer (if implemented).
 * 	This makes sure the background is updated if there is a fundamental
 * 	change that is not checked by this class. */
    public void update()
	{
		if (path == null){
			path = new GeneralPath();
		}
		path.reset();
		
		CoordinateSystem coord = getGraphArea().getCoordinateSystem();

		float ppx, ppy, startX;
		float px, py, startY;

		Point2D.Double pathPoint = new Point2D.Double();

		Point2D.Double dataPoint = new Point2D.Double(bounds.x, bounds.y);
		coord.transformToDisplay(dataPoint, pathPoint);
		ppx = (float)pathPoint.x;
		ppy = (float)pathPoint.y;
		startX = ppx;
		startY = ppy;
		path.moveTo(ppx, ppy);

		dataPoint.setLocation(bounds.x + bounds.width, bounds.y);
		coord.transformToDisplay(dataPoint, pathPoint);
		ppx = (float)pathPoint.x;
		ppy = (float)pathPoint.y;
		path.lineTo(ppx, ppy);

		dataPoint.setLocation(bounds.x + bounds.width, 
							  bounds.y + bounds.height);
		coord.transformToDisplay(dataPoint, pathPoint);
		ppx = (float)pathPoint.x;
		ppy = (float)pathPoint.y;
		path.lineTo(ppx, ppy);

		dataPoint.setLocation(bounds.x, bounds.y + bounds.height);
		coord.transformToDisplay(dataPoint, pathPoint);
		ppx = (float)pathPoint.x;
		ppy = (float)pathPoint.y;
		path.lineTo(ppx, ppy);

		path.lineTo(startX, startY);

	}

	/** 
	 * Returns a copy of itself 
	 */
	public Graphable getCopy()
	{
		return new DashedBox();
	}

	/**
	 * Checks if the location is within enough proximity to be selected. 
	 * If the location is close enough of the object and
	 * the object wants to be selected, returns true. 
	 * The location is given in screen coordinates. 
	 * The axis is provided to transform coordinates.
	 */
	public boolean isPointInProximity(Point location)
	{
		if (newBox){
			setDragMode(DRAGMODE_RESIZE_SE);
			return true;
		}
		
		CoordinateSystem coord = graphArea.getCoordinateSystem();
		
		Point2D sw, ne;// nw, se;
		double threshold = 7.0;
		
		double xmin, xmax, ymin, ymax, temp;
		
		xmin = bounds.getX();
		sw = coord.transformToDisplay(new Point2D.Double(bounds.getX(),bounds.getY()));
		//nw = coord.transformToDisplay(new Point2D.Double(bounds.getX(),bounds.getY() + bounds.getHeight()));
		//se = coord.transformToDisplay(new Point2D.Double(bounds.getX() + bounds.getWidth(),bounds.getY()));
		ne = coord.transformToDisplay(new Point2D.Double(bounds.getX() + bounds.getWidth(),bounds.getY() + bounds.getHeight()));

		xmin = sw.getX();
		xmax = ne.getX();
		ymin = ne.getY();
		ymax = sw.getY();

		if (xmin > xmax){
			temp = xmin;
			xmin = xmax;
			xmax = temp;
		}
		if (ymin > ymax){
			temp = ymin;
			ymin = ymax;
			ymax = temp;
		}

		//System.out.println("xmin:"+xmin+" xmax:"+xmax);
		//System.out.println("ymin:"+ymin+" ymax:"+ymax);
		//System.out.println("--------");

		if (Math.abs(location.getX() - xmin) <= threshold){ 
			if (Math.abs(location.getY() - ymax) <= threshold){
				setDragMode(DRAGMODE_RESIZE_SW);
				return true;
			}
			else if (Math.abs(location.getY() - ymin) <= threshold){
				setDragMode(DRAGMODE_RESIZE_NW);
				return true;
			}
			else if (location.getY() > ymin && location.getY() < ymax){
				setDragMode(DRAGMODE_RESIZE_W);
				return true;
			}
		}
		else if (Math.abs(location.getX() - xmax) <= threshold){ 
			if (Math.abs(location.getY() - ymax) <= threshold){
				setDragMode(DRAGMODE_RESIZE_SE);
				return true;
			}
			else if (Math.abs(location.getY() - ymin) <= threshold){
				setDragMode(DRAGMODE_RESIZE_NE);
				return true;
			}
			else if (location.getY() > ymin && location.getY() < ymax){
				setDragMode(DRAGMODE_RESIZE_E);
				return true;
			}
		}
		else if (Math.abs(location.getY() - ymax) <= threshold){ 
			if (location.getX() > xmin && location.getX() < xmax){
				setDragMode(DRAGMODE_RESIZE_S);
				return true;
			}
		}
		else if (Math.abs(location.getY() - ymin) <= threshold){ 
			if (location.getX() > xmin && location.getX() < xmax){
				setDragMode(DRAGMODE_RESIZE_N);
				return true;
			}
		}

		if (path.contains(location)){
			setDragMode(DRAGMODE_TRANSLATE);
			return true;
		}
		return false;
	}
	
	public void setDragMode(int dragMode)
	{
		if (this.dragMode != dragMode){
			this.dragMode = dragMode;
			updateCursor();
		}
	}
	
	private void updateCursor()
	{
		if (dragMode == DRAGMODE_NONE){
			setCursor(null);
		}
		else if (dragMode == DRAGMODE_TRANSLATE){
			setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_N){
			setCursor(Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_S){
			setCursor(Cursor.getPredefinedCursor(Cursor.S_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_E){
			setCursor(Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_W){
			setCursor(Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_NE){
			setCursor(Cursor.getPredefinedCursor(Cursor.NE_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_NW){
			setCursor(Cursor.getPredefinedCursor(Cursor.NW_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_SE){
			setCursor(Cursor.getPredefinedCursor(Cursor.SE_RESIZE_CURSOR));
		}
		else if (dragMode == DRAGMODE_RESIZE_SW){
			setCursor(Cursor.getPredefinedCursor(Cursor.SW_RESIZE_CURSOR));
		}
	}

	/**
	 * @see org.concord.graph.engine.MouseControllable#mousePressed(java.awt.Point)
	 */
	public boolean mousePressed(Point p)
	{
		if (super.mousePressed(p)){
			if (newBox){
				setBounds((float)clickPointWorld.getX(), 
						(float)clickPointWorld.getY(),0,0);
			}
			originalBounds = bounds.getBounds2D();
		}
		
		return false;
	}
	
	/**
	 * @see org.concord.graph.engine.MouseControllable#mouseDragged(java.awt.Point)
	 */
	public boolean mouseDragged(Point p)
	{
		if (dragMode == DRAGMODE_TRANSLATE){
			return super.mouseDragged(p);
		}
		
		Point2D newLocation;
		
		newLocation = graphArea.getCoordinateSystem().transformToWorld(p);

		double distanceX;
		double distanceY;
		if (dragMode == DRAGMODE_RESIZE_N){
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)originalBounds.getX(), (float)originalBounds.getY(), 
					(float)originalBounds.getWidth(), (float)(originalBounds.getHeight() + distanceY));
		}
		else if (dragMode == DRAGMODE_RESIZE_S){
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)originalBounds.getX(), (float)(originalBounds.getY() + distanceY), 
					(float)originalBounds.getWidth(), (float)(originalBounds.getHeight() - distanceY));
		}
		else if (dragMode == DRAGMODE_RESIZE_E){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			setBounds((float)originalBounds.getX(), (float)originalBounds.getY(), 
					(float)(originalBounds.getWidth() + distanceX), (float)originalBounds.getHeight());
		}
		else if (dragMode == DRAGMODE_RESIZE_W){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			setBounds((float)(originalBounds.getX() + distanceX), (float)originalBounds.getY(), 
					(float)(originalBounds.getWidth() - distanceX), (float)originalBounds.getHeight());
		}
		else if (dragMode == DRAGMODE_RESIZE_NE){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)originalBounds.getX(), (float)originalBounds.getY(), 
					(float)(originalBounds.getWidth() + distanceX), (float)(originalBounds.getHeight() + distanceY));
		}
		else if (dragMode == DRAGMODE_RESIZE_NW){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)(originalBounds.getX() + distanceX), (float)originalBounds.getY(), 
					(float)(originalBounds.getWidth() - distanceX), (float)(originalBounds.getHeight() + distanceY));
		}
		else if (dragMode == DRAGMODE_RESIZE_SE){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)originalBounds.getX(), (float)(originalBounds.getY() + distanceY), 
					(float)(originalBounds.getWidth() + distanceX), (float)(originalBounds.getHeight() - distanceY));
		}
		else if (dragMode == DRAGMODE_RESIZE_SW){
			distanceX = newLocation.getX() - clickPointWorld.getX();
			distanceY = newLocation.getY() - clickPointWorld.getY();
			setBounds((float)(originalBounds.getX() + distanceX), (float)(originalBounds.getY() + distanceY), 
					(float)(originalBounds.getWidth() - distanceX), (float)(originalBounds.getHeight() - distanceY));
		}
		
		return true;
	}
	
	/**
	 * @see org.concord.graph.engine.MouseControllable#mouseReleased(java.awt.Point)
	 */
	public boolean mouseReleased(Point p)
	{
		newBox = false;
		return super.mouseReleased(p);
	}


	//////////////////////////////////////

	//////////////////////////////////////
	// Methods from Cursorable
	//////////////////////////////////////
	/**
	 * Sets the current cursor
	 */
	public void setCursor(Cursor cursor)
	{
		if (currentCursor != cursor){
			currentCursor = cursor;
			notifyCursorListeners();
		}
	}

	/**
	 * Returns the current cursor
	 */
	public Cursor getCursor()
	{
		return currentCursor;
	}
	
	/**
	 * Adds a listener that will listen to a cursor change
	 */
	public void addCursorListener(CursorListener l)
	{
		cursorListeners.add(l);
	}

	/**
	 * Removes a cursor change listener 
	 */
	public void removeCursorListener(CursorListener l)
	{
		cursorListeners.remove(l);
	}
	
	protected void notifyCursorListeners()
	{
		CursorListener l;
		for (int i=0;i<cursorListeners.size();i++){
			l = (CursorListener)cursorListeners.elementAt(i);
			l.cursorChanged(new EventObject(this));
		}
	}

	//////////////////////////////////////
	
	public void zoom()
	{
		//Change the coordinate system and the graph area to zoom
		Dimension size = graphArea.getSize();	
		CoordinateSystem coord = graphArea.getCoordinateSystem();

		Point2D.Double scale = new Point2D.Double(size.width/bounds.width, 
												  size.height/bounds.height);

		coord.setScale(scale);

		Point2D lowerLeft = coord.transformToWorld(new Point2D.Double(0, size.height));
		Point2D.Double origin = new Point2D.Double(lowerLeft.getX()-bounds.x,
				lowerLeft.getY()-bounds.y);
		Point2D displayOffset = coord.transformToDisplay(origin);
		System.out.println(displayOffset);
		coord.setOriginOffsetDisplay(displayOffset);
	}

}
