/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.6 $
 * $Date: 2004/06/24 21:57:02 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.ui;

import java.util.*;
import java.awt.event.*;
import java.awt.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;

/**
 * Default manager of the mouse events in the 
 * Graph Window
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid Moncada<p>
*/
public class DefaultGraphMouseManager 
	implements GraphMouseManager, CursorListener
{	
    public static final int ONCLICK_NONE = 0;
    public static final int ONCLICK_DELETE = 1;	//Not implemented yet
    public static final int ONCLICK_SELECT = 2;
    public static final int ONCLICK_SELECT_DESELECT = 3;

    private int mouseBehavior = ONCLICK_SELECT_DESELECT;

	private boolean clickAccepted = false;
	
	protected SelectableList selectableObjects;

	protected GraphWindow graph;
	
	private MouseSensitive clickedObject;
	private MouseSensitive mousedObject;
	
	public DefaultGraphMouseManager()
	{
	}

	public DefaultGraphMouseManager(GraphWindow g)
	{
		this();
		
		graph = g;
		
		//Temp
		if (graph != null){
			graph.addMouseMotionListener(this);
		}

	}
	
	/**
	 * Sets the list of Selectable objects
	 * that need to be selected with the mouse. 
	 * The objects in this list that are
	 * MouseControllable objects, need also
	 * to receive the mouse events (pressed,
	 * dragged, released)
	 */
	public void setSelectableList(SelectableList selectables)
	{
		this.selectableObjects = selectables;
	}

	/**
	 * Sets the behavior with the mouse
	 */
	public void setMouseBehavior(int behavior)
	{
		if (mouseBehavior == behavior) return;
				
		if (behavior == ONCLICK_NONE || behavior == ONCLICK_DELETE ||
				behavior == ONCLICK_SELECT || behavior == ONCLICK_SELECT_DESELECT){
			
			mouseBehavior = behavior;
			
			if (graph != null){
				graph.setCursor(null);
			}
		}
	}

	/**
	 * Returns the behavior with the mouse
	 */
	public int getMouseBehavior()
	{
		return mouseBehavior;
	}

	////////////////////////////////
	// Mouse events handling
	////////////////////////////////
	public void mousePressed(MouseEvent e) 
	{
		clickAccepted = false;
		clickedObject = null;
		
		if (graph != null && !graph.isEnabled()) return;
		
		Point location = e.getPoint();
		if (mouseBehavior == ONCLICK_SELECT || mouseBehavior == ONCLICK_SELECT_DESELECT){
			
			if (selectableObjects==null) return;
			
			clickedObject = getFirstObjectToSelect(location);
			
			doMousePressSelection(clickedObject, location);

			if (clickedObject != null){
				clickAccepted = true;
			}
			
		}
	}
	
	public void mouseDragged(MouseEvent e) 
	{
		if (graph != null && !graph.isEnabled()) return;
		if (graph != null && !graph.isEditable()) return;
		
		Point location = e.getPoint();
		if (mouseBehavior == ONCLICK_SELECT || mouseBehavior == ONCLICK_SELECT_DESELECT){

			//Only delegate the mouse dragged event if some graphable accepted the mouse pressed
			if (!clickAccepted) return;

			// Loop trough the selected MouseControllables to delegate the mouse dragged event
			doMouseDragged(location);
			
		}
	}

	public void mouseReleased(MouseEvent e) 
	{
		if (graph != null && !graph.isEnabled()) return;
		if (graph != null && !graph.isEditable()) return;

		Point location = e.getPoint();
		if (mouseBehavior == ONCLICK_SELECT || mouseBehavior == ONCLICK_SELECT_DESELECT){

			//Only delegate the mouse released event if some graphable accepted the mouse pressed
			if (!clickAccepted) return;

			// Loop trough the selected MouseControllables to delegate the mouse released event
			doMouseReleased(location);
			
			//Check if the current cursorable is still in the location of the mouse
			if (!mousedObject.isPointInProximity(location)){

				//Send a mouse exited event to the mousedObject
				if (mousedObject instanceof MouseMotionReceiver){
					MouseMotionReceiver mm = (MouseMotionReceiver)mousedObject;
					mm.mouseExited(location);
				}

				//Clear the current mousedObject
				setCurrentMousedObject(null);
			}

		}
	}

	/**
	 * Listener for the mouse move. 
	 * For now, there is a hack for scaling x and y axes
	*/
	public void mouseMoved(MouseEvent e)
	{
		boolean accepted = false;
		if (graph != null && !graph.isEnabled()) return;
		
		Point location = e.getPoint();

		if (selectableObjects==null) return;
			
		MouseSensitive ms = getFirstObjectToSelect(location);
		
		if (ms != mousedObject){
			
			if (mousedObject != null){

				//Delegate mouse exited on the old mousedObject
				if (mousedObject instanceof MouseMotionReceiver){
					MouseMotionReceiver mm = (MouseMotionReceiver)mousedObject;
					mm.mouseExited(location);
				}
				
			}
			
			if (ms != null){

				//Delegate mouse entered on the new MouseSensitive object ms
				if (ms instanceof MouseMotionReceiver){
					MouseMotionReceiver mm = (MouseMotionReceiver)ms;
					mm.mouseEntered(location);
				}
				
			}
			else{
				if (graph != null){
					graph.setCursor(null);
				}
			}

			setCurrentMousedObject(ms);
		}
		
		//Delegate mouse moved on the mousedObject
		if (mousedObject instanceof MouseMotionReceiver){
			MouseMotionReceiver mm = (MouseMotionReceiver)mousedObject;
			mm.mouseMoved(location);
		}
	}

	public void mouseExited(MouseEvent e) 
	{
	}
	public void mouseClicked(MouseEvent e)
	{
	}
	public void mouseEntered(MouseEvent e)
	{
	}
	////////////////////////////////////

	public MouseSensitive getFirstObjectToSelect(Point location)
	{
		return getFirstObjectToSelect(location, selectableObjects);
	}
	
	public MouseSensitive getFirstObjectToSelect(Point location, Vector list)
	{
		//Get the first object to select, going through the selected objects first
		return getFirstObjectToSelect(location, list, true);
	}

	public MouseSensitive getFirstObjectToSelect(Point location, Vector list, boolean checkSelectedFirst)
	{
		Object obj;
		Graphable g;
		MouseSensitive ms;
		Selectable s;
		int i;
		
		if (list == null || list.size() <= 0) return null;

		//In the vector aux, we will store the NON-SELECTED elements
		Vector nonSelectedObjs = new Vector();
			
		// Loop trough all the objects to see which one accepts the mouse click
		// Look for selection candidate in backwards order
		for (i = list.size()-1; i>=0; i--){
			obj = list.elementAt(i);
			if (obj instanceof GraphableList){

				/* FIXME Bug. You are making the assumption that every
                GraphableList is a Selectable. This will lead to the
				obvious ClassCast exception.*/
				ms = getFirstObjectToSelect(location, (SelectableList)obj);
				if (ms != null){
					return ms;
				}
			}
			else{
				if (obj instanceof Selectable){
					s = (Selectable)obj;
					
					//Don't pay attention to selectables with selection disabled
					if (!s.isSelectionEnabled()){
						continue;
					}
					
					if (checkSelectedFirst){
						//Don't pay attention to non-selected selectables yet
						if (!s.isSelected()){
							//Add the object to the non-selected vector
							nonSelectedObjs.add(0,s);
							continue;
						}
					}
				}
				if (obj instanceof Graphable){
					g = (Graphable)obj;
					//Don't pay attention to invisible graphables
					if (!g.isVisible()){
						continue;
					}
				}

				/* BUG. FIXME. I think the intention here is to loop through
				   the selected objects first. It only checks the non-selected
				   objects when no selected object claims the click.

				   The problem is when there is a visible graphable in the
				   list. This graphable will grab the click before the
				   non-selected selectables have a chance.
				   
				   This use of the continue statement is a bit awkward. We
				   should rewrite this. */
	
				if (obj instanceof MouseSensitive){
					ms = (MouseSensitive)obj;
					if (ms.isPointInProximity(location)){
						return ms;
					}
				}
			}
		} // for
		
		//Now search in the non-selected objects
		if (nonSelectedObjs.size() > 0){
			return getFirstObjectToSelect(location, nonSelectedObjs, false); 
		}
		
		return null;
	}

	public void doMousePressSelection(MouseSensitive ms, Point location)
	{
		Selectable s;
		if (ms != null){
			
			//Select the object
			if (ms instanceof Selectable){
				s = (Selectable)ms;
				selectSelectable(s);
			}

			//Mouse pressed events are not delegated if the graph is not editable
			if (graph != null && !graph.isEditable()) return;
			
			//Delegate the mouse pressed to all the mouse controllables
			doMousePressed(location);
		}
		else{  // (ms == null)
			
			if (mouseBehavior == ONCLICK_SELECT_DESELECT){
				// If no Selectable accepted the mouse click, deselect all
				deselectAllGraphables();
			}
		}
	}
	
	/**
	 * Delegate the mousePressed event to the Mouse Controllables
	 */
	public void doMousePressed(Point location)
	{
		MouseControllable mc;
		
		if (selectableObjects == null) return;

		//Delegate the mouse pressed event to the object that accepted the click
		if (clickedObject instanceof MouseControllable){
			mc = (MouseControllable)clickedObject;
			mc.mousePressed(location);
		}
		
		//Now loop trough all the selected objects and 
		//if they are MouseControllables, delegate the mouse pressed event
		Enumeration objs;
		objs = selectableObjects.getMousePropagationObjects().elements();
		while (objs.hasMoreElements()) {
			Object obj = objs.nextElement();
			//Be careful to not send again the mouse pressed event to the clicked object
			if (obj != clickedObject && obj instanceof MouseControllable){
				mc = (MouseControllable)obj;
				mc.mousePressed(location);
			}

		}
	}

	/**
	 * Delegate the mouseDragged event to the Mouse Controllables
	 */
	public void doMouseDragged(Point location)
	{
		MouseControllable mc;

		if (selectableObjects == null) return;

		//Delegate the mouse dragged event to the object that accepted the click
		if (clickedObject instanceof MouseControllable){
			mc = (MouseControllable)clickedObject;
			mc.mouseDragged(location);
		}

		//Now loop trough all the selected objects and 
		//if they are MouseControllables, delegate the mouse dragged event
		Enumeration objs;
		objs = selectableObjects.getMousePropagationObjects().elements();
		while (objs.hasMoreElements() ) {
			Object obj = objs.nextElement();
			//Be careful to not send again the mouse pressed event to the clicked object
			if (obj != clickedObject && obj instanceof MouseControllable){
				mc = (MouseControllable)obj;
				mc.mouseDragged(location);
			}
		}
	}
	
	/**
	 * Delegate the mouseReleased event to the Mouse Controllables
	 */
	public void doMouseReleased(Point location)
	{
		MouseControllable mc;

		if (selectableObjects == null) return;
		
		//Delegate the mouse released event to the object that accepted the click
		if (clickedObject instanceof MouseControllable){
			mc = (MouseControllable)clickedObject;
			mc.mouseReleased(location);
		}

		Enumeration objs;
		objs = selectableObjects.getMousePropagationObjects().elements();		
		while (objs.hasMoreElements() ) {
			Object obj = objs.nextElement();
			//Be careful to not send again the mouse pressed event to the clicked object
			if (obj != clickedObject && obj instanceof MouseControllable){
				mc = (MouseControllable)obj;
				mc.mouseReleased(location);
			}
		}
	}
	
	protected void setCurrentMousedObject(MouseSensitive s)
	{
		Cursorable cursorable;

		if (mousedObject instanceof Cursorable){
			cursorable = (Cursorable)mousedObject;
			//Stop listening to its cursor changed events
			cursorable.removeCursorListener(this);
		}

		mousedObject = s;
		
		if (mousedObject == null){
			changeCursor(null);
		}
		else if (mousedObject instanceof Cursorable){
			cursorable = (Cursorable)mousedObject;
			
			//Set the cursor of the current cursorable
			changeCursor(cursorable);
			
			//Listen to its cursor changed events
			cursorable.addCursorListener(this);
		}
		else{
			changeCursor(null);
		}
		//
	}

	//////////////////	
	//CursorListener interface
	//cursor change event handler
	public void cursorChanged(EventObject evt)
	{
		changeCursor((Cursorable)evt.getSource());
	}
	//////////////////	
	
	protected void changeCursor(Cursorable c)
	{
		if (graph == null) return;
		
		if (c==null){
			graph.setCursor(null);
		}
		else{
			graph.setCursor(c.getCursor());
		}
	}
	
	private void selectSelectable(Selectable s)
	{	
		if (s.isSelected()) return;
		//if (mouseBehavior == ONCLICK_SINGLESELECT){
		////Deselect all the other graphables
		//	deselectAllGraphables();
		//}
		s.select();
	}

	private void deselectAllGraphables()
	{	
		int selectedSize;
		if (selectableObjects == null) return;
		selectableObjects.deselectAll();
	}
}
