/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.7 $
 * $Date: 2004/06/07 17:55:42 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.ui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;

/**
 * The main graph panel that displays graphables
 * Mouse events are handled by a GraphMouseManager
 * Draw functionality is handled by a GraphDrawManager 
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid C Moncada<p>
 *
 */

public class GraphWindow extends JPanel 
{
    /** A vector of Graphable objects that are decorations in the background */
    private GraphableList decorations;

    /** 
	 * List of Selectable objects.
	 * The GraphWindow will select these objects with the mouse pressed event
	 * and it will pass each mouse event to the mouseControllable objects 
	 */
    private SelectableList mouseGraphables;

    /** 
	 * A vector of Graphable objects. Each element contains an object
	 * that is drawn on the screen, but does not respond to mouse events 
	 */
//    private GraphableList graphables;

    /** The GraphWindowListeners that receive events */
    private Vector listeners; 

	private boolean antialias = false;

    /** Sets if it should let the user edit (change) the graphables with the graph */
    protected boolean editable = true;
	
	/** Mouse events manager */
	private GraphMouseManager mouseManager;
	
	/** Draw managers */
	private GraphDrawManager decorationsDrawManager;
	private GraphDrawManager graphablesDrawManager;
	private GraphDrawManager mouseGraphablesDrawManager;
	
	/** Default Graph Area */
	private GraphArea defaultGraphArea;
	
	/** listener to the graph area changes */
	ChangeListener coordinateSystemListener;
	
	/** Creates and initializes a graphwindow */
    public GraphWindow() 
	{
		listeners = new Vector();
		
		///////////////////
		//Draw managers
		//There is no generic way to set this yet
		//setDrawManager(new DefaultGraphDrawManager());
		decorationsDrawManager = new DefaultGraphDrawManager();
//		graphablesDrawManager = new DefaultGraphDrawManager();
		mouseGraphablesDrawManager = new DefaultGraphDrawManager();
		///////////////////

		///////////////////
		//Mouse managers
		setMouseManager(new DefaultGraphMouseManager(this));
		///////////////////

		///////////////////
		// Graphables Lists
		setDecorationsList(new GraphableList());
//		setGraphablesList(new GraphableList());
		setSelectablesList(new SelectableList());
		///////////////////
		
		setBackground(Color.white);
		
		//Listener to the component resizing
		addComponentListener( new ComponentAdapter() 
		{
			public void componentResized(ComponentEvent e) 
			{
				notifyResize();

//				update();
			}
		});

		//Listener to the axis changes
		coordinateSystemListener = new ChangeListener()
		{
			 public void stateChanged(ChangeEvent e)
			 {
//				 update();
//				 repaint();
				 notifyChange();
			 }
		};

		///////////////////
		//Default graph area
		/** The default graph area uses an affine transformation matrix used to draw
		the contents of this window */
		setDefaultGraphArea(new GraphArea(new DefaultCoordinateSystem2D()));
		///////////////////
	}
    
    /** Adds a Graphable */
	public void add(Drawable g) 
	{
		if (g==null) return;
//		if (g instanceof Selectable){
//		//The Graphable is a Selectable, so it will be added 
//		//to the mouseGraphables list (SelectableList)
			if (mouseGraphables.contains(g)) return;
			mouseGraphables.add(g);
//		}
//		else{
//		//Graphables that are not Selectables will be added 
//		//to the graphables list (GraphableList)		
//			if (graphables.contains(g)) return;
//			graphables.add(g);	
//		}
		
		updateGraphArea(g);
	}

    /** Adds a Graphable */
	public void add(GraphableList l) 
	{
		if (l == null) return;
//		if (l instanceof SelectableList){
//		//The Graphable is a Selectable, so it will be added 
//		//to the mouseGraphables list (SelectableList)
			if (mouseGraphables.contains(l)) return;
			mouseGraphables.add(l);
//		}
//		else{
//		//Graphables that are not Selectables will be added 
//		//to the graphables list (GraphableList)		
//			if (graphables.contains(l)) return;
//			graphables.add(l);	
//		}
		l.addGraphableListListener(listListener);
		
		updateGraphArea(l);
	}

    /** Adds a Decoration */
	public void addDecoration(Drawable g) 
	{
		if (g==null) return;
		if (!decorations.contains(g)) {
			decorations.add(g);
		}

		updateGraphArea(g);
	}

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		Graphics2D g2 = (Graphics2D)g;

		if (antialias) {
			g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
								 RenderingHints.VALUE_ANTIALIAS_ON);
		}

		// Draw Decorations
		decorationsDrawManager.draw(g2);

		// Draw mouse Graphables
		mouseGraphablesDrawManager.draw(g2);
		
		// Draw Graphables
//		graphablesDrawManager.draw(g2);
	}

    /** Removes a Graph from the GraphWindow. */
    public void remove(Graphable graph) 
	{
		if (graph == null) return;
//		if (graphables.contains(graph)) {
//			graphables.remove(graph);
//		}
		if (mouseGraphables.contains(graph)){
			mouseGraphables.remove(graph);
		}
		repaint();
    }

    /** Removes a Graph from the GraphWindow. */
    public void remove(GraphableList l) 
	{
		if (l == null) return;
//		if (graphables.contains(l)) {
//			graphables.remove(l);
//		}
		if (mouseGraphables.contains(l)){
			mouseGraphables.remove(l);
		}
		l.removeGraphableListListener(listListener);
		repaint();
    }

	/** Removes all graphables in the window */
	public void removeAllGraphables()
	{
//		graphables.removeAllElements();
		mouseGraphables.removeAllElements();
		repaint();
	}

    /** Removes a Graph from the GraphWindow. */
    public void removeDecoration(Graphable graph) 
	{
		if (decorations.contains(graph)) {
			decorations.remove(graph);
		}
		repaint();
    }

	/** Removes all graphables in the window */
	public void removeAllDecorations()
	{
		decorations.removeAllElements();
		repaint();
	}

	/** Removes all graphables in the window */
	public void removeAll()
	{
		removeAllGraphables();
		removeAllDecorations();
	}

	public void addGraphWindowListener(GraphWindowListener listener) 
	{
		if (! listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	public void removeGraphWindowListener(GraphWindowListener listener) 
	{
		if (listeners.contains(listener)) {
			listeners.remove(listener);
		}
	}

	private void notifyChange() 
	{
		Enumeration lists = listeners.elements();
		
		EventObject e = new EventObject(this);
		
		while (lists.hasMoreElements()) {
			((GraphWindowListener)lists.nextElement()).windowChanged(e);
		}
	}

	private void notifyResize() 
	{
		Enumeration lists = listeners.elements();
		
		GraphWindowResizeEvent e = new GraphWindowResizeEvent(this,this.getSize());
		
		while (lists.hasMoreElements()) {
			((GraphWindowListener)lists.nextElement()).windowResized(e);
		}
	}

	public void setEditable(boolean b)
	{
		editable = b;
	}
	
	public boolean isEditable()
	{
		return editable;
	}
	
	public void setMouseBehavior(int mode)
	{
		if (mouseManager==null) return;
		mouseManager.setMouseBehavior(mode);
	}
	
	public int getMouseBehavior()
	{
		if (mouseManager==null) return -1;
		return mouseManager.getMouseBehavior();
	}

	public void setAntialias(boolean aa) 
	{
		antialias = aa;
	}
	
	public boolean getAntialias() 
	{
		return antialias;
	}

	private void update()
	{	
		updateGraphables(decorations);
//		updateGraphables(graphables);
		updateGraphables(mouseGraphables);
	}	

	private void updateGraphables(Vector gs)
	{	
		if (gs == null) return;
		
		Object obj;
		Graphable g;
		Enumeration graphs = gs.elements();
		while (graphs.hasMoreElements()) {
			obj = graphs.nextElement();
			if (obj instanceof Graphable){
				g = (Graphable)obj;
				//SUPER HACK
				//FIXME 
				if (g instanceof Grid){
					((Grid)g).setSize(getWidth(), getHeight());
				}
				g.update();
			}
		}
	}

	public void setMouseManager(GraphMouseManager mouseManager)
	{
		if (this.mouseManager!=null){
			this.removeMouseListener(this.mouseManager);
			this.removeMouseMotionListener(this.mouseManager);
		}		
		this.mouseManager = mouseManager;
		if (this.mouseManager!=null){
			this.addMouseListener(this.mouseManager);
			this.addMouseMotionListener(this.mouseManager);
			this.mouseManager.setSelectableList(mouseGraphables);
		}
	}
	
	SelectableListListener listListener = new SelectableListListener()
	{
		public void listGraphableChanged(EventObject e)
		{
			repaint();
		}
		
		public void listGraphableAdded(EventObject e)
		{
			Object obj = e.getSource();
			if (obj instanceof GraphableList){
				updateGraphArea((GraphableList)obj);
			}
			else if (obj instanceof Graphable){
				updateGraphArea((Graphable)obj);
			}
			repaint();
		}
		
		public void listGraphableRemoved(EventObject e)
		{
			repaint();
		}
		
		public void listGraphableSelected(EventObject e)
		{
			repaint();
		}
		
		public void listGraphableDeselected(EventObject e)
		{
			repaint();
		}
	};

	public void setSelectablesList(SelectableList list)
	{
		if (mouseGraphables!=null){
			mouseGraphables.removeGraphableListListener(listListener);
		}
		mouseGraphables = list;
		if (mouseGraphables!=null){
			mouseGraphables.addGraphableListListener(listListener);
		}

		if (mouseManager!=null){
			this.mouseManager.setSelectableList(mouseGraphables);
		}
		if (mouseGraphablesDrawManager!=null){
			mouseGraphablesDrawManager.setGraphableList(mouseGraphables);
		}
		updateGraphArea(mouseGraphables);
	}

//	private void setGraphablesList(GraphableList list)
//	{
//		if (graphables!=null){
//			graphables.removeGraphableListListener(listListener);
//		}
//		graphables = list;
//		if (graphables!=null){
//			graphables.addGraphableListListener(listListener);
//		}
//
//		if (graphablesDrawManager!=null){
//			graphablesDrawManager.setGraphableList(graphables);
//		}
//		updateGraphArea(graphables);
//	}

	private void setDecorationsList(GraphableList list)
	{
		if (decorations!=null){
			decorations.removeGraphableListListener(listListener);
		}
		decorations = list;
		if (decorations!=null){
			decorations.addGraphableListListener(listListener);
		}

		if (decorationsDrawManager!=null){
			decorationsDrawManager.setGraphableList(decorations);
		}
		updateGraphArea(decorations);
	}
	
	public void setMousePropagationMode(int m)
	{
		mouseGraphables.setMousePropagationMode(m);
	}
	
	public void setSelectionMode(int m)
	{
		mouseGraphables.setSelectionMode(m);
	}

	public void setDefaultGraphArea(GraphArea area)
	{
		if (defaultGraphArea != null){
			defaultGraphArea.removeChangeListener(coordinateSystemListener);
			this.removeGraphWindowListener(defaultGraphArea);
		}
		defaultGraphArea = area;
		if (defaultGraphArea != null){
			//listen to the graph area
			defaultGraphArea.addChangeListener(coordinateSystemListener);
			//Make the area listen to resize events
			addGraphWindowListener(defaultGraphArea);
		}

		updateGraphArea(decorations);
//		updateGraphArea(graphables);
		updateGraphArea(mouseGraphables);
		
	}
	
	public GraphArea getDefaultGraphArea()
	{
		return defaultGraphArea;
	}
	
	protected void updateGraphArea(Drawable g)
	{
		//If the graphable doesnt have a graph area, set is a the default one
		if (g.getGraphArea() == null){
			g.setGraphArea(defaultGraphArea);
		}
		else if (g.getGraphArea() != defaultGraphArea){
		//The graphable has already a graph area
			//listen to the graph area
			g.getGraphArea().addChangeListener(coordinateSystemListener);
			//Make the area listen to resize events
			addGraphWindowListener(g.getGraphArea());
		}
	}
	
	private void updateGraphArea(Vector gs)
	{	
		if (gs == null) return;
		
		Graphable g;
		Enumeration graphs = gs.elements();
		while (graphs.hasMoreElements()) {
			g = (Graphable)graphs.nextElement();
			updateGraphArea(g);
		}
	}

	public GraphMouseManager getMouseManager()
	{
		return mouseManager;
	}

}


