/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.9 $
 * $Date: 2004/07/27 23:03:35 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.ui;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.text.Format;
import java.text.DecimalFormat;

import javax.swing.event.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;

/**
 * This Class draws and maintains a standard grid for a graph. 
 * It keeps track of the number of units for each grid line.
 * 
 * @author Eric Brown-Munoz<p>
 * @author Ingrid Moncada<p>
*/

public class Grid 
	implements Graphable, ImageObserver, ChangeListener
{
    protected Image buffer;
    private int height;
    private int width;
//    long axisChecksum;

    private double xInterval;
    private double yInterval;

	// Automatically set of the interval in a Smart(tm) way. This
	//   is designed for use when the scale is not dragged.
	private boolean smartInterval = false; 

	private final int [] preferredIntervals = { 1 , 2, 5, 10, 20, 
												50, 100, 150, 200 };


	// Interval set initially (before scaling by dragging)
	private Point2D initialInterval;
	private Point2D initialScale;
	
    Color axisColor = Color.black;
    Color gridColor = Color.lightGray;
    Color labelColor = Color.black;

    Font labelFont = new Font("SansSerif",Font.PLAIN,10);
    FontMetrics labelMetrics;
	
	Format formatObj;

	private boolean autoAdjustLabels = true;
	private boolean showGridLabels = true;
	private boolean showGridLines = true;
	private boolean showAxisLabels = true;
	
	private String labelX;
	private String labelY;
	
	private boolean needUpdate = true;
	
	protected GraphArea graphArea;
	
	protected CoordinateSystemChangeEvent coordChange;
	
	protected Vector graphableListeners;
	
	private boolean useBufferedImage = false;
	
    /** Contstructs and initializes a Grid */
    public Grid() 
	{
		graphableListeners = new Vector();
		
		// Default behavior:
		
		// Default interval is 1
		setInterval(1.0,1.0);
		// Labels don't show decimals by default
		setLabelFormat(new DecimalFormat("#"));
	}

	/**
	 * Sets the interval in units to paint the grid
	*/
	public void setInterval(double intervalX, double intervalY)
	{
		if ( intervalX > 0.1 && intervalY > 0.1){
			setIntervalInternal(intervalX, intervalY);
			initialInterval = new Point2D.Double(xInterval, yInterval);
			needUpdate = true;
		}
	}
	
	/**
	 * Internal private method to set the interval
	*/
	private void setIntervalInternal(double intervalX, double intervalY)
	{
		if ( intervalX > 0.1 && intervalY > 0.1){
			xInterval = intervalX;
			yInterval = intervalY;
		}
	}

	/**
	 * Returns the interval in units in X
	*/
	public double getIntervalX()
	{
		return xInterval;
	}

	/**
	 * Returns the interval in units in X
	*/
	public double getIntervalY()
	{
		return yInterval;
	}

	/**
	 * Sets the object used to format the labels
	*/
	public void setLabelFormat(Format format)
	{
		formatObj = format;
	}
	
	public void setLabelSize(float labelSize)
	{
		labelFont = labelFont.deriveFont(labelSize);
	}
	
	public void setSize(int width, int height)
	{
		this.height = height;
		this.width = width;
	}
	
    /** Draws the grid on the screen. The result is stored in buffer for
	probable later use. */
	public void draw(Graphics2D g) 
	{

		if (useBufferedImage){
			if ((buffer == null) || needUpdate) {//|| (this.width != width) ||	(this.height != height) || (axis.getChecksum() != axisChecksum)) {
				// need to update buffer
				//axisChecksum = axis.getChecksum();
				update();
				needUpdate = false;
			}
			if (buffer != null) {
				g.drawImage(buffer,0,0,this);
			}
		}
		else{

			update();

			drawImage((Graphics2D)g.create());
		}
	}
	
	/**
	 * Updates the buffer image with the info from the axis, and leaves it ready to draw
	*/
    public void update() 
	{
		if (graphArea == null || width <= 0 || height <= 0) return;


		if (smartInterval) {

			
			Point2D scale = graphArea.getCoordinateSystem().getScale();

			double scalex = scale.getX();
			double scaley = scale.getY();

			double targetIntervalX =  Math.floor(30.0 / scalex); // FIXME hardcoded Value
			int i;
			// Change targetInterval to the next preffered Interval
			for (i= 0; 
				 (i < preferredIntervals.length) && 
					 (preferredIntervals[i] < targetIntervalX);
				 i++) {}
			if (i < preferredIntervals.length)  
				targetIntervalX = preferredIntervals[i];
				

			double targetIntervalY = Math.ceil(30.0 / scaley);

			// Change targetInterval to the next preffered Interval
			for (i= 0; 
				 (i < preferredIntervals.length) && 
					 (preferredIntervals[i] < targetIntervalY);
				 i++) {}
			if (i < preferredIntervals.length)  
				targetIntervalY = preferredIntervals[i];
			
			setInterval ( Math.ceil(targetIntervalX),
						  Math.ceil(targetIntervalY));
		}
			
		
		if (autoAdjustLabels){
			//Update the labels in the grid
			updateGridInterval();
		}

		
		if (useBufferedImage){
			if (buffer == null){
				buffer = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
				//BufferedImage.TYPE_3BYTE_BGR );
			}
			Graphics2D bg = (Graphics2D)buffer.getGraphics();
			
			Composite comp = bg.getComposite();
			bg.setComposite(AlphaComposite.Clear);
			bg.setColor(new Color(0,0,0,0));
			bg.fillRect(0,0,width,height);
			bg.setComposite(comp);
			
			drawImage(bg);
		}
	}
	
	public void drawImage(Graphics2D bg)
	{

		

		long ttime = System.currentTimeMillis();

		Line2D.Double scratchLine = new Line2D.Double();
		
		CoordinateSystem coord = graphArea.getCoordinateSystem();

		bg.setFont(labelFont);
	
		labelMetrics = bg.getFontMetrics(labelFont);

		// Axis Lines
		Point2D origin = new Point2D.Double(0.0,0.0);
		Point2D screenOrigin;
		
		screenOrigin = coord.transformToDisplay(origin);

		String label;
		int labelWidth;
		int labelHeight;

		/////////////////////////////
		// Turn off anti-aliasing to fix performance bug 
		Object savedAlias = 
			bg.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
		
		bg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							RenderingHints.VALUE_ANTIALIAS_OFF);
		
		/////////////////////////////
		//Labels in the grid
		if (showGridLabels || showGridLines){
			bg.setStroke(new BasicStroke(1.0f));
			
			double screenXInterval = xInterval * coord.getScale().getX();
			double screenYInterval = yInterval * coord.getScale().getY();
		
			// Interval number to start counting
			int xIntervalNumber = - (int) 
				Math.floor(screenOrigin.getX()/screenXInterval );
		
			double xi = screenOrigin.getX() + ((double)xIntervalNumber * 
							  screenXInterval);
					
			while (xi < width) {
				if (xIntervalNumber != 0) { // don't draw over axis
					
					if (showGridLines){
						bg.setColor(gridColor);
						scratchLine.setLine(xi, 0.0, xi, height);
						bg.draw(scratchLine);
					}
			
					if (showGridLabels){
						// Draw x axis labels
						Double val;
						if (formatObj!=null){
							label = formatObj.format(new Double(xIntervalNumber * xInterval));
						}
						else{
							label = Double.toString(xIntervalNumber * xInterval);
						}
						
						labelWidth = labelMetrics.stringWidth(label); // XXXXX
						labelHeight = labelMetrics.getHeight(); // XXXXX
						bg.setColor(labelColor); // YYYYYYYYYYY
						if ((2 * xi) > labelWidth) {
							bg.drawString(label, (int)(xi - labelWidth/2), 
										  (int)screenOrigin.getY() + 
										  labelHeight +2);
						}
					}
				}
		
				xIntervalNumber ++ ;
				xi = screenOrigin.getX() + ((double)xIntervalNumber * 
						   screenXInterval);
		
			}

			// System.println was here

			// Interval number to start counting
			int yIntervalNumber = - (int) 
				Math.floor(screenOrigin.getY()/screenYInterval );
		
			double yi = screenOrigin.getY() + ((double)yIntervalNumber * 
							  screenYInterval);


			while (yi < height) {
				if (yIntervalNumber != 0) { // Don't draw over axis

					if (showGridLines){
						bg.setColor(gridColor);
						
						scratchLine.setLine(0.0,yi, (double)width, yi);

						bg.draw(scratchLine);
					}

			
					if (showGridLabels){
						// Draw y axis labels
						if (formatObj!=null){
							label = formatObj.format(new Double(-(yIntervalNumber * yInterval)));
						}
						else{
							label = Double.toString( -(yIntervalNumber * yInterval));
						}
				
						labelWidth = labelMetrics.stringWidth(label); // XXXXX
						labelHeight = labelMetrics.getHeight(); // XXXXX
						bg.setColor(labelColor);
						bg.drawString(label, (int)(screenOrigin.getX() - labelWidth - 2), 
								  (int)(yi));
					}
				}
		
				yIntervalNumber ++ ;
				yi = screenOrigin.getY() + ((double)yIntervalNumber * 
							   screenYInterval);
				
			} // while (yi < height)

		}//(showGridLabels)
		///////////////

		
		///////////////
		//Axis labels
		if (showAxisLabels){
			if (labelX!=null && !labelX.equals("")){
				label = labelX;
				labelWidth = labelMetrics.stringWidth(label); // XXXXX
				labelHeight = labelMetrics.getHeight(); // XXXXX
				bg.setColor(labelColor);
				int posX = (int)(screenOrigin.getX() + ((width - screenOrigin.getX())/2) - (labelWidth/2));
				int posY = (int)(screenOrigin.getY() + 2*labelHeight);
				bg.drawString(label, posX, posY);
			} 
			if (labelY!=null && !labelY.equals("")){
				label = labelY;
				labelWidth = labelMetrics.stringWidth(label); // XXXXX
				labelHeight = labelMetrics.getHeight(); // XXXXX
				bg.setColor(labelColor);
				int posX = 0;

				//Draw label vertically
				int posY = (int)((screenOrigin.getY()/2) - (labelWidth/2));
				AffineTransform savedTransform = bg.getTransform();

				bg.translate(screenOrigin.getX(), screenOrigin.getY());
				bg.rotate(-Math.PI/2);
				bg.drawString(label, posY , - (labelHeight + 5) );

				bg.setTransform(savedTransform);

			} 


		}//(showAxisLabels)
		///////////////
		
		bg.setColor(axisColor);
		bg.setStroke(new BasicStroke(3.0f));

		if ((screenOrigin.getX() > 0.0) && (screenOrigin.getX() < width)) {
			scratchLine.setLine(screenOrigin.getX(),0.0,
								screenOrigin.getX(),height);

			bg.draw(scratchLine);
		}
	
		if ((screenOrigin.getY() > 0.0) && (screenOrigin.getY() < height)) {
			scratchLine.setLine(0.0, screenOrigin.getY(),
								width, screenOrigin.getY());

			bg.draw(scratchLine);
		}

		bg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							 savedAlias); // UUUUUUU
	    
    } // drawImage()
	
	/** Adds a GraphableListener to listen for events */
	public void addGraphableListener(GraphableListener gl)
	{
		graphableListeners.add(gl);
	}
	
	/** Removes a GraphableListener.*/
	public void removeGraphableListener(GraphableListener gl)
	{
		graphableListeners.remove(gl);
	}

    protected void notifyChange() 
	{
		EventObject e = new EventObject(this);
		Enumeration lists = graphableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			gl.graphableChanged(e);
		}
    }

	/** Removes the object from any enclosing view. */
	//NOT IMPLEMENTED
	public void remove(){}
	
	/** Sets whether it is visible in a graph */
	public void setVisible(boolean visible){}
	
	/** Returns true if it is visible in a graph */
	public boolean isVisible()
	{
		//Always visible
		return true;
	}

	/** Returns a copy of itself */
	public Graphable getCopy()
	{
		Grid g = new Grid();
		g.setInterval(xInterval, yInterval);
		return g;
	}

	/**
	 * Sets if the grid will auto update the interval and labels
	 * WHILE changing the axis scale
	 * By default it is turned ON
	*/
	public void setAutoAdjustLabels(boolean b)
	{
		autoAdjustLabels = b;
	}
	
	/**
	 * Returns if the grid will auto update the interval and labels
	 * WHILE changing the axis scale
	*/
	public boolean isAutoAdjustLabels()
	{
		return autoAdjustLabels;
	}

	public void setSmartInterval (boolean yes) {
		smartInterval = yes;
	}

	public boolean getSmartInterval () {
		return smartInterval;
	}

	/**
	 * Sets if the grid will show the interval number labels 
	 * By default it is turned ON
	*/
	public void setShowGridLabels(boolean b)
	{
		if (showGridLabels != b){
			needUpdate = true;
		}
		showGridLabels = b;
	}
	
	/**
	 * Returns if the grid will show the interval number labels 
	*/
	public boolean getShowGridLabels()
	{
		return showGridLabels;
	}

	/**
	 * Sets if the grid will show the grid lines 
	 * By default it is turned ON
	*/
	public void setShowGridLines(boolean b)
	{
		if (showGridLines != b){
			needUpdate = true;
		}
		showGridLines = b;
	}
	
	/**
	 * Returns if the grid will show the grid lines 
	*/
	public boolean getShowGridLines()
	{
		return showGridLines;
	}

	/**
	 * Sets if the grid will show the unit labels in the axis 
	 * By default it is turned ON
	*/
	public void setShowAxisLabels(boolean b)
	{
		if (showAxisLabels != b){
			needUpdate = true;
		}
		showAxisLabels = b;
	}
	
	/**
	 * Returns if the grid will show the unit labels in the axis 
	*/
	public boolean getShowAxisLabels()
	{
		return showAxisLabels;
	}

	/**
	 * Sets the label displayed under the x axis
	*/
	public void setLabelXAxis(String labelX)
	{
		this.labelX = labelX;
		needUpdate = true;
		notifyChange();
	}

	/**
	 * Sets the label displayed to the left of the y axis
	*/
	public void setLabelYAxis(String labelY)
	{
		this.labelY = labelY;
		needUpdate = true;
		notifyChange();
	}

	/**
	 * Returns the label displayed under the x axis
	*/
	public String getLabelXAxis()
	{
		if (this.labelX==null) return "";
		return this.labelX;
	}

	/**
	 * Returns the label displayed to the left of the y axis
	*/
	public String getLabelYAxis()
	{
		if (this.labelY==null) return "";
		return this.labelY;
	}

	/**
	 * This method is used to update the grid interval WHILE the axes are being dragged
	 * It tries to maintain the same grid interval in pixels in the screen,
	 * so it changes the interval accordingly, when the scale of the axis is changing.
	 * The grid labels will always be in multiples of 1, 2, 5 or 10
	*/
	private void updateGridInterval()
	{		
		if (!autoAdjustLabels) return;
		if (coordChange == null) return;
		if (!coordChange.isChanging()) return;		
		if (!showGridLabels) return;

		CoordinateSystem coord;
		coord = graphArea.getCoordinateSystem();

		if (coordChange.isScaleChange() && coordChange.isFirstAxisChange()){
			double interval;
			double intervalPx;
			double newInterval;
			double newIntervalPx;
	
			if (initialInterval == null) return;
	
			interval = initialInterval.getX();
			intervalPx = interval * initialScale.getX();
	
			newInterval = intervalPx / coord.getScale().getX();
			
			int expInterval = log10(newInterval);
			newInterval = newInterval / Math.pow(10, expInterval);
			
			if (newInterval<=1) newInterval =1;
			else if (newInterval<=2) newInterval =2;
			else if (newInterval<=5) newInterval =5;
			else newInterval = 10;
			
			newInterval = newInterval * Math.pow(10, expInterval);
	
			if (newInterval != xInterval){
				setIntervalInternal(newInterval, yInterval);
			}
		}
		
		if (coordChange.isScaleChange() && coordChange.isSecondAxisChange()){
			double interval;
			double intervalPx;
			double newInterval;
			double newIntervalPx;

			if (initialInterval == null) return;

			interval = initialInterval.getY();
			intervalPx = interval * initialScale.getY();

			newInterval = intervalPx / coord.getScale().getY();

			int expInterval = log10(newInterval);
			newInterval = newInterval / Math.pow(10, expInterval);
			
			if (newInterval<=1) newInterval =1;
			else if (newInterval<=2) newInterval =2;
			else if (newInterval<=5) newInterval =5;
			else newInterval = 10;
			
			newInterval = newInterval * Math.pow(10, expInterval);

			if (newInterval != yInterval){
				setIntervalInternal(xInterval, newInterval);
			}
		}
		
		coordChange = null;

	}
    
    /** This is required for ImageObserver Inteface. Here we shouldn't need to
	worry about threads, so just return false. No Problem. */
    public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) 
	{
		return false; // The image is already "loaded" (i.e. OK)
    }

	//Temporary. It will be more efficient
	private static int log10(double num)
	{
		double n;
		int i=0;
		
		if (num == 0.0) return 0;
		
		n = Math.abs(num);
		
		if (n >= 1.0){
			while(n >= 10.0){
				i++;
				n = n/10.0;
			}
		}
		/*
		In this case no need to do negative exponents (should be 0)
		else if (n < 1.0){
			while(n < 1.0){
				i--;
				n = n*10.0;
			}
		}
		*/
		return i;
	}	

	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		if (graphArea != null){
			graphArea.removeChangeListener(this);
		}
		this.graphArea = area;
		if (graphArea != null){
			graphArea.addChangeListener(this);
		}
		
		initialScale = graphArea.getCoordinateSystem().getScale();
	}

	/** 
	 * returns the graph area where the graphable will be drawn 
	 */
	public GraphArea getGraphArea()
	{
		return this.graphArea;
	}

	/**
	 * Handler of the changes in the graph area
	 */
	public void stateChanged(ChangeEvent e)
	{
		if (e instanceof CoordinateSystemChangeEvent){
			
			//If graphArea changed size...
			if (e instanceof GraphAreaChangeEvent){
				if (((GraphAreaChangeEvent)e).isSizeChange()){
					Dimension d = graphArea.getSize();
					setSize((int)d.getWidth(), (int)d.getHeight());
				}
			}
			
			coordChange = (CoordinateSystemChangeEvent)e;
			 
			needUpdate = true;
			notifyChange();
		}
	}
	
	public void setUseBufferedImage(boolean b)
	{
		useBufferedImage = b;
	}
}




