/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.5 $
 * $Date: 2004/05/28 22:12:19 $
 * $Author: scytacki $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.ui;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.text.Format;
import java.text.DecimalFormat;

import javax.swing.event.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;

/**
 * @author Ingrid Moncada<p>
*/

public class SingleAxisGrid extends DefaultGraphable
	implements ChangeListener
{
	private double interval;		// Number of UNITS (world coordinates) for each grid line
	private int majorInterval = 0;	// Number of minor intervals in each major interval (in relationship with the interval)
	private int coordinateNumber;	//This number indicates which axis ths grid belongs to
									//1 is X axis (first coordinate), 2 is Y axis (second coordinate)

	// Interval set initially (before scaling by dragging)
	private double initialInterval;
	private double initialScale;

	//Default colors for lines and labels
	private Color defaultAxisLinesColor = Color.black;
	private Color defaultGridLinesColor = Color.lightGray;
	private Color defaultLabelsColor = Color.black;

	////////////////
	//Properties of the axes lines and labels
	private boolean showAxisLines = true;
	private boolean showAxisLabels = true;
	private Color axisLinesColor;
	private Color axisLabelColor;
	private Font axisLabelFont = new Font("SansSerif",Font.PLAIN,10);
	private Stroke axisLinesStroke = new BasicStroke(3.0f);
	//Properties of the grid lines and number labels
	private boolean showGridLines = true;
	private boolean showGridLabels = true;
	private boolean showGridLabelsMajorLinesOnly = true;
	private Color gridMajorLinesColor;
	private Color gridMinorLinesColor;
	private Color gridLabelColor;
	private Font gridLabelFont = new Font("SansSerif",Font.PLAIN,10);
	private Stroke gridMinorLinesStroke = new BasicStroke(1.0f);
	private Stroke gridMajorLinesStroke = new BasicStroke(2.0f);
	
	private boolean showLabelsOnMajorOnly = false;
	////////////////
	
	private boolean autoAdjustLabels = true;

	private boolean drawGridOnAxis = true;
	
	////////////////

	private Format formatObj;	
	private String axisLabel;
	private boolean needUpdate = true;
	protected CoordinateSystemChangeEvent coordChange;

	protected Point2D startPosition; 	// World coordinates
	protected Point2D endPosition; 		// World coordinates
	
	/** 
	 * Constructs and initializes a Single Axis Grid 
	 * for the given coordinate number,
	 * with the default behavior:
	 * Interval of 1 unit, labels with no decimals,
	 * No minor lines
	 * @throws IllegalArgumentException - if coordinateNumber
	 *			is different than 1 or 2
	 */
	public SingleAxisGrid(int coordinateNumber) 
		throws IllegalArgumentException
	{
		//Only supports 1 and 2 for now
		if (!(coordinateNumber == 1 || coordinateNumber == 2)){
			throw new IllegalArgumentException("Grid only supports coordinate number of 1 or 2");
		}
		
		graphableListeners = new Vector();
		
		this.coordinateNumber = coordinateNumber;
		
		// Default behavior:
		
		// Default interval is 1
		setInterval(1.0);
		// Labels don't show decimals by default
		setLabelFormat(new DecimalFormat("#"));
	}
	
	////////////////////////////
	// Color business :p
	
	/**
	 * Sets the color of the axis lines
	 */
	public void setColorAxisLines(Color color)
	{
		axisLinesColor = color;
	}
	
	/**
	 * Returns the color of the axis lines
	 */
	public Color getColorAxisLines()
	{
		if (axisLinesColor == null){
			return defaultAxisLinesColor;
		}
		return axisLinesColor;
	}
	
	/**
	 * Sets the color of the axis label
	 */
	public void setColorAxisLabel(Color color)
	{
		axisLabelColor = color;
	}
	
	/**
	 * Returns the color of the axis label
	 */
	public Color getColorGridLabels()
	{
		if (gridLabelColor == null){
			return defaultLabelsColor;
		}
		return gridLabelColor;
	}
	
	/**
	 * Sets the color of the labels drawn on each grid line
	 * @param color
	 */
	public void setColorGridLabels(Color color)
	{
		gridLabelColor = color;
	}
	
	/**
	 * Returns the color of the label drawn on the axis
	 * @return the color of the label drawn on the axis
	 */
	public Color getColorAxisLabel()
	{
		if (axisLabelColor == null){
			return defaultLabelsColor;
		}
		return axisLabelColor;
	}

	/**
	 * Sets the color of the grid lines (minor and major)
	 * @param color
	 */
	public void setColorGridLines(Color color)
	{
		setColorMajorGridLines(color);
		setColorMinorGridLines(color);
	}
	
	/**
	 * Sets the color of the major grid lines
	 * @param color
	 */
	public void setColorMajorGridLines(Color color)
	{
		gridMajorLinesColor = color;
	}

	/**
	 * Returns the color of the major grid lines
	 */
	public Color getColorMajorGridLines()
	{
		if (gridMajorLinesColor == null){
			return defaultGridLinesColor;
		}
		return gridMajorLinesColor;
	}
	
	/**
	 * Sets the color of the minor grid lines
	 * @param color
	 */
	public void setColorMinorGridLines(Color color)
	{
		gridMinorLinesColor = color;
	}

	/**
	 * Returns the color of the minor grid lines
	 */
	public Color getColorMinorGridLines()
	{
		if (gridMinorLinesColor == null){
			return defaultGridLinesColor;
		}
		return gridMinorLinesColor;
	}
	////////////////////////////
	
	/**
	 * Sets the number of minor lines between each major
	*/
	public void setMajorInterval(int interval)
	{
		if ( interval > 0){
			majorInterval = interval;
			needUpdate = true;
		}
	}

	/**
	 * Sets the interval in units to paint the grid
	*/
	public void setInterval(double interval)
	{
		if ( interval > 0.1){
			setIntervalInternal(interval);
			initialInterval = interval;
			needUpdate = true;
		}
	}
	
	/**
	 * Internal private method to set the interval
	*/
	private void setIntervalInternal(double interval)
	{
		if ( interval > 0.1){
			this.interval = interval;
		}
	}
	
	/**
	 * Returns the interval in units in X
	*/
	public double getInterval()
	{
		return interval;
	}
	
	/**
	 * Sets the object used to format the labels
	*/
	public void setLabelFormat(Format format)
	{
		formatObj = format;
	}
	
	/**
	 * Sets the size of the label that is drawn on the axis
	 * @param labelSize	size of the font (in pixels)
	 */
	public void setAxisLabelSize(float labelSize)
	{
		axisLabelFont = axisLabelFont.deriveFont(labelSize);
	}
	
	/**
	 * Sets the size of the number labels that are drawn 
	 * for each grid line
	 * @param labelSize	size of the font (in pixels)
	 */
	public void setGridLabelSize(float labelSize)
	{
		gridLabelFont = gridLabelFont.deriveFont(labelSize);
	}
	
	/**
	 * Sets if the grid labels (numbers) will be drawn in ONLY
	 * the major grid lines or not. The default values is false,
	 * so the grid labels will be drawn for every minor grid line. 
	 * @param b
	 */
	public void setShowLabelsOnMajorOnly(boolean b)
	{
		showLabelsOnMajorOnly = b;
	}

	public void update() 
	{
	}
	
	/**
	 * Draws a line between two points given in world coordinates
	 */
	private void drawLine(Graphics2D g, Point2D worldP1, Point2D worldP2)
	{
		if (graphArea == null) return;

		CoordinateSystem coord = graphArea.getCoordinateSystem();		

		Point2D displayP1;
		Point2D displayP2;
		
		displayP1 = coord.transformToDisplay(worldP1);
		displayP2 = coord.transformToDisplay(worldP2);
		
		Line2D l = new Line2D.Double(displayP1, displayP2);
		
		g.draw(l);
	}
	
	/**
	 * Draws a label on a given point in world coordinates
	 */
	private void drawLabel(Graphics2D g, String strLabel, FontMetrics labelMetrics, 
			Point2D p, boolean fromgrid)
	{
		if (graphArea == null) return;

		double posX = 0.0;
		double posY = 0.0;
		boolean doDraw = false;
		
		CoordinateSystem coord = graphArea.getCoordinateSystem();		

		int labelWidth = labelMetrics.stringWidth(strLabel);
		int labelHeight = labelMetrics.getHeight();
		
		Point2D display;
		
		display = coord.transformToDisplay(p);
		
		if (fromgrid){
			if (coordinateNumber == 1){
				if ((2 * display.getX()) > labelWidth) {
					posX = display.getX() - labelWidth/2; 
					posY = display.getY() + labelHeight + 2;
					doDraw = true;
				}
			}
			else if (coordinateNumber == 2){
				posX = display.getX() - labelWidth - 2;
				posY = display.getY() + (labelHeight / 4);
				doDraw = true;
			}
		}
		else{
			if (coordinateNumber == 1){
				//In the middle of the positive axis
				//posX = display.getX() + ((graphArea.getSize().getWidth() - display.getX())/2) - (labelWidth/2);
				//In the middle of the screen
				posX = (graphArea.getSize().getWidth() / 2) - (labelWidth / 2);
				posY = display.getY() + 2*labelHeight;
				doDraw = true;
			}
			else if (coordinateNumber == 2){
				//
				// Draw label vertically
				//
				//In the middle of the positive axis
				//posX = (display.getY() / 2) - (labelWidth / 2);
				//In the middle of the screen
				posX = display.getY() - (graphArea.getSize().getHeight() / 2) - (labelWidth / 2);
				posY = - (labelHeight + 5);
				
				AffineTransform savedTransform = g.getTransform();
				g.translate(display.getX(), display.getY());
				g.rotate(-Math.PI/2);
				g.drawString(strLabel, (int)posX, (int)posY);

				g.setTransform(savedTransform);
			}
		}
		
		if (doDraw){
			g.drawString(strLabel, (int)posX, (int)posY);
		}
	}
	
	/** 
	 * Draws the grid on the screen. 
	 */
	public void draw(Graphics2D g) 
	{
		if (graphArea == null) return;

		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		Font oldFont = g.getFont();
		
		startPosition = graphArea.getLowerLeftCornerWorld();
		endPosition = graphArea.getUpperRightCornerWorld();
		
		//Grid lines and labels
		drawGridLinesAndLabels(g);
		
		//Axis lines
		drawAxisLinesAndLabels(g);
	    
		g.setColor(oldColor);
		g.setStroke(oldStroke);
		g.setFont(oldFont);
    }
	
	/** 
	 * Draws the grid lines and number labels in the grid
	 */
	public void drawGridLinesAndLabels(Graphics2D g)
	{
		String label;
		int labelWidth;
		int labelHeight;
		Point2D p1 = new Point2D.Double();
		Point2D p2 = new Point2D.Double();
		Point2D p3 = new Point2D.Double();

		if (showGridLabels || showGridLines){
			
			FontMetrics labelMetrics = g.getFontMetrics(gridLabelFont);
			g.setFont(gridLabelFont);
			
			if (coordinateNumber == 1){
			//Vertical lines
				//Start position
				double startX = absFloor(startPosition.getX() / interval) * interval;
				double startMajorX = absFloor(startX / majorInterval) * majorInterval;
				boolean isMajor;
				
				p1.setLocation(startX, startPosition.getY());
				p2.setLocation(startX, endPosition.getY());
				
				while (p1.getX() <= endPosition.getX()){
					
					isMajor = false;
					if (majorInterval != 0 && startMajorX == p1.getX()){
						isMajor = true;
						startMajorX = startMajorX + majorInterval;
					}
					
					//////////////
					// Grid lines
					if (showGridLines){
						if (isMajor){
							g.setStroke(gridMajorLinesStroke);
							g.setColor(this.getColorMajorGridLines());
						}
						else{
							g.setStroke(gridMinorLinesStroke);
							g.setColor(this.getColorMinorGridLines());
						}
						
						//Don't draw the axis as a grid line or label if you dont have to 
						if (!drawGridOnAxis && p1.getX() == 0){	
						}
						else{
							drawLine(g, p1, p2);
						}
					}
					//////////////
					
					//////////////
					// Grid labels
					if (showGridLabels){
						if (showLabelsOnMajorOnly && !isMajor){
						}
						else{
							if (formatObj != null){
								label = formatObj.format(new Double(p1.getX()));
							}
							else{
								label = Double.toString(p1.getX());
							}
							
							p3.setLocation(p1.getX(), 0.0);
							
							g.setColor(this.getColorGridLabels());
	
							//Don't draw the axis as a grid line or label if you dont have to 
							if (!drawGridOnAxis && p1.getX() == 0){	
							}
							else{
								drawLabel(g, label, labelMetrics, p3, true);
							}
						}
					}
					//////////////

					p1.setLocation(p1.getX() + interval, p1.getY());
					p2.setLocation(p2.getX() + interval, p2.getY());
				}
			}
			else if (coordinateNumber == 2){
			//Horizontal lines
				//Start position
				double startY = absFloor(startPosition.getY() / interval) * interval;
				double startMajorY = absFloor(startY / majorInterval) * majorInterval;
				boolean isMajor;
				
				p1.setLocation(startPosition.getX(), startY);
				p2.setLocation(endPosition.getX(), startY);
				
				while (p1.getY() <= endPosition.getY()){
					
					isMajor = false;
					if (majorInterval != 0 && startMajorY == p1.getY()){
						isMajor = true;
						startMajorY = startMajorY + majorInterval;
					}

					//////////////
					// Grid lines
					if (showGridLines){
						if (isMajor){
							g.setStroke(gridMajorLinesStroke);
							g.setColor(this.getColorMajorGridLines());
						}
						else{
							g.setStroke(gridMinorLinesStroke);
							g.setColor(this.getColorMinorGridLines());
						}

						//Don't draw the axis as a grid line or label if you dont have to 
						if (!drawGridOnAxis && p1.getY() == 0){	
						}
						else{
							drawLine(g, p1, p2);
						}
					}
					//////////////
					
					//////////////
					// Grid labels
					if (showGridLabels){
						if (showLabelsOnMajorOnly && !isMajor){
						}
						else{
							if (formatObj != null){
								label = formatObj.format(new Double(p1.getY()));
							}
							else{
								label = Double.toString(p1.getY());
							}
							
							p3.setLocation(0.0, p1.getY());
							
							g.setColor(this.getColorGridLabels());
	
							//Don't draw the axis as a grid line or label if you dont have to 
							if (!drawGridOnAxis && p1.getY() == 0){	
							}
							else{
								drawLabel(g, label, labelMetrics, p3, true);
							}
						}
					}
					//////////////
					
					p1.setLocation(p1.getX(), p1.getY() + interval);
					p2.setLocation(p2.getX(), p2.getY() + interval);
				}
			}
			
		}
	}
	
	/**
	 * Draws the axis lines and the labels on the axis
	 */
	public void drawAxisLinesAndLabels(Graphics2D g)
	{
		Point2D p1 = new Point2D.Double();
		Point2D p2 = new Point2D.Double();
		Point2D p3 = new Point2D.Double();

		if (showAxisLines){
			g.setColor(this.getColorAxisLines());
			g.setStroke(axisLinesStroke);

			if (coordinateNumber == 1){
			//X Axis
				if (0.0 >= startPosition.getY() &&
					0.0 <= endPosition.getY()){

					p1.setLocation(startPosition.getX(), 0.0);
					p2.setLocation(endPosition.getX(), 0.0);
					
					drawLine(g, p1, p2);
				}
			}
			else if (coordinateNumber == 2){
			//Y Axis
				if (0.0 >= startPosition.getX() &&
					0.0 <= endPosition.getX()){

					p1.setLocation(0.0, startPosition.getY());
					p2.setLocation(0.0, endPosition.getY());
					
					drawLine(g, p1, p2);
				}
			}
		}
		if (showAxisLabels){
			g.setColor(this.getColorAxisLabel());
			g.setFont(axisLabelFont);
			FontMetrics labelMetrics = g.getFontMetrics(axisLabelFont);

			if (coordinateNumber == 1){
			//X Axis
				if (0.0 >= startPosition.getY() &&
					0.0 <= endPosition.getY()){

					p3.setLocation(0.0, 0.0);
					if (axisLabel != null && !axisLabel.equals("")){
						drawLabel(g, axisLabel, labelMetrics, p3, false);
					}
				}
			}
			else if (coordinateNumber == 2){
			//X Axis
				if (0.0 >= startPosition.getX() &&
					0.0 <= endPosition.getX()){

					p3.setLocation(0.0, 0.0);
					if (axisLabel != null && !axisLabel.equals("")){
						drawLabel(g, axisLabel, labelMetrics, p3, false);
					}
				}
			}
		}
	}
	
	private static double absFloor(double v)
	{
		double r;
		r = Math.floor(Math.abs(v));
		if (v < 0){
			r = -r;
		}
		return r;
	}
	
	/** Returns a copy of itself */
	public Graphable getCopy()
	{
		SingleAxisGrid g = new SingleAxisGrid(coordinateNumber);
		g.setInterval(interval);
		return g;
	}

	/**
	 * Sets if the grid will auto update the interval and labels
	 * WHILE changing the axis scale
	 * By default it is turned ON
	*/
	public void setAutoAdjustLabels(boolean b)
	{
		autoAdjustLabels = b;
	}
	
	/**
	 * Returns if the grid will auto update the interval and labels
	 * WHILE changing the axis scale
	*/
	public boolean isAutoAdjustLabels()
	{
		return autoAdjustLabels;
	}

	/**
	 * Sets if the grid will show the interval number labels 
	 * By default it is turned ON
	*/
	public void setShowGridLabels(boolean b)
	{
		if (showGridLabels != b){
			needUpdate = true;
		}
		showGridLabels = b;
	}
	
	/**
	 * Returns if the grid will show the interval number labels 
	*/
	public boolean getShowGridLabels()
	{
		return showGridLabels;
	}

	/**
	 * Sets if the grid will show the grid lines 
	 * By default it is turned ON
	*/
	public void setShowGridLines(boolean b)
	{
		if (showGridLines != b){
			needUpdate = true;
		}
		showGridLines = b;
	}
	
	/**
	 * Returns if the grid will show the grid lines 
	*/
	public boolean getShowGridLines()
	{
		return showGridLines;
	}

	/**
	 * Sets if the grid will show the axis (lines)
	 * By default it is turned ON
	*/
	public void setShowAxisLines(boolean b)
	{
		showAxisLines = b;
	}
	
	/**
	 * Returns if the grid will show the axis (lines)
	*/
	public boolean getShowAxisLines()
	{
		return showAxisLines;
	}

	/**
	 * Sets if the grid will show the unit labels in the axis 
	 * By default it is turned ON
	*/
	public void setShowAxisLabels(boolean b)
	{
		if (showAxisLabels != b){
			needUpdate = true;
		}
		showAxisLabels = b;
	}
	
	/**
	 * Returns if the grid will show the unit labels in the axis 
	*/
	public boolean getShowAxisLabels()
	{
		return showAxisLabels;
	}

	/**
	 * Sets the label displayed on the axis
	*/
	public void setAxisLabel(String label)
	{
		this.axisLabel = label;
		needUpdate = true;
	}

	/**
	 * Returns the label displayed on the axis
	*/
	public String getAxisLabel()
	{
		if (this.axisLabel==null) return "";
		return this.axisLabel;
	}
    
	public void setDrawGridOnAxis(boolean b)
	{
		drawGridOnAxis = b;
	}
	
	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		super.setGraphArea(area);
		
		if (coordinateNumber == 1){
			initialScale = graphArea.getCoordinateSystem().getScale().getX();
		}
		else if (coordinateNumber == 2){
			initialScale = graphArea.getCoordinateSystem().getScale().getY();
		}
	}

	/**
	 * Handler of the changes in the graph area
	 */
	 public void stateChanged(ChangeEvent e)
	 {
		 if (e instanceof CoordinateSystemChangeEvent){
			 coordChange = (CoordinateSystemChangeEvent)e;
			 
			 //Store the scale if the coordinate system is not changing anymore
			 if (!coordChange.isChanging()){
				if (coordinateNumber == 1){
					initialScale = graphArea.getCoordinateSystem().getScale().getX();
				}
				else if (coordinateNumber == 2){
					initialScale = graphArea.getCoordinateSystem().getScale().getY();
				}
			 }
			 needUpdate = true;
			 notifyChange();
		 }
	 }
}




