/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.1 $
 * $Date: 2004/05/28 20:32:54 $
 * $Author: scytacki $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graph.ui;

import java.awt.*;
import java.awt.geom.*;
import java.awt.font.*;


/**
 * Considerations: It might be better if this had a setText
 * and getText method.  Then the width and height could be 
 * returned.  Also this creates a new TextLayout each time
 * the draw method is called so it might be better to keep
 * these objects around.
 * 
 * Also this class isn't directly depended on the Graph 
 * it could be put in the general concord ui package.  
 * 
 * @author Scott Cytacki
 *
 */
public class TextLabel
{
    // Constants these are mutiples of PI/2
    public final static float RIGHT = 0;
    public final static float DOWN = 1;	
    public final static float LEFT = 2;
    public final static float UP = 3;
 
    // these are for the anchorPosition
    public final static Point2D RIGHT_EDGE = new Point2D.Double(1,0.5);
    public final static Point2D LEFT_EDGE = new Point2D.Double(0,0.5);
    public final static Point2D TOP_EDGE = new Point2D.Double(0.5,1);
    public final static Point2D BOTTOM_EDGE = new Point2D.Double(0.5,0);

	/**
	 * Font to use for text
	 */
    private Font font     = null;

	/**
	 * Text color
	 */
    private Color color = Color.black;

    /**
     * the text direction
     */
    private float direction = 0;
    
    /**
     * anchor position this is a percentage offset from the lower 
     * left hand corner.  So 0.5 is half way across a side.
     */
    private Point2D anchorPosition = new Point2D.Double(0,0);
    
    public void setColor(Color c)
    {
    	color = c;
    }

    public Color getColor()
    {
    	return color;		
    }
    
    public void setFont(Font f)
    {
    	font = f;
    }
    
    public Font getFont()
    {
    	return font;
    }
    
    /**
     * This is the direction the text should be rotated.
     * there are predefined rotations for drawing text:
     * UP, LEFT, RIGHT, and DOWN.  (Note: left text will be upside
     * down)
     * @param d
     */
    public void setDirection(float d)
    {    	
    	direction = d;
    }
    
    public float getDirection()
    {
    	return direction;
    }
    
    /**
     * This is the position before rotation where
     * the text should be anchored.  The x and y values
     * are from 0 to 1 and they represent fractions of the
     * height and width from the lower left corner of the 
     * text.  The actual anchor point is passed in when the 
     * text is drawn.  This just indicates the position of this
     * point relative to the text itself.
     * 
     * There are predefined posistions of:
     * LEFT_EDGE - center of the left edge of the text 
     * RIGHT_EDGE - center of the right edge of the text
     * TOP_EDGE - center of the top edge of the text
     * BOTTOM_EDGE - center of the bottom edge of the text
     *  
     * @param pos
     */
    public void setAnchorPosition(Point2D pos)
    {
    	anchorPosition = pos;
    }
    
    public Point2D getAnchorPosition()
    {
    	return anchorPosition;
    }
    
    /**
     * 
     * @param g
     * @param anchor point in display coordinates
     * @param text
     */
    public void draw(Graphics2D g, Point2D anchor, String text)
    {
    	g.setColor(color);

    	FontRenderContext frc = g.getFontRenderContext();
    	TextLayout layout = new TextLayout(text, font, frc);
    	
    	Rectangle2D bounds = layout.getBounds();
    	
    	float posX = (float)(-anchorPosition.getX()*bounds.getWidth());
    	float posY = (float)(anchorPosition.getY()*bounds.getHeight());
    	    	
    	if (direction != 0){
			AffineTransform savedTransform = g.getTransform();
			g.translate(anchor.getX(), anchor.getY());
			g.rotate(direction*Math.PI/2);
			layout.draw(g, posX, posY);

			g.setTransform(savedTransform);    		
    	}
    	else {
    		layout.draw(g, posX, posY);
    	}

		return;
    }


}

