/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.8 $
 * $Date: 2004/08/02 21:21:06 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.engine;

import java.util.*;

import java.awt.*;
import java.awt.geom.*;

import javax.swing.event.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;

import org.concord.graphfunction.event.*;

/**
 * RTT specific version of Default Graphfunction. 
 * DefaultGraphfunction is  having problems with 
 * new graph package and RTT applets. RTTGraphfunction 
 * will support new Interactives while DefaultGraphfunction
 * will keep old stuff compiling.
 *
 * Eventually these should be combined again.
 *
 * Date created: 05/06/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid C Moncada<p>
 *
 */

public abstract class DefaultGraphFunction
	implements GraphFunction, ChangeListener
{
	//Expression form constants
	public static final int FORM_NONE = 0;

    private Color color;						// Color of the function
	private int expressionForm = FORM_NONE;		// Expression form

	//////////////////////////////////////////////////////
	//Properties for Graphable and Selectable interfaces 
	protected Vector graphableListeners;					//graphable listeners	
	protected Vector selectableListeners;					//selectable listeners	

	private boolean visible = true;		//is the graphable currently visible
	private boolean selected = false;  	//is the graphable currently selected
	private boolean selectable = true; 	//is the graphable selectable
	//////////////////////////////////////////////////////

	private String functionName = "";	//Name of the function (f,g, etc)
	private int functionOrdinal = -1;	//Ordinal of the function (1,2,3...) function will be f1, f2, f3
	
    private Stroke lineStroke;
    private Stroke selectedLineStroke;    

    protected DragMode dragMode;

    protected FunctionPainter painter = null;
	
	protected boolean showGhost = false;
	protected GraphFunctionGhost ghost;
	
	private boolean mouseClicked = false;
	
	private Cursor currentCursor;
	private Vector cursorListeners;
	
	private GraphArea graphArea;
	
	//////////////////////////////////////////////////////
	private GeneralPath path;		//For default drawing
	//////////////////////////////////////////////////////

	/**
	 * Default constructor. 
	 * Initializes stroke values and dragMode
	 */
	public DefaultGraphFunction()
	{
		//////////////////////////////////////////////////////
		//Properties for Graphable interface 
		graphableListeners = new Vector();
		selectableListeners = new Vector();
		
		lineStroke = new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    	selectedLineStroke = new BasicStroke(3.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
		
		dragMode = new DragMode(DragMode.NONE);
		//////////////////////////////////////////////////////
		cursorListeners = new Vector();

	}
	
	/**
	 * Constructor. 
	 * Creates a function with the specified color and expression form
	 * Initializes stroke values and dragMode
	 */
	public DefaultGraphFunction(Color color, int expressionForm)
	{
		this();
		this.color = color;
		this.expressionForm = expressionForm;		
	}

	/*
	 * Sets the color of the graph 
	 */
	public void setColor(Color c)
	{
		color = c;
	}

	/*
	 * Returns the color of the graph 
	 */
	public Color getColor()
	{
		return color;
	}

    /** 
	 * Returns the string representation of this function
	 * depending on the expression form this function has
	 */
	public String getExpression()
	{
		return getExpression(expressionForm);
	}
	
	/**
	 * Returns the expression form of this function
	*/
	public int getExpressionForm()
	{
		return expressionForm;
	}
	
	/**
	 * Sets the expression form of this function
	*/
	public void setExpressionForm(int form)
	{
		expressionForm = form;
	}	
	
	/**
	 * Sets the drag mode of this function
	*/
	public void setDragMode(DragMode dragMode)
	{
		this.dragMode = dragMode;
	}

	/**
	 * Sets the drag mode of this function
	*/
	public void setDragMode(int dragMode)
	{
		this.dragMode = new DragMode(dragMode);
	}

	/**
	 * Returns the drag mode of this function
	*/
	public DragMode getDragMode()
	{
		return this.dragMode;
	}

	/**
	 * Returns the drag mode of this function
	*/
	public int getDragModeValue()
	{
		return this.dragMode.getValue();
	}

    /** 
	 * Returns the name of the function (f, g, etc)
	 */
	public String getFunctionName()
	{
		return functionName;
	}

    /** 
	 * Changes the name of the function (f, g, etc)
	 */
	public void setFunctionName(String fName)
	{
		functionName = fName;
		notifyChange();
	}

    /** 
	 * Returns the ordinal of the function (1,2, etc)
	 * if the ordinal number is 1 and the name is f,
	 * the function will appear as f1
	 */
	public int getFunctionOrdinal()
	{
		return functionOrdinal;
	}

    /** 
	 * Changes the ordinal of the function (1,2, etc)
	 * if the ordinal number is 1 and the name is f,
	 * the function will appear as f1
	 * @param ordinal	if < 0, the ordinal will be set to -1,
	 *					and the function label will 
	 * 					consider only the function name
	 */
	public void setFunctionOrdinal(int ordinal)
	{
		if (ordinal<0){
			functionOrdinal = -1;
		}
		else{
			functionOrdinal = ordinal;
		}
	}

	//////////////////////////////////////
	// Methods from Graphable, Selectable 
	// or MouseControllable interface
	//////////////////////////////////////
	public void setVisible(boolean visible) 
	{
		this.visible = visible;
		notifyChange();
	}
	
	public boolean isVisible() 
	{
		return visible;
	}
	
	/** Adds a GraphableListener to listen for events */
	public void addGraphableListener(GraphableListener gl) 
	{
		graphableListeners.add(gl);
	}
	
	/** Removes a GraphableListener */
	public void removeGraphableListener(GraphableListener gl) 
	{
		graphableListeners.remove(gl);
	}
	
	/*
	 * Adds a SelectableListener to listen for selection events 
	 */
	public void addSelectableListener(SelectableListener l)
	{
		selectableListeners.add(l);
	}
	
	/*
	 * Removes a SelectableListener.
	 */
	public void removeSelectableListener(SelectableListener l)
	{
		selectableListeners.remove(l);
	}

	/** Sends a remove event to any enclosing class. This should remove
	the Graphable from any view. */
	public void remove() 
	{

		notifyRemove();
	}
	
	/** 
	 * Select this GraphFunction 
	 */
	public void select() 
	{
		if (!selected){
			selected = true;		
			notifySelect();
		}
	}
	
	public void deselect() 
	{
		if (selected){
			selected = false;
			notifyDeselect();
			activateGhost(false); // get rid of ghost (if saved)
		}
	}
	
	/** 
	 * Returns if the Graph Function is selected 
	 */
	public boolean isSelected() 
	{
		return selected;
	}

	/**
	 * Sets if the object can be selected or not 
	 */
	public void setSelectionEnabled(boolean b)
	{
		selectable = b;
	}

	/**
	 * Returns if the object can be selected or not 
	 */
	public boolean isSelectionEnabled()
	{
		return selectable;
	}

	public boolean isMouseControlled()
	{
		return mouseClicked;
	}

	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		if (graphArea != null){
			graphArea.removeChangeListener(this);
		}
		this.graphArea = area;
		if (ghost!=null){
			ghost.setGraphArea(graphArea);
		}
		if (graphArea != null){
			graphArea.addChangeListener(this);
		}
	}

	/** 
	 * Returns the graph area where the graphable will be drawn 
	 */
	public GraphArea getGraphArea()
	{
		return this.graphArea;
	}

	public void stateChanged(ChangeEvent e)
	{
		update();
		notifyChange();
	}

	//////////////////////////////////////
	// Methods from Cursorable
	//////////////////////////////////////
	/**
	 * Sets the current cursor
	 */
	public void setCursor(Cursor cursor)
	{
		if (currentCursor != cursor){
			currentCursor = cursor;
			notifyCursorListeners();
		}
	}

	/**
	 * Returns the current cursor
	 */
	public Cursor getCursor()
	{
		return currentCursor;
	}
	
	/**
	 * Adds a listener that will listen to a cursor change
	 */
	public void addCursorListener(CursorListener l)
	{
		cursorListeners.add(l);
	}

	/**
	 * Removes a cursor change listener 
	 */
	public void removeCursorListener(CursorListener l)
	{
		cursorListeners.remove(l);
	}
	
	protected void notifyCursorListeners()
	{
		CursorListener l;
		for (int i=0;i<cursorListeners.size();i++){
			l = (CursorListener)cursorListeners.elementAt(i);
			l.cursorChanged(new EventObject(this));
		}
	}

	//////////////////////////////////////
	
	//////////////////////////////////////
    protected void notifyChange() 
	{
		notifyChange(true);
	}
	
    protected void notifyChange(boolean functionChange) 
	{
//System.out.println("--- notifying change: "+graphableListeners.size());
		GraphFunctionEvent e = new GraphFunctionEvent(this, functionChange);
		Enumeration lists = graphableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			gl.graphableChanged(e);
		}
    }

	protected void notifyRemove() 
	//This method has been modified 
	//after notifying a listener, that listener could have removed
	//itself as a listener, so the looping needs to avoid this situation
	//We might want to use an EventList for this. For now it works.
	{
		GraphFunctionEvent e = new GraphFunctionEvent(this);
		int i = 0;
		int t = graphableListeners.size();
//System.out.println("# listeners: "+t);
		while (i < t){
			GraphableListener gl = (GraphableListener)graphableListeners.elementAt(i);
			gl.graphableRemoved(e);
			t = graphableListeners.size();
			if (i < t && graphableListeners.elementAt(i) == gl){
				i++;
			}
		}
	}

	protected void notifySelect() 
	{
//System.err.println("--- notifying selection: "+this);
		GraphFunctionEvent e = new GraphFunctionEvent(this);
		Enumeration lists = selectableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			if (gl instanceof SelectableListener){
				((SelectableListener)gl).graphableSelected(e);
			}
		}
	}

	protected void notifyDeselect() 
	{
//System.err.println("--- notifying deselection: "+this);
		GraphFunctionEvent e = new GraphFunctionEvent(this);
		Enumeration lists = selectableListeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			if (gl instanceof SelectableListener){
				((SelectableListener)gl).graphableDeselected(e);
			}
		}
	}

	/**
	 * Creates an empty painter that listens to the variable
	*/
	public void setPainter(Value variable, boolean on)
	{
		FunctionPainter p = new FunctionPainter(variable);
		setPainter(p,on);
	}

    /** 
	 * Returns the object to maintain the part of the line that is currently
	 * painted. This is used by the draw method. If this object is
	 * null, it is ignored.
    */
	public void setPainter(FunctionPainter painter)
	{
		this.painter = painter;
	}
	
    /** 
	 * Returns the object to maintain the part of the line that is currently
	 * painted. This is used by the draw method. If this object is
	 * null, it is ignored.
    */
	public void setPainter(FunctionPainter painter, boolean on)
	{
		setPainter(painter);
		if (on){
			setPainterOn();
		}
		else{
			setPainterOff();
		}
	}

	/** Resets the LinePainter object if it is defined. */
	public void resetPainter() 
	{
		if (painter != null) {
			painter.reset();
		}
	}
	
	/**
	 * Turns ON the painter on the function
	 */
	public void setPainterOn()
	{
		if (painter!=null){
			painter.setOn();
		}
	}
	
	/**
	 * Turns OFF the painter on the function
	 */
	public void setPainterOff()
	{
		if (painter!=null){
			painter.setOff();
		}
	}
	
	/**
	 * Changes the line stroke
	 */
	public void setLineStroke(Stroke lineStroke)
	{
		//System.out.println("DefaultGraphFunction: setLineStroke " + lineStroke);
		if (this.lineStroke != lineStroke){		//FIXME: should equals be used?
			this.lineStroke = lineStroke;
			notifyChange(false);
		}
	}

	/**
	 * Returns the line stroke
	 */
	public Stroke getLineStroke()
	{
		//System.out.println("DefaultGraphFunction: getLineStroke " + this.lineStroke);

		return this.lineStroke;
	}

	/**
	 * Changes the line stroke
	 */
	public void setSelectedLineStroke(Stroke selectedLineStroke)
	{

	//System.out.println("DefaultGraphFunction: setSelectedLineStroke " + selectedLineStroke);
		if (this.selectedLineStroke != selectedLineStroke){		//FIXME: should equals be used?
			this.selectedLineStroke = selectedLineStroke;
			notifyChange(false);
		}
	}

	/**
	 * Returns the line stroke
	 */
	public Stroke getSelectedLineStroke()
	{
	//System.out.println("DefaultGraphFunction: getSelectedLineStroke " + this.selectedLineStroke);

		return this.selectedLineStroke;
	}
	//////////////////////////////////////

	/**
	 * Indicates if the function will show a ghost when dragged
	*/
	public void setShowGhost(boolean showGhost)
	{
		this.showGhost = showGhost;
		if (showGhost && ghost==null){
			//By default it creates a SingleGhost
			ghost = new SingleGhost(this);
			ghost.setGraphArea(graphArea);
		}
	}
	
	/**
	 * Activates or deactivates the ghost
	*/
	public void activateGhost(boolean b)
	{
		if (ghost==null || !showGhost) return;
		if (b){
			ghost.activate();
		}
		else{
			ghost.deactivate();
		}
	}
	
	/**
	 * Returns the ghost of the function
	*/
	public GraphFunctionGhost getGhost()
	{
		return ghost;
	}
	
	/**
	 * Sets the ghost for the function
	*/
	public void setGhost(GraphFunctionGhost ghost)
	{
		this.ghost = ghost;
		ghost.setGraphArea(graphArea);
	}

	public void	drawGhost(Graphics2D g)
	{
		if (showGhost && ghost!=null){
			ghost.draw(g);
		}
	}
	
	/*
	 * Creates a Graphical object using the string. The format of the
	 * string is determined by the subclass of Graphical. 
	 */
    public void loadString(String instring)
	{
	}

    /*
	 * Gets a string representation of this object. The format is
	 * determined by the subclass of Graphical 
	 */
    public String getLoadString()
	{
		return "";
	}

	/** 
	 * Updates a path with the function using yval, and ALWAYS
	 * connecting all the points (no singularities or piecewises)
	 */
	public void updateFunction()
	{
		Point2D start, end;
		boolean startPath;
		CoordinateSystem coord;
		
		if (graphArea == null) return;

		if (path == null){
			path = new GeneralPath();
		}
		path.reset();

		coord = graphArea.getCoordinateSystem();
		start = graphArea.getLowerLeftCornerDisplay();
		end = graphArea.getUpperRightCornerDisplay();

		double px, py, x, y;
		Point2D ppoint, point;
		startPath = true;
		//Painting a path Pixel by pixel!!!
		for (px = start.getX(); px < end.getX(); px++){
			point = coord.transformToWorld(new Point2D.Double(px, 0.0));
			y = yval(point.getX());
			if (Double.isNaN(y) || y == Double.NEGATIVE_INFINITY || y == Double.POSITIVE_INFINITY){
				startPath = true;
				continue;
			}
			ppoint = coord.transformToDisplay(new Point2D.Double(0.0, y));
			py = ppoint.getY();
			if (startPath){
				path.moveTo((float)px,(float)py);
				startPath = false;
			}
			else{
				path.lineTo((float)px,(float)py);
			}
		}
	}
	
	/** 
	 * Draws the function using yval, and ALWAYS
	 * connecting all the points (no singularities or piecewises)
	 */
	public void drawFunction(Graphics2D g)
	{
		if (graphArea == null) return;

		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		
		if (path == null){
			updateFunction();
		}
		
		g.setColor(this.getColor());
		if (this.isSelected()){
			g.setStroke(this.getSelectedLineStroke());
		}
		else{
			g.setStroke(this.getLineStroke());
		}
		g.draw(path);
		
		g.setColor(oldColor);
		g.setStroke(oldStroke);
	}

}







