/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * @(#)DomainManager.java
 * 
 * Last modification information:
 * $Revision: 1.4 $
 * $Date: 2004/07/29 15:27:35 $
 * $Author: jqian $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.util.*;

import org.concord.graph.event.*;
import org.concord.graphfunction.event.*;

/**
 * Manages the Domains for Segments in a Piecewise object and keeps a Vector
 * of the places they overlap.
 * The isBad(double) method returns a boolean if it is not a function at this
 * point.
 * The isBadMove(double) method checks if the value of double (presumably a
 * trace) passed through a bad spot since the last time it was checked.
 * isBad() without an argument returns true if there is a problem in the 
 * entire function.
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 *
 */

public class DomainManager 
	implements DomainListener, GraphableListener 
{
	Piecewise piecewise;
	Vector badRanges = new Vector();
	
	
	public DomainManager(Piecewise piecewise) 
	{
		this.piecewise = piecewise;
		piecewise.addDomainListener(this);
		piecewise.addGraphableListener(this);
	}
	
	
	/** Check to see if xval is in a "bad" place */
	public boolean isBad(double xval) 
	{
		boolean retval = false;
		Enumeration baddies = badRanges.elements();
		while ((baddies.hasMoreElements()) && (!retval)) {
			Domain baddie = (Domain)baddies.nextElement();
			retval = baddie.isInside(xval);
		}
		return retval;
	}
	
	/** Check to see if xval has passed over a "bad" place */
	public boolean isBad(double last, double xval) 
	{
		boolean retval = false;
		
		if (Double.isNaN(last)) {
			retval = isBad(xval);
		} else {
			Enumeration baddies = badRanges.elements();
			while ((baddies.hasMoreElements()) && (!retval)) {
			Domain baddie = (Domain)baddies.nextElement();
			retval = baddie.isOver(last,xval);
			}
		}
		return retval;	
	}
	
	public void resetBad() 
	{
		// Clear the (old) bad ranges
		badRanges.clear();
		
		for (int t = 0 ; t < piecewise.segments.size() - 1; t++ ) {
			SegmentDomain dom1 = 
				((Segment)
				 piecewise.segments.elementAt(t)).getDomain();
			for ( int u = t+1; u< piecewise.segments.size(); u++) {
				Domain dom2 = 
					((Segment)
					 piecewise.segments.elementAt(u)).getDomain();
				Domain badRange = dom1.getIntersection(dom2);
				if (badRange != null) {
					badRanges.add(badRange);
				}
		
			} // for u
		} // for t 
		
	}
	
	// Methods to implement DomainListener
	public void domainChanged(DomainEvent e) {
		resetBad();
	}
	
	public void multipleDomains(DomainEvent d) {} // a stub for now
	public void noDomain(DomainEvent d) {} // a stub for now
	
	public void graphableChanged(EventObject e){

		if (e instanceof PiecewiseEvent) {
			if (
                (((PiecewiseEvent)e).getId() ==	PiecewiseEvent.PIECEWISE_SEGMENT_REMOVED) ||
                (((PiecewiseEvent)e).getId() ==	PiecewiseEvent.PIECEWISE_SEGMENT_ADDED) ||
				(((PiecewiseEvent)e).getId() == PiecewiseEvent.PIECEWISE_ALL_SEGMENTS_REMOVED)
               )
				{
					resetBad();
				}
		}
	}

	public void graphableRemoved(EventObject e)
	{
		resetBad();
	}

}


