/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * @(#)EquationTrace.java
 * 
 * Last modification information:
 * $Revision: 1.3 $
 * $Date: 2004/07/14 17:12:13 $
 * $Author: ebrown-munoz $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import org.concord.graph.engine.CoordinateSystem;

/**
 * A trace that paints equations differently
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid Moncada<p>
*/
public class EquationTrace extends Trace
{
	private double threshold = 0.05;
	private Font labelFont = new Font("SansSerif",Font.PLAIN,12);
	
	private boolean multipleSolution;
	private int numGraphsMultipleSolution;
	
	private Stroke projectionBaseStroke = new BasicStroke(3.0f,BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
	private Stroke projectionStroke = new BasicStroke(1.0f,BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);

    /** Constructs a Trace Object initialized with the provided
	Graphable and Value objects */
    public EquationTrace(Value xvalue) 
	{
		super(xvalue);
    }

    /** Constructs a Trace Object initialized with the provided
	Graphable and Value objects */
    public EquationTrace(GraphFunction subject, Value xvalue ) 
	{
		super(subject, xvalue);
    }
	
	private boolean isMultipleSolution()
	{
		int n=0;
		double value = Double.NaN;
		double yvalue = Double.NaN;
		// If this is a GraphableList we should loop through each element
		if (functionList != null) { 
			Enumeration graphables = functionList.elements();
			while (graphables.hasMoreElements ()) {
				GraphFunction graph = (GraphFunction)graphables.nextElement();
				if (graph instanceof FunctionEquation){
					FunctionEquation eq = (FunctionEquation)graph;
					value = xvalue.getValue();
					yvalue = graph.yval(value);
					if ( Math.abs(value - eq.getSolution()) < threshold &&
						Math.abs(yvalue - eq.getResultValue().getValue()) < threshold ){
					//The x solves the equation!
						n++;
					}
				}
			} // while graphables.hasMoreElements()
		}		
		
		if (n>1){
			return true;
			//return new Point2D.Double(value, yvalue);
		}
		else{
			return false;
			//return null;
		}
	}
	
    public void draw(Graphics2D g) 
	{
		multipleSolution = isMultipleSolution();
		numGraphsMultipleSolution = 0;
		super.draw(g);		
	}
	
	protected void drawGraph(Graphics2D g, GraphFunction graph) 
	{
		super.drawGraph (g, graph);
		if (graph instanceof FunctionEquation){
			FunctionEquation eq = (FunctionEquation)graph;
			if ( tracePoint!=null ) {
				CoordinateSystem cs = getGraphArea().getCoordinateSystem();

				Color oldColor = g.getColor();
				Stroke oldStroke = g.getStroke();
				
				double value = xvalue.getValue();
				double yval = graph.yval(value);

				double solution = eq.getSolution(); // xvalue that solves it
				
				if ( Double.isInfinite(solution) ||
					 (Math.abs(value - solution) < threshold) ) {

					//The x solves the equation!
				
					Point2D.Double pOrigin = new Point2D.Double();
					cs.transformToDisplay ( new Point2D.Double(0, 0), pOrigin);
					
					//Draw projection on x axis 
					g.setColor(Color.yellow);
					g.setStroke(projectionBaseStroke);
					g.drawLine((int)tracePoint.x, (int)tracePoint.y, (int)tracePoint.x, (int)pOrigin.y);
					g.setColor(Color.black);
					g.setStroke(projectionStroke);
					g.drawLine((int)tracePoint.x, (int)tracePoint.y, (int)tracePoint.x, (int)pOrigin.y);

					//Draw projection on y axis 
					g.setColor(Color.yellow);
					g.setStroke(projectionBaseStroke);
					g.drawLine((int)tracePoint.x, (int)tracePoint.y, (int)pOrigin.x, (int)tracePoint.y);
					g.setColor(Color.black);
					g.setStroke(projectionStroke);
					g.drawLine((int)tracePoint.x, (int)tracePoint.y, (int)pOrigin.x, (int)tracePoint.y);
					
					//Draw label with the value x
					String label = "x = " + MathUtil.getStringRound(value,2);
					FontMetrics labelMetrics = g.getFontMetrics(labelFont);
					int w = labelMetrics.stringWidth(label);
					int h = labelMetrics.getHeight();
					int xPos, yPos;
					
					xPos = (int)(tracePoint.x - (w/2));
					if (yval>=-0.1){
						yPos = (int)(pOrigin.y + 20);
					}
					else{
						yPos = (int)(pOrigin.y - h);
					}
					g.setColor(Color.yellow);
					g.fill(new Rectangle2D.Double(xPos, yPos, w + 3, h));

					g.setColor(graph.getColor());
					g.draw(new Rectangle2D.Double(xPos, yPos, w + 3, h));
					
					g.setColor(Color.black);
					g.setFont(labelFont);
					g.drawString(label, xPos + 2, yPos + h - 2);
					
					if (multipleSolution){
						//Paint multiple solution effect!
						g.setColor(graph.getColor());
						g.setStroke(projectionBaseStroke);
						g.drawOval((int)(tracePoint.x)- (7 + (numGraphsMultipleSolution*3)), 
						           (int)(tracePoint.y)- (7 + (numGraphsMultipleSolution*3)),
								   14 + (numGraphsMultipleSolution*6),
								   14 + (numGraphsMultipleSolution*6));
						numGraphsMultipleSolution++;
					}
				}
				
				g.setColor(oldColor);
				g.setStroke(oldStroke);
				
			} // if ( tracePoint!=null )
		}//if (graph instanceof Equation)
	} // drawGraph()      

}



