/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * @(#)FunctionEquation.java
 * 
 * Last modification information:
 * $Revision: 1.3 $
 * $Date: 2004/07/14 17:12:13 $
 * $Author: ebrown-munoz $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

/**
 * An equation with one variable (x)
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid Moncada<p>
 *
*/

import java.util.*;

import java.awt.*;
import java.awt.geom.*;

import org.concord.graphfunction.event.*;
import org.concord.graphfunction.engine.*;
import org.concord.graph.event.*;
import org.concord.graph.engine.*;

public abstract class FunctionEquation extends DefaultGraphFunction
	implements Equation, ValueListener, GraphableListener
{
	//Expression form constants
	public static final int FORM_FUNCTION = 20;
	public static final int FORM_EQUATION = 21;

	protected GraphFunction function;	// Left-side expression of the equation
	protected Value valueResult;		// Right-side value of the equation
	
	//Equation point moving
	private boolean moveEquationPoint;	// Internal variable that indicates if the equation point (yellow dot) is being moved
	protected boolean allowMoveEquationPoint = false;
	protected Value originalValueResult;
	
	public FunctionEquation()
	{
		setExpressionForm(FORM_EQUATION);
	}
	/**
	 * Sets the function associated with this equation
	*/
	public void setFunction(GraphFunction f)
	{
		if (function!=null){
			function.removeGraphableListener(this);
		}
		function = f;
		if (function!=null){
			function.addGraphableListener(this);

			function.setGraphArea(getGraphArea());
		}

	}

	/**
	 * Returns the function associated with this equation
	*/
	public GraphFunction getFunction()
	{
		return function;
	}

	public void setGraphArea(GraphArea area) {
		super.setGraphArea(area);
		if (function != null) {
			function.setGraphArea(area);
		}
	}

	/**
	 * Sets the result value of the equation
	 * i.e. the value on the right side of the equation
	*/
	public void setResultValue(Value val)
	{
		if (valueResult!=null){
			valueResult.removeValueListener(this);
		}
		valueResult = val;
		if (valueResult!=null){
			valueResult.addValueListener(this);
		}
		notifyChange();
	}
	
	/**
	 * Returns the result value of the equation
	 * i.e. the value on the right side of the equation
	*/
	public Value getResultValue()
	{
		return valueResult;
	}

	/**
	 * Listener for the changes in the variable value
	*/
    public void valueChanged(ValueEvent e)
	{
		this.notifyChange();
	}

	/**
	 * Sets if the user will be able to drag the equation point (yellow dot)
	 */
	public void setAllowMoveEquationPoint(boolean allow)
	{
		allowMoveEquationPoint = allow;
	}

	/**
	 * Returns if the user will be able to drag the equation point (yellow dot)
	 */
	public boolean getAllowMoveEquationPoint()
	{
		return allowMoveEquationPoint;
	}
	
	//////////////////////////////////////
	// Methods from GraphFunction class
	//////////////////////////////////////
    /** 
	 * Returns a string representation of the equation
	*/
    public String getExpression(int form) 
	{
		if (function==null) return "";
		
		if (form == FORM_FUNCTION){
			return function.getExpression();
		}
		else if (form == FORM_EQUATION){
			double value = Double.NaN;
			if (valueResult!=null){
				value = valueResult.getValue();
			}
			return function.getExpression() + " = " +
				MathUtil.roundTenth(getResultValue().getValue(), 10.0);
		}
		else{
			return "Equation: form " + form + " not implemented";
		}
    }

    /** Returns the value of the function at 'xval' */
    public double yval( double xval)
	{
		if (function==null) return Double.NaN;
		
		return function.yval(xval);
	}

    /** Creates a Graphical object using the string. The format of the
	string is determined by the subclass of Graphical. */
    public void loadString(String instring )
	{
	}

    /** Gets a string representation of this object. The format is
	determined by the subclass of Graphical */
    public String getLoadString()
	{
		return "";
	}
	//////////////////////////////////////

	//////////////////////////////////////
	// Listener to the changes of the Graphable (function)
	// When the function changes, I change too (Equation)
    public void graphableChanged(EventObject e) 
	{
		notifyChange();
    }

    public void graphableRemoved(EventObject e) 
	{
		//notifyRemove();
    }
	//////////////////////////////////////

	////////////////////////////////////////////////////////////////
	// Methods of the interface Graphable
	////////////////////////////////////////////////////////////////
    /** Forces classes to redraw any internal image buffer (if implemented).
	 * 	This makes sure the background is updated if there is a fundamental
	 * 	change that is not checked by this class. 
	 */
    public void update()
	{
	}

    /** Draws this object on Graphics g */
    public void draw( Graphics2D g)
	{
		//super.draw( g);

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
				
		if (function==null) return;
			
		function.draw(g);
		
		if (!function.isVisible()) return;
		
		//Draw the solution of the equation as a yellow spot

		Color oldColor = g.getColor();
	    // A spot for the tracePoint in screen coords
	    Point2D.Double point = new Point2D.Double();

	    double xval = this.getSolution();
	    double yval = this.getResultValue().getValue();
		
	    if (! Double.isNaN(xval) && ! Double.isNaN(yval)) {
			cs.transformToDisplay ( new Point2D.Double( xval, yval), point);
			g.setColor(this.getColor());
			g.fillOval((int)(point.x)-6, (int)(point.y)-6,12,12);
			g.setColor(new Color(255,255,0));
			g.fillOval((int)(point.x)-5, (int)(point.y)-5,10,10);
	    }
		
		g.setColor(oldColor);
	}

    /** Checks if the location is close enough to the equation point (yellow dot) */
    protected boolean isPointInProximityToEquationPoint( Point location)
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

	    Point2D.Double point = new Point2D.Double();
	    double xval = this.getSolution();
		if (Double.isInfinite(xval)) { // any point on line is a solution
			return isPointInProximity(location); // ... if close to line
		} else {
			double yval = this.getResultValue().getValue();
			double dx = 4.0; //threshold in pixels
			double dy = 4.0; //threshold in pixels
		
			if (! Double.isNaN(xval) && ! Double.isNaN(yval)) {
				cs.transformToDisplay( new Point2D.Double(xval, yval), point);
			
				if (MathUtil.equalsDouble(point.x,location.x,dx) && 
					MathUtil.equalsDouble(point.y,location.y,dy)){
			
					return true;
				}
			}
		}
		return false;
	}

    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
    public boolean isPointInProximity(Point location)
	{
		if (function==null) return false;
		return function.isPointInProximity(location);
	}

    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
    public boolean mousePressed(Point location)
	{
		//super.mousePressed(location);
		if (function==null) return false;
		moveEquationPoint = false;
		//Detect if the equation point will be moved
		if (allowMoveEquationPoint){
			if (isPointInProximityToEquationPoint(location)) {
				originalValueResult = new Value(valueResult.getValue());
				moveEquationPoint = true;
				return true;
			}
		}
		//
		return function.mousePressed(location);
	}

    /** Checks for a mouse drag. */
    public boolean mouseDragged(Point location)
	{
		if (function==null) return false;
		if (allowMoveEquationPoint){
			if (moveEquationPoint){
				return mouseDraggedEquationPoint(location);
			}
		}
		return function.mouseDragged(location);
	}

    /** Checks for a mouse release. */
    public boolean mouseReleased(Point location)
	{
		//super.mouseReleased(location);

		moveEquationPoint = false;
		if (function==null) return false;
		return function.mouseReleased(location);
	}
	
	protected abstract boolean mouseDraggedEquationPoint(Point location);
	////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////
	//Methods to delegate to the function
    public void select() 
	{
		if (function==null) return;
		function.select();
		super.select();
    }

    public void deselect() 
	{
		if (function==null) return;
		function.deselect();
		super.deselect();
    }
	
    /** Removes the object from any enclosing view. */
    public void remove()
	{
		if (function==null) return;
		function.remove();  //???????
		super.remove();
	}

	public void setColor(Color c)
	{
		super.setColor(c);
		if (function==null) return;
		function.setColor(c);
	}
	
	/**
	 * Creates an empty painter that listens to the variable
	*/
	public void setPainter(Value variable, boolean on)
	{
		super.setPainter(variable, on);
		if (function==null) return;
		function.setPainter(variable, on);
	}

    /** 
	 * Returns the object to maintain the part of the line that is currently
	 * painted. This is used by the draw method. If this object is
	 * null, it is ignored.
    */
	public void setPainter(FunctionPainter painter)
	{
		super.setPainter(painter);
		if (function==null) return;
		function.setPainter(painter);
	}
	
    /** 
	 * Returns the object to maintain the part of the line that is currently
	 * painted. This is used by the draw method. If this object is
	 * null, it is ignored.
    */
	public void setPainter(FunctionPainter painter, boolean on)
	{
		super.setPainter(painter, on);
		if (function==null) return;
		function.setPainter(painter, on);
	}

	/** Resets the LinePainter object if it is defined. */
	public void resetPainter() 
	{
		super.resetPainter();
		if (function==null) return;
		function.resetPainter();
	}
	
	/**
	 * Turns ON the painter on the function
	 */
	public void setPainterOn()
	{
		super.setPainterOn();
		if (function==null) return;
		function.setPainterOn();
	}
	
	/**
	 * Turns OFF the painter on the function
	 */
	public void setPainterOff()
	{
		super.setPainterOff();
		if (function==null) return;
		function.setPainterOff();
	}

	/**
	 * Changes the line stroke
	 */
	public void setLineStroke(Stroke lineStroke)
	{
		if (function==null) return;
		function.setLineStroke(lineStroke);
	}

	/**
	 * Changes the line stroke
	 */
	public void setSelectedLineStroke(Stroke selectedLineStroke)
	{
		if (function==null) return;
		function.setSelectedLineStroke(selectedLineStroke);
	}
	
	
	////////////////////////////////////////////////////////////////
}

