/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.6 $
 * $Date: 2004/07/16 20:44:16 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;

import org.concord.graph.engine.*;
/**
 * Class name and description
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid C Moncada<p>
 *
 */

public class LinearFunction extends DefaultGraphFunction
{
	//Expression form constants	
	public static final int FORM_SLOPEINTERCEPT = 1;
	public static final int FORM_POINTSLOPE = 2;
	public static final int FORM_FUNCTIONPOINTSLOPE = 3;

	// The equation for the line.
	protected double slope;			//slope
	protected double yIntercept;	//intercept with the y axis
	
	private double slopeInterval = 0.1;
	private double yInterceptInterval = 0.1;
	private int slopePrecision = 10;
	private int yInterceptPrecision = 10;
	
	//Temp
    protected Point2D.Double clickLocation; 
    protected DragState original;
	
	/**
	 * Default constructor. 
	 * By default it creates
	 * y = x black function, in the slope-intercept form
	 */
	public LinearFunction()
	{
		this(1.0, 0.0, Color.black, FORM_SLOPEINTERCEPT);
	}
	
	/**
	 * Constructor that creates
	 * a function with the specified slope and y intercept
	 * black and in the slop-intercept form
	 */
	public LinearFunction(double slope, double yIntercept)
	{
		this(slope, yIntercept, Color.black, FORM_SLOPEINTERCEPT);
	}

	/**
	 * Constructor that creates
	 * a function with the specified slope, y intercept, color and form
	 */
	public LinearFunction(double slope, double yIntercept, Color color, int expressionForm)
	{
		super(color,expressionForm);
		this.slope = slope;
		this.yIntercept = yIntercept;
		//Default drag mode is translate?
		dragMode = new DragMode(DragMode.TRANSLATE);
		clickLocation = new Point2D.Double();
	}
	
	/**
	 * Returns the slope of the linear function
	 */
	public double getSlope() 
	{
		return slope;
	}
	
	/**
	 * Sets the slope of the linear function
	 */
	public void setSlope(double slope) 
	{
		this.slope = slope;
		notifyChange();
	}

	/**
	 * Returns the value where the linear function intersects with the y axis
	 */
	public double getYIntercept()
	{
		return yIntercept;
	}
	
	/**
	 * Sets the value where the linear function intersects with the y axis
	 */
	public void setYIntercept(double yIntercept) 
	{
		this.yIntercept = yIntercept;
		notifyChange();
	}

	/** Reflect this function about the y axis. */
	public void reflectX() {
		slope = (-slope);
		yIntercept = (-yIntercept);
		notifyChange();
	}
	
	/** Reflects this function about the x axis. */
	public void reflectY() {
		slope = - slope;
		notifyChange();
	}

    /** 
	 * Returns the string representation of this function
	 * for a specific expression form
	 */
	public String getExpression(int form)
	{
		if (form == FORM_SLOPEINTERCEPT){
			return new String( "y = " + 
				MathUtil.roundTenth(slope, getSlopePrecision()) + "x + " + 
				MathUtil.roundTenth(yIntercept, getYInterceptPrecision()));
		}
		else if (form == FORM_POINTSLOPE){
		//No pushpin... so? assume x0 = 0 for now...
			return new String( "y - "+
			MathUtil.roundTenth(yIntercept, getYInterceptPrecision())+" = " + 
			MathUtil.roundTenth(slope, getSlopePrecision()) + " ( x - 0.0 )");		
		}
		else if (form == FORM_FUNCTIONPOINTSLOPE){
		//No pushpin... so? assume x0 = 0 for now...
			return new String( "y = " + 
			MathUtil.roundTenth(slope, getSlopePrecision()) + " ( x - 0.0 ) + " +
			MathUtil.roundTenth(yIntercept, getYInterceptPrecision()));		
		}
		return "Linear Function form " + form + " not implemented";
	}

	/**
	 * Sets the default form expression for this function
	*/
	public void setExpressionForm(int mode)
	{
		if (mode == FORM_SLOPEINTERCEPT || mode == FORM_POINTSLOPE ||
		    mode == FORM_FUNCTIONPOINTSLOPE){
				
			super.setExpressionForm(mode);
			notifyChange();
		}
	}

	public double getSlopeInterval()
	{
		return slopeInterval;
	}
	
	public void setSlopeInterval(double i)
	{
		if (i<0.0) return;
		slopeInterval = i;
		if (slopeInterval > 0 && slopeInterval < 1){
			slopePrecision = Math.abs(MathUtil.log10(slopeInterval));
			slopePrecision = (int)Math.pow(10,slopePrecision);
		}
		else{
			slopePrecision = 1;
		}
	}

	public double getYInterceptInterval()
	{
		return yInterceptInterval;
	}
	
	public void setYInterceptInterval(double i)
	{
		if (i<0.0) return;
		yInterceptInterval = i;
		if (yInterceptInterval > 0 && yInterceptInterval < 1){
			yInterceptPrecision = Math.abs(MathUtil.log10(yInterceptPrecision));
			yInterceptPrecision = (int)Math.pow(10,yInterceptPrecision);
		}
		else{
			yInterceptPrecision = 1;
		}
	}

	public int getSlopePrecision()
	{
		return slopePrecision;
	}

	public int getYInterceptPrecision()
	{
		return yInterceptPrecision;
	}
	
    /*
	 * Returns the value of the function at 'xval' 
	 */
    public double yval(double xval)
	{
		return (slope * xval) + yIntercept;
	}

    /*
	 * Creates a Graphical object using the string. The format of the
	 * string is determined by the subclass of Graphical. 
	 */
    public void loadString(String instring)
	{
	}

    /*
	 * Gets a string representation of this object. The format is
	 * determined by the subclass of Graphical 
	 */
    public String getLoadString()
	{
		return "";
	}

	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
    /** Forces classes to redraw any internal image buffer (if implemented).
	 * 	This makes sure the background is updated if there is a fundamental
	 * 	change that is not checked by this class. 
	 */

    /** Draws this object on the specified graphics context */
	public void draw(Graphics2D g) 
	{	
		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		if (isVisible()) {
			double domainStart;
			double domainEnd;

			if (isSelected()) {
				drawGhost(g); // FIXME
			}
		
			if ( (painter != null ) && (painter.isOn()) ) {
				domainStart = painter.xMin;
				domainEnd = painter.xMax;
			} else {
				Rectangle clip = g.getClipBounds();
			
				Point2D leftPoint = new Point2D.Double( clip.x , 0.0 );
				Point2D rightPoint = 
					new Point2D.Double( clip.x + clip.width , 0.0 );
		
			// Left and right edges in real coordinates go here.
				Point2D.Double leftScreen = new Point2D.Double();
				Point2D.Double rightScreen = new Point2D.Double();
				cs.transformToWorld ( leftPoint, leftScreen) ;
				cs.transformToWorld ( rightPoint, rightScreen) ;
				domainStart = leftScreen.x;
				domainEnd = rightScreen.x;
			}	
			// Translate back to screen coordinates (with yvals) and draw
		
			Point2D start = new Point2D.Double();
			cs.transformToDisplay(new Point2D.Double(domainStart,
													 yval(domainStart)),
								  start);
			Point2D end = new Point2D.Double();
			cs.transformToDisplay(new Point2D.Double(domainEnd, 
													 yval(domainEnd)),
								  end);
		
			Stroke saveStroke = g.getStroke();
			Color saveColor = g.getColor();
		
			if (isSelected()) {
				g.setStroke(getSelectedLineStroke());
			}
			else{
				g.setStroke(getLineStroke());
			}			
			
			g.setColor(this.getColor());
			g.draw(new Line2D.Double(start, end));
		
			g.setStroke(saveStroke);
			g.setColor(saveColor);
		
		} // if visible
		
	}//draw

    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
	// NOT IMPLEMENTED YET
    public boolean isPointInProximity(Point location)
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		boolean retval = false; // the return value
		
		if (isVisible()) {	

			// Transform the Y intercept into screen coords
			Point2D.Double interceptInScreen = new Point2D.Double();
			cs.transformToDisplay(new Point2D.Double(0.0,yIntercept),
					  interceptInScreen);
	
			// Transform other point to screen coords
			Point2D.Double otherPointInScreen = new Point2D.Double();
			cs.transformToDisplay(new Point2D.Double(1.0, slope+yIntercept),
					  otherPointInScreen);
	
			// Angle of tranformed line (on screen)
			double angle = 
				Math.atan2((otherPointInScreen.y -interceptInScreen.y),
					   (otherPointInScreen.x -interceptInScreen.x));
					  
			// Since the line can have any slope, you must calculate the
			// distance between the click and the *nearest* point on the
			// line. (Subtracting the y-values of the points will not 
			// work well for lines with a steep slope.
			double dist_from_line = Math.sin(Math.PI/2.0 - angle) *
				((double)location.y - 
				 ((double)(location.x - interceptInScreen.x) *
				  Math.tan(angle) + 
				  interceptInScreen.y));
	
			if (Math.abs(dist_from_line) < 6.0) { // this is in pixels
				// Calculate the handle (screen coords) you want
				double handleX =  location.x + 
				dist_from_line * Math.sin(angle);
				double handleY =  interceptInScreen.y + 
				(handleX - interceptInScreen.x) * Math.sin(angle);
				retval = true;
	
				cs.transformToWorld(location, clickLocation);
	
			}
		} // if visible

		return retval;
	}

	/** Update the case (do nothing for now */
	public void update() {
	}

    /** Handles a mousePressed event. 
	Only does translations.
    */
	public boolean mousePressed(Point location) 
	{

		activateGhost(true);

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		cs.transformToWorld(location, clickLocation); // store the click
		
		original = new DragState(this);

		return true;
		
	}//mousePressed
 
	public boolean mouseDragged( Point location) 
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		boolean retval = false;
	
		if (dragMode.getValue() == DragMode.TRANSLATE) {

			Point2D.Double transformedLocation = new Point2D.Double();
			cs.transformToWorld(location, transformedLocation);
			
			double changeX = transformedLocation.x - clickLocation.x;	
			double changeY = transformedLocation.y - clickLocation.y;	
	
			retval = true;
			
			// move line horizontally
			setYIntercept( original.yIntercept - original.slope * changeX);
			
			// move line vertically
			setYIntercept(yIntercept + changeY);
	
			notifyChange();
		}  	
		return retval;
	
	}//mouseDragged

	public boolean mouseReleased(Point location) 
	{

		activateGhost(false);
		return true;
		
	}//mouseReleased

	/*
	 * Returns a clone of itself 
	 */
	public Graphable getCopy()
	{
		LinearFunction retval = new LinearFunction();
		retval.slope = this.slope;
		retval.yIntercept = this.yIntercept;
		retval.setColor(this.getColor());
		if (this.painter!=null){
			retval.painter = this.painter.getCopy();
		}
		return retval;
	}
	//////////////////////////////////////
	
	class DragState {
		double slope;
		double yIntercept;
	
		public DragState(LinearFunction parent) 
		{
			slope = parent.slope;
			yIntercept = parent.yIntercept;
		}	
	
	}
} // class LinearFunction
