/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * @(#)LinearFunctionPushpin.java
 * 
 * Last modification information:
 * $Revision: 1.6 $
 * $Date: 2004/08/31 15:22:44 $
 * $Author: ebrown-munoz $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;

import org.concord.graph.engine.*;

/**
 * The class to manage a Linear Function with a pushpin. This is the main
 * Graphable for the LinearTransformer Applet.
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 *
 */

public class LinearFunctionPushpin extends LinearFunction 
{
	// use lineStroke (in LinearObject) to change the width of the line.
	protected Stroke pushpinStroke = new BasicStroke(3.0f);
	
	protected boolean pushpinOn = false; // i.e. the pushpin object.
		
	protected LinearPushpinObject pushpin;
	private double pushpinX; // temp storage for drag operations
	private double pushpinY;
	
	public LinearFunctionPushpin() 
	{
		super();
		pushpin = new LinearPushpinObject(this);
	}
	
	public void setSlope(double slope) 
	{
	
		if ( pushpinOn ) {
	
			double pushpinY = yval ( pushpin.xPosition);
			this.slope = slope;
			// Make line rotates around pushpin
			this.yIntercept = pushpinY - slope*pushpin.xPosition;
		} else {
			this.slope = slope;
		}
	
		notifyChange();
	}
	
	public void setPushpin(boolean state) 
	{
		if ( this.pushpinOn != state) { // only send event if changed
			this.pushpinOn = state;
			//System.err.println("Pushpin state = " + state);
	
			notifyChange();
		}
	}
	
	/** Returns the pushpin object */
	public LinearPushpinObject getPushpin() 
	{
		return pushpin;
	}
	
	
	/** Reflect this function about the x axis. */
	public void reflectX() 
	{
		slope = (-slope);
		yIntercept = (-yIntercept);
		notifyChange();
	}
	
	/** Reflects this function about the y axis. */
	public void reflectY() 
	{
		slope = - slope;
		pushpin.setXPosition(- pushpin.getXPosition() );	//This notifies the change
	}

	/** 
	 * Changes the x0 (x of pushpin) WITHOUT changing ANY other coefficient
	 * in the equation y - y0 = m (x - x0)
	 */
	public void setPushpinX(double x0)
	{
		// Changing the y0 just moves the y-intercept
		double y0 = yval(pushpin.getXPosition());
		yIntercept = y0 - (getSlope() * x0);
		pushpin.setXPosition(x0);						//This notifies the change
	}
	
	/** 
	 * Returns the x0 (x of pushpin)
	 * in the equation y - y0 = m (x - x0)
	 */
	public double getPushpinX()
	{
		return getPushpin().getXPosition();
	}

	/** 
	 * Changes the y0 (y of pushpin) WITHOUT changing ANY other coefficient
	 * in the equation y - y0 = m (x - x0)
	 * yIntercept = y0 - m*x0
	 */
	public void setPushpinY(double y0)
	{
		// Changing the y0 just moves the y-intercept
		double x0 = pushpin.getXPosition();
		setYIntercept(y0 - getSlope()*x0);
	}

	/** 
	 * Gets the y0 (y of pushpin)
	 * in the equation y - y0 = m (x - x0)
	 */
	public double getPushpinY()
	{
		return yval(getPushpin().getXPosition());
	}

	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
	public void draw( Graphics2D g) 
	{

		if (isVisible()) {
			super.draw(g);
				
			Stroke saveStroke = g.getStroke();
			Color saveColor = g.getColor();
			g.setStroke(pushpinStroke);
	
			if (pushpinOn) {
				pushpin.draw(g);
			}
	
			g.setStroke(saveStroke);
			g.setColor(saveColor);
		} // if visible
	}
	
	/** Handles a mousePressed event. This is different than its analogue
	in LinearPiecewise segment because there are no handles. This class
	will use the pushpin and the point on the line closest to the
	click as handles. 
	*/
	
	public boolean mousePressed(Point location) 
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		boolean retval = false; // the return value
		
		if (isVisible()) {	
			if (pushpinOn) {
				retval = pushpin.mousePressed(location);
			}
	
			if (retval) {
				setDragMode(DragMode.MOVE_PUSHPIN);
			} 
			else { 	// retval == false 
				retval = super.mousePressed(location);
				if (retval){
					if (pushpinOn) {
						pushpinX = pushpin.xPosition; // MOVE this to correct place
						pushpinY = yval(pushpinX);
						
						setDragMode(DragMode.DILATE);
					} 
					else {
						setDragMode(DragMode.TRANSLATE);
					}
				}
			} // (retval == false)
		} // if visible
	
		return retval;
	
	}
	
	public boolean mouseDragged(Point location) 
	{
		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		boolean retval = false;
	
		if (dragMode.getValue() == DragMode.MOVE_PUSHPIN) {
			
			retval = pushpin.mouseDragged (location);
		} 
		else if (dragMode.getValue() == DragMode.TRANSLATE){
			
			retval = super.mouseDragged( location);
		}
		else if (dragMode.getValue() == DragMode.DILATE){
			
			Point2D.Double transformedLocation = new Point2D.Double();
			cs.transformToWorld(location,
								transformedLocation);

			/* Check to make sure slope does not pass 90 degrees.
			 * (the location must be on the same side of the pushpin as
			 * The original click
			 */
			if ((( clickLocation.x > pushpinX) &&
 				 ( transformedLocation.x > pushpinX)) ||
				((clickLocation.x < pushpinX) &&
				 ( transformedLocation.x < pushpinX))) {
				
				double changeX = transformedLocation.x - clickLocation.x;	
				double changeY = transformedLocation.y - clickLocation.y;	
	
				setSlope((clickLocation.y - pushpinY + changeY) /
						 ( clickLocation.x - pushpinX + changeX));
	
				notifyChange();
			}
		
		} // dragMode == DILATE
		
		return retval;
	
	}
	
	public boolean mouseReleased(Point location) 
	{
		boolean retval = super.mouseReleased(location);
		
		setDragMode(DragMode.NONE);
		
		if (pushpinOn) {
			pushpin.mouseReleased(location);
		}
		
		return retval;
	}

	/** Creates a clone of this obect */
	public Graphable getCopy() 
	{
		LinearFunctionPushpin retval = new LinearFunctionPushpin();
		retval.slope = this.slope;
		retval.yIntercept = this.yIntercept;
		retval.setColor(this.getColor());
		retval.pushpinOn = this.pushpinOn;
		retval.showGhost = this.showGhost;
		retval.ghost = this.ghost;
		retval.pushpin.xPosition = this.pushpin.xPosition;
		if (this.painter!=null){
			retval.painter = this.painter.getCopy();
		}
		return retval;
	}
	//////////////////////////////////////
	
	/** Returns true if the pushpin in on */
	public boolean isPushpinOn() 
	{
		return pushpinOn;
	}
	
	public void setPushpinFixed(boolean fixed) 
	{
		pushpin.setFixed(fixed);
	}

	/** Returns a string representation of the pushpinlinear in the specified form. */
	public String getExpression(int form)
	{
		//Only override the point-slope form
		if (form == FORM_POINTSLOPE){
			//Use pushpin as (x0,y0)
			double xval = pushpin.getXPosition();
			return new String( "y - "+
				MathUtil.roundTenth(yval(xval), getYInterceptPrecision())+" = " + 
				MathUtil.roundTenth(slope, getSlopePrecision()) + " ( x - " +
				MathUtil.roundTenth(xval, getSlopePrecision()) + " )");
		}
		else if (form == FORM_FUNCTIONPOINTSLOPE){
			//Use pushpin as (x0,y0)
			double xval = pushpin.getXPosition();
			return new String( "y = "+
				MathUtil.roundTenth(slope, getSlopePrecision()) + " ( x - " +
				MathUtil.roundTenth(xval, getSlopePrecision()) + " ) + " +
				MathUtil.roundTenth(yval(xval), getYInterceptPrecision())); 
		}
		else{
			return super.getExpression(form);
		}
	}

} // class PushpinLinear
