/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.7 $
 * $Date: 2004/08/02 21:21:06 $
 * $Author: ebrown-munoz $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;

import sun.dc.pr.PRException;
import org.concord.graphfunction.event.*;
import org.concord.graph.engine.*;

/**
 * LinearSegment represents one segment of a LinearPiecewise graph.
 * The LinearPiecewise class contains a vectors of these segments.
 * Each LinearSegment contains two Points -- start and end. The
 * segment is a straight line between these two points.
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid C Moncada<p>
 *
 */

public class LinearSegment extends LinearFunction
	implements Segment
{
     /** The domain of this segment */
    protected SegmentDomain segmentDomain;
	
	/** The function to display in the segment */
	//protected GraphFunction function;

	//Temp
    /** value to determine whether the user clicked the center or
	an endpoint of the line. It is set by the isPointInProximity() method */
    protected static final int MOVE_START = 1;
    protected static final int MOVE_END = 2;
    protected static final int MOVE_LINE = 3;
    protected int moveMode;

    /** Show handles or not */
    protected boolean handles = true;

    /** Store the value of the click. This is used by mouseDrag */
    Point2D.Double clickLocation = new Point2D.Double(); // xformed-axis coords
    DragState original = null; // Original location for mouse drag

	private double topSlope = 100000000.0;

	/** An arrow shape for the tip of an infinite segment */
	private GeneralPath arrowHead;

	/** The parent of the segment */
	private GraphFunction parent;
	
	/** 
	 * Constructs a Linear segment "y = x" from 0 to 10 
	 */
	public LinearSegment() 
	{
		this(0.0,10.0);
	}
	
    /*
	 * Constructs a Linear segment "y = x" with specified start and 
	 * end x values. 
	 * The default is that the start is "closed" and the end
	 * is "open". 
	 */
	public LinearSegment(double start, double end) 
	{
		this(1.0,0.0,start, true, end, false);
	}

    /*
	 * Constructs a Linear segment from the point
	 * (x0,y0) to the point (x1,y1)
	 * The default is that the start is "closed" and the end
	 * is "open". 
	 */
	public LinearSegment(double x0, double y0, double x1, double y1) 
	{
		this((y1-y0)/(x1-x0), 
			((y1-y0)/(x1-x0))*(-x0) + y0, 
			x0, true, x1, false);
	}

    /*
	 * Constructs a Linear segment with specified slope,
	 * yIntercept, start, end x values, and booleans
	 * that indicate if the interval is closed of open at the start and end. 
	 * The default is that the start is "closed" and the end
	 * is "open". 
	 */
	public LinearSegment(double slope, double yIntercept, double start, boolean closedStart, double end, boolean closedEnd) 
	{
		super(slope, yIntercept);
		setDomain(new SegmentDomain( start , end));
		getDomain().setClosedStart(closedStart);
		getDomain().setClosedEnd(closedEnd);
		setSlopeInterval(0.01);
	}

	/** Constructs a LinearPiecewise segment with specified start and 
	end points. The openness of the two endpoints is specified by
	startopen and endopen.
	public LinearPiecewiseSegment( Point2D start, Point2D end, 
				   boolean startopen, boolean endopen) {}
	*/
	
	protected DomainListener domainListener = new DomainListener() 
	{
		public void domainChanged(DomainEvent e) 
		{
			notifyChange();
		}
		
		public void multipleDomains (DomainEvent e)
		{
		}
		
		// This sould be fired by Piecewise, but it is not implemented yet.
		public void noDomain (DomainEvent e)
		{
		} 
	};
	
	/** Gets the Domain of this Segment as determined by start and end 
	points. */
	public SegmentDomain getDomain() 
	{
		return segmentDomain;
	}	
	
	/** provides a way to call yval without checking to domain */
	public double yval( double xval, boolean checkDomain) 
	{
		if (!checkDomain) {
			return xval * slope + yIntercept;
		} else {
			return yval(xval);
		}
	}	
	
	public double yval( double xval) 
	{
		double retval = Double.NaN;
		// If in domain
		if ( (segmentDomain.getStart() <= xval) && 
			 (xval < segmentDomain.getEnd() )) {
			retval =  xval * slope + yIntercept;
		} else {
			retval =  Double.NaN;
		}
		
		return retval;
	}
	
	/** Sets the slope and domain to a "grid" */
	public void fix() 
	{
		if (getSlopeInterval()!=0){		
			slope = Math.floor(slope/getSlopeInterval() + 0.5) * getSlopeInterval();   
		}
		segmentDomain.fix();
		
	}   
	
	/** Creates a Graphical object using the string. The format of the
	string is determined by the subclass of Graphical. */
	public void loadString (String instring ) {}
		
	/** Gets a string representation of this object. The format is
	determined by the subclass of Graphical */
	public String getString() 
	{
		return "Not Implemented";
	}	
	
	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
	/** 
	 * Draws this segment the same color as the parent,
	 * and thicker if the parent is selected 
	 * Draws handles in the selected segment
	*/
	public void draw(Graphics2D g) 
	{	
		boolean parentSelected = isSelected();
		GraphArea ga = getGraphArea();
		CoordinateSystem cs = ga.getCoordinateSystem();
		
		if (parent != null){
			parentSelected = parent.isSelected();
		}
		
		//If the interval is empty (i.e. start > end)
		//Don't draw anything!
		if (segmentDomain.getStart() > segmentDomain.getEnd()){
			return;
		}
		
		if (showGhost && ghost!=null){
			ghost.draw(g);
		}
		
		// Create original points in axis space
			
		double sdStart = segmentDomain.getStart();
		boolean startInf = Double.isInfinite(sdStart);
		double sdEnd = segmentDomain.getEnd();
		boolean endInf = Double.isInfinite(sdEnd);

	   
		
		// Set infinite termini to the edge of the clip region (in 
		// real coordinates).
		if (startInf ) {
			startInf = true;
			Point2D.Double screenReference = new Point2D.Double();
		
			// Calculates the point coordinates that map to the pixels
			// (0,0) and puts in screenReference;
			cs.transformToWorld(new Point2D.Double(0.0,0.0),
								screenReference);
			sdStart = screenReference.x;
		}
		
		if (endInf ) {
			endInf = true;
			Point2D.Double screenReference = new Point2D.Double();
			Rectangle clip = g.getClipBounds();
			cs.transformToWorld(new Point2D.Double((double)(clip.x+clip.width),
												   0.0),
								screenReference);
			sdEnd = screenReference.x;
		}
									   
		
		
		Point2D startPoint = new Point2D.Double( sdStart,
							 slope * sdStart + 
							 yIntercept);
			
		Point2D endPoint = new Point2D.Double ( sdEnd,
							slope * sdEnd +
							yIntercept);
			
		
		// Create destination points
		Point2D.Double tStartPoint = new Point2D.Double();
		Point2D.Double tEndPoint = new Point2D.Double();
			
		// Transform axis space to screen space
		cs.transformToDisplay(startPoint, tStartPoint);
		cs.transformToDisplay(endPoint, tEndPoint);
			
			
		Color saveColor = g.getColor();
		Stroke saveStroke = g.getStroke();
			
		g.setColor(this.getColor());
		
		if (parentSelected && isSelected()){
			//System.out.println("-LinearSegment " + this.getColor() + " selected stroke");
			g.setStroke(getSelectedLineStroke()); 
		}
		else{
			//System.out.println("-LinearSegment" + this.getColor() + "NonSelected stroke");
			g.setStroke(getLineStroke()); 
		}
            g.draw(new Line2D.Double(tStartPoint, tEndPoint));
		// ---- Draw Arrowheads for infinite segments
		if (startInf || endInf) {
			if (arrowHead == null) createArrowHead();

			AffineTransform infSaveTransform = g.getTransform();
			double angle = Math.atan( (tEndPoint.y - tStartPoint.y) / 
									  (tEndPoint.x - tStartPoint.x));

			if (endInf) {
				g.translate(tEndPoint.x, tEndPoint.y);
				g.rotate(angle);
				g.draw(arrowHead);
				g.setTransform(infSaveTransform);
			}

			if (startInf) {
				g.translate(tStartPoint.x, tStartPoint.y);
				g.rotate(angle + Math.PI);
				g.draw(arrowHead);
				g.setTransform(infSaveTransform);
			}

		}
			
		float lineThickness;
		//lineThickness = ((BasicStroke)g.getStroke()).getLineWidth();
		lineThickness = ((BasicStroke)getLineStroke()).getLineWidth();
			
		g.setColor(Color.black);
		if (parentSelected && handles) {
			if (parentSelected) {
				double handleSize = lineThickness + 4;
				Shape handleShape;
				if (!startInf){
					//Instead of drawing a rectangle, now it draws a circle
					//handleShape = new Rectangle2D.Double(tStartPoint.x-(handleSize/2),
					//			  tStartPoint.y-(handleSize/2),
					//			  handleSize, handleSize);
					handleShape = new Ellipse2D.Double(tStartPoint.x-(handleSize/2),
								  tStartPoint.y-(handleSize/2),
								  handleSize, handleSize);

					g.setStroke(new BasicStroke(1.0f));
					if (segmentDomain.getClosedStart()){
						g.fill(handleShape);
					}
					else{
						g.draw(handleShape);
					}
				}
				if (!endInf ){
					//Instead of drawing a rectangle, now it draws a circle
					//handleShape = new Rectangle2D.Double(tEndPoint.x-(handleSize/2), 
					//			  tEndPoint.y-(handleSize/2), 
					//			  handleSize, handleSize);
					handleShape = new Ellipse2D.Double(tEndPoint.x-(handleSize/2), 
								  tEndPoint.y-(handleSize/2), 
								  handleSize, handleSize);
					
					g.setStroke(new BasicStroke(1.0f));
					if (segmentDomain.getClosedEnd()){
						g.fill(handleShape);
					}
					else{
						g.draw(handleShape);
					}
				}
			}
		}
		
		g.setColor(saveColor);
		g.setStroke(saveStroke);
	}

	/** Loads global variable arrowHead with an arrowhead shape
	   The arrow points up with the tip at (0,0).
	*/ 
	public void createArrowHead () {
		arrowHead = new GeneralPath();
		arrowHead.moveTo(-15.0f,5.0f);
		arrowHead.lineTo(-3.0f, 0.0f);
		arrowHead.lineTo(-15.0f, -5.0f);
	}

	/** Returns true if the click is on the line between the two points */
	public boolean isPointInProximity(Point location) 
	{
		boolean retval = false;
		GraphArea ga = getGraphArea();
		CoordinateSystem cs = ga.getCoordinateSystem();
		Point2D coordScale = cs.getScale();
		
		double angle = Math.atan(slope * coordScale.getY()/coordScale.getX()); // angle on screen
	
		Point2D.Double interceptInScreen = new Point2D.Double();
		
		cs.transformToDisplay(new Point2D.Double(0.0, yIntercept), 
							  interceptInScreen);
	
		double dist_from_line = Math.cos( angle) *
			((double)location.y - 
			 (interceptInScreen.y - (double)(location.x - interceptInScreen.x) *
			  Math.tan(angle)));
	
	
		if ( Math.abs(dist_from_line) < 6.0) {
			// project the clicked point onto the line 
			Point2D.Double projection = 
			new Point2D.Double( (double)location.x + 
						Math.sin(angle) * dist_from_line,
						(double)location.y + 
						Math.cos(angle) * dist_from_line);
								  
	
			Point2D.Double startInScreen = new Point2D.Double();
			Point2D.Double endInScreen = new Point2D.Double();
			
			cs.transformToDisplay(new Point2D.Double(segmentDomain.getStart(),
													 yval(segmentDomain.getStart(),false)),
								  startInScreen);
			
			cs.transformToDisplay(new Point2D.Double(segmentDomain.getEnd(),
													 yval(segmentDomain.getEnd(),false)),
								  endInScreen);
			
			if ( startInScreen.distance(projection) < 6.0) {

				moveMode = MOVE_START;
				retval = true;
	
				//For now: it doesn't let dragging with the handles if the segment is infinite
				if (this.isInfinite()){
					moveMode = MOVE_LINE;
				}
				
			} else if ( endInScreen.distance(projection) < 6.0) {

				moveMode = MOVE_END;
				retval = true;
	
				//For now: it doesn't let dragging with the handles if the segment is infinite
				if (this.isInfinite()){
					moveMode = MOVE_LINE;
				}
				
			} else if (( projection.x >= startInScreen.x) && ( projection.x <= endInScreen.x)) {

				moveMode = MOVE_LINE;
				retval = true;	
			}
		}
	
		return retval;
	}// isPointInProximity

	/** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
	public boolean mousePressed(Point location) 
	{
		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		super.mousePressed(location);
		cs.transformToWorld(location, clickLocation);
		original = new DragState(this);
				
		return true;
	} // public boolean mousePressed ()
	
	/** Checks for a mouse drag. */
	public boolean mouseDragged(Point location) 
	{	

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		//Hack. Remove the listener to the domain WHILE I change the domain properties
		segmentDomain.removeDomainListener(domainListener);
		//
		
		Point2D.Double transformedLocation = new Point2D.Double();
		cs.transformToWorld(location,
							  transformedLocation);
		
		double changeX = transformedLocation.x - clickLocation.x;	
		double changeY = transformedLocation.y - clickLocation.y;	
		
		if (moveMode == MOVE_LINE) {
			// move line horizontally
			yIntercept = original.yIntercept - original.slope * changeX;
		
			// move domain horizontally
			segmentDomain.setStart(original.start + changeX);
			segmentDomain.setEnd(original.end + changeX);
		
			// move line vertically
			yIntercept = yIntercept + changeY;
		} else if (moveMode == MOVE_START) {

			if ( (original.start + changeX) < original.end) {
			
				double m =  
					(original.slope * (original.end - original.start) - changeY)/
					( original.end - original.start - changeX);
		
				slope = m;
			
				yIntercept = 
					original.yIntercept + 
					original.end * (original.slope - slope);
				segmentDomain.setStart(original.start + changeX);
			}
		
		} else if (moveMode == MOVE_END) {
			if ( (original.end + changeX) > original.start) {
				double m =  
					(original.slope * (original.start - original.end) - changeY)/
					( original.start - original.end - changeX);
		
				if (Math.abs(m) > topSlope) return false;
			
				slope = m;
				if (getSlopeInterval()!=0){
					slope = Math.floor(slope/getSlopeInterval() + 0.5) * 
						getSlopeInterval(); // set Slope to "grid"  
				}
				yIntercept = 
					original.yIntercept + original.start * (original.slope - slope);
		
				segmentDomain.setEnd(original.end + changeX);
		
			} // if original.end + changeX > original.start
		} // if moveMode = ...
		
		fixDomain(transformedLocation);
		
		fix();
		
		//Hack. Add the listener to the domain AGAIN ( removed WHILE changing the domain properties)
		segmentDomain.addDomainListener(domainListener);
		//

		notifyChange();
		
		return true;	
	}
	
	/** Checks for a mouse release. */
	public boolean mouseReleased(Point location) 
	{
		super.mouseReleased(location);
		mouseDragged(location); // Do we really need (or want) to do this?
		return false;
	}	

	/*
	 * Creates a clone of this obect <br>
	*/
	public Graphable getCopy() 
	{
		LinearSegment retval = new LinearSegment(slope, yIntercept,
			segmentDomain.getStart(),segmentDomain.getClosedStart(),
			segmentDomain.getEnd(),segmentDomain.getClosedEnd());
		retval.setColor(getColor());
		return retval;
	}
	//////////////////////////////////////
	
	/** 
	 * Flip domains if start > end
	*/
	private void fixDomain(Point2D.Double transformedLocation)
	{
		if (segmentDomain.getStart() > segmentDomain.getEnd() + (0.1) && Math.abs(slope) < topSlope){
			double end = segmentDomain.getEnd();
			segmentDomain.setEnd(segmentDomain.getStart());
			segmentDomain.setStart(end);
			if (moveMode == MOVE_START){
				moveMode = MOVE_END;
			}
			else if (moveMode == MOVE_END){
				moveMode = MOVE_START;
			}
			original = new DragState(this);
			clickLocation = transformedLocation;
		}
	}	
	
	/** Changes the expresion of the function that draws the
	line that this segment is a part of. */
	public void setExpression(String expression) {} // throws ExpressionException {}
	
	/** Sets the Domain of this Segment as determined by start and end 
	points. */
	public void setDomain(SegmentDomain domain)
	{
		if (segmentDomain != null){
			segmentDomain.removeDomainListener(domainListener);
		}
		segmentDomain = domain;
		if (segmentDomain != null){
			segmentDomain.addDomainListener(domainListener);
		}
	} //throws DomainException{} {}
	
	public void addDomainListener(DomainListener list) 
	{
		segmentDomain.addDomainListener(list);
	}
	
	public void removeDomainListener(DomainListener list) 
	{
		segmentDomain.removeDomainListener(list);
	}
	
	public void setParent(GraphFunction parent)
	{
		this.parent = parent;
	}

	public String getExpression(int form)
	{
		String strDomain;
		String startOp;
		String endOp;
	
		if (getDomain().getClosedStart()) {
			startOp = "<font face=\"Symbol\">&#8804;</font>";
		} else {
			startOp = "&#60;";
		}
	
		if (getDomain().getClosedEnd()) {
			endOp = "<font face=\"Symbol\">&#8804;</font>";
		} else {
			endOp = "&#60;";
		}
		strDomain = "  for " + 
				MathUtil.roundTenth(getDomain().getStart()) + 
				startOp +  " x " + endOp +
				MathUtil.roundTenth(getDomain().getEnd());
				
		if (form == FORM_FUNCTIONPOINTSLOPE){
			//"Pushpin" will be the start point FOR NOW
			//If the start point is infinite, then it will be the end point
			//if the end point is infinite, it will be 0.0
			String strExpression="";
			double x0 = getDomain().getStart();
			if (Double.isInfinite(x0)){
				x0 = getDomain().getEnd();
			}
			if (Double.isInfinite(x0)){
				x0 = 0.0;
			}
			double y0 = yval(x0, false);
			
			strExpression = "<html>" + 
				MathUtil.roundTenth(slope, getSlopePrecision()) + " ( x - "+
				MathUtil.roundTenth(x0) + " ) + " +
				MathUtil.roundTenth(y0) + strDomain + "</html>";

			return strExpression;
		}
		else if (form == FORM_SLOPEINTERCEPT){
			return "<html>" + 
				MathUtil.roundTenth(slope, getSlopePrecision()) + " x + "+
				MathUtil.roundTenth(yIntercept, getYInterceptPrecision()) +
				strDomain + "</html>";
		}
		else{
			return super.getExpression(form) + strDomain;
		}
	}
	
	/**
	 * Returns if a segment is infinite or not (i.e.
	 * if the domain start OR end is infinite)
	*/
	public boolean isInfinite()
	{
		if (Double.isInfinite(getDomain().getStart()) ||
			Double.isInfinite(getDomain().getEnd())){
	
			return true;
		}
		return false;
	}

	class DragState 
	{
		double slope;
		double yIntercept;
		double start;
		double end;
		Point2D.Double startPoint;
		Point2D.Double endPoint;
	
		public DragState(LinearSegment parent) 
		{
			slope = parent.slope;
			yIntercept = parent.yIntercept;
			start = parent.getDomain().getStart();
			end = parent.getDomain().getEnd();
		
			startPoint = new Point2D.Double(start, parent.yval(start));
			endPoint = new Point2D.Double(end, parent.yval(end));
		}
		
	}
}

