/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * @(#)Piecewise.java
 * 
 * Last modification information:
 * $Revision: 1.7 $
 * $Date: 2004/07/30 18:10:02 $
 * $Author: jqian $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.util.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;
import org.concord.graph.ui.*;
import org.concord.graphfunction.event.*;

/**
 * A component that represents a symbolic definition of a piecewise function.
 * The piecewise is represented in standard form with a line for each
 * segment that specifies its domain and expression.
 * This definition is put into a panel using Ascii characters. The
 * expressions are editable. The values in the domains are editable
 * and the comparison operators toggle between "\<" and "<=" (the fancy
 * symbol) when the user clicks.
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 *
 */
 
public class Piecewise extends DefaultGraphFunction 
	implements GraphableListener, SelectableListener, DomainListener 
{	
	/** The segments to display. Each element of this Vector implements
	the Segment interface. */
	//It is now implemented with Graphable List
	SelectableList segments;
	
	/** A Vector of DomainListeners */
	Vector domainListeners;
	
	public DomainManager dManager;
	
	// These are variables used by checkFunction() methods.
	private double lastValidXval = Double.NaN;
	private boolean inBadRange = false;
	
	//Current segment clicked
	Segment currentSegmentClickAccepted;
	
	//Draw manager to draw the list of segments
	GraphDrawManager drawManager;
	
	//Mouse manager to manage the clicks on the list of segments
	DefaultGraphMouseManager mouseManager;

	/**
	 * Always call this default constructor.
	 * Initializes attributes
	 */
	public Piecewise()
	{
		segments = new SelectableList();
		//Only one segment can be selected at once
		segments.setSelectionMode(SelectableList.SELECTIONMODE_SINGLE);
		
		drawManager = new DefaultGraphDrawManager();
		drawManager.setGraphableList(segments);
		
		mouseManager = new DefaultGraphMouseManager();
		mouseManager.setSelectableList(segments);
		mouseManager.setMouseBehavior(DefaultGraphMouseManager.ONCLICK_SELECT_DESELECT);
		
		domainListeners = new Vector();
		dManager = new DomainManager(this);
	}
	
	/** Adds a listener for a Piecewise listener */
	public void addDomainListener(DomainListener listener) 
	{
		if (! domainListeners.contains(listener)) {
			domainListeners.add(listener);
		}
	}
	
	/** Adds a listener for a Piecewise listener */
	public void removeDomainListener(DomainListener listener) 
	{
		if (domainListeners.contains(listener)) {
			domainListeners.remove(listener);
		}
	}
		
	/** Adds a segment to the piecewise */
	public void add(Segment newSegment) 
	{
		if (!segments.contains(newSegment)) {
			segments.add(newSegment);
		
			newSegment.setParent(this);
			
			// Add a listener to each Segment
			newSegment.addGraphableListener(this);
			newSegment.addSelectableListener(this);
			newSegment.addDomainListener(this);
			
			//Same color on the segment
			newSegment.setColor(getColor());
			//Same graph area
			newSegment.setGraphArea(getGraphArea());
			//Same ghost
			newSegment.setShowGhost(showGhost);
			
			notifyAdd(newSegment);
			
			//Always Select the last segment that was added
			newSegment.select();			
		}
	}
	
	public void removeSelected() 
	{
		segments.removeAllSelected();
	}

	public void removeAllElements()
	{
		PiecewiseEvent e;
		segments.removeAllElements();
		e = new PiecewiseEvent (this, null, 
								PiecewiseEvent.PIECEWISE_ALL_SEGMENTS_REMOVED,
								true);
		notifyChange(e);

	}
	
	/** Removes a segment from the piecewise */
	public void remove(Segment deadSegment) 
	{
		if (segments.contains(deadSegment)){
			deadSegment.remove();
		}
	}

	public void notifyAdd(Segment newSegment) 
	{
		EventObject e = new PiecewiseEvent(this, newSegment, PiecewiseEvent.PIECEWISE_SEGMENT_ADDED, true); 		
		notifyChange(e);
        
	}
	
	public void notifyRemove(Segment goneSegment) 
	{
		EventObject e = new PiecewiseEvent(this, goneSegment, PiecewiseEvent.PIECEWISE_SEGMENT_REMOVED, true); 
		notifyChange(e);
	}
	
	public void notifyRemove(EventObject e) 
	{
		Enumeration lists = graphableListeners.elements();
		while (lists.hasMoreElements()) {
			((GraphableListener) lists.nextElement()).graphableRemoved(e);
		}
	}
	
	public void notifyChange(Segment segmentChanged) 
	{
		EventObject e = new PiecewiseEvent(this, segmentChanged, true);
		notifyChange(e);
	}
	
	public void notifyChange(EventObject e) 
	{
		Enumeration lists = graphableListeners.elements();
		
		while (lists.hasMoreElements()) {
			GraphableListener l = (GraphableListener) lists.nextElement();
			l.graphableChanged(e);
		}
	}
		
	public void notifySelect(Segment segmentChanged) 
	{
		EventObject e = new PiecewiseEvent(this, segmentChanged, PiecewiseEvent.PIECEWISE_SEGMENT_SELECTED, true);
		notifyChange(e);
	}
	
	public void notifySelect(EventObject e) 
	{
		Enumeration lists = selectableListeners.elements();
		
		while (lists.hasMoreElements()) {
			SelectableListener l = (SelectableListener)lists.nextElement();
			l.graphableSelected(e);
		}
	}

	public void notifyDeselect(Segment segmentChanged) 
	{
		EventObject e = new PiecewiseEvent(this, segmentChanged, PiecewiseEvent.PIECEWISE_SEGMENT_DESELECTED, true);
		notifyChange(e);
	}
	
	public void notifyDeselect(EventObject e) 
	{
		Enumeration lists = selectableListeners.elements();
		
		while (lists.hasMoreElements()) {
			SelectableListener l = (SelectableListener)lists.nextElement();
			l.graphableDeselected(e);
		}
	}

	public void notifyMultipleDomains() 
	{
		DomainEvent e = new DomainEvent(this);
		
		Enumeration lists = domainListeners.elements();
		
		while (lists.hasMoreElements()) {
			((DomainListener) lists.nextElement()).multipleDomains(e);
		}
	}

	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
    /** Forces classes to redraw any internal image buffer (if implemented).
	 * 	This makes sure the background is updated if there is a fundamental
	 * 	change that is not checked by this class. 
	 */
    public void update()
	{
	}

	/** Draws the piecewise (i.e. each segment in the segments vector)
	on the specified graphics area. */
	public void draw(Graphics2D g) 
	{

		GraphArea ga = getGraphArea();
		CoordinateSystem cs = ga.getCoordinateSystem();

		drawManager.draw(g);

/*		
		if (isVisible()) {
			//Draw all non-selected segments
			Enumeration segs = segments.elements();
			Segment seg;
			while (segs.hasMoreElements ()) {
				seg = ((Segment)segs.nextElement());
				if (seg != currentSegmentSelected){
					seg.draw(axis,g,isSelected());
				}
			}
			//Draw selected segment
			if (currentSegmentSelected!=null){
				currentSegmentSelected.draw(axis,g,isSelected());
			}
		}
*/

	}
	
    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
    public boolean isPointInProximity(Point location)
	{
		currentSegmentClickAccepted = (Segment)mouseManager.getFirstObjectToSelect(location);
		
		if (currentSegmentClickAccepted != null){
			return true;
		}
		return false;

/*		
		//First, look if the selected segment accepts the mouse click
		seg = currentSegmentSelected;
		if (seg != null){
			accepted = seg.isPointInProximity(axis, location);
		}
		
		if (!accepted){
			segs = segments.elements(); 
			//Loop through all the segments to see if one of them accepts the mouse click 
			while ( (accepted == false) && (segs.hasMoreElements())  ) {
				seg = ((Segment) segs.nextElement());
				accepted = seg.isPointInProximity(axis, location);
			}
		}
		
		if (accepted){
			////Select the segment!
			//seg.select();
			//Mark this segment as accepting the click
			currentSegmentClickAccepted = seg;
		}
		else{
			currentSegmentClickAccepted = null;
		}
		
		return accepted;
*/
	}
	
	/** Handles a mouse click. The Container that the Piecewise is
	in calls this function with the location of the click. Piecewise
	passes these coordinates to each Segment. If any segment "claims
	the click than this method stores a pointer to that segment. This
	method returns true if the click is on a segment. If not it returns
	false */
	public boolean mousePressed( Point location ) 
	{
		//Select the segment that accepted the click
		mouseManager.doMousePressSelection(currentSegmentClickAccepted, location);

		if (currentSegmentClickAccepted != null){
			return true;
		}

		return false;
	}
	
	/** Passes the mouseDrag location to the Segment that was clicked
	(stored by mousePressed()). */
	public boolean mouseDragged(Point location)
	{
		if (currentSegmentClickAccepted != null){
			mouseManager.doMouseDragged(location);
			return true;
		}
		return false;
	}
	
	/** Passes the mouseDrag location to the Segment that was clicked
	(stored by mousePressed()). */
	public boolean mouseReleased(Point location)
	{
		// super.mouseReleased(location);
		if (currentSegmentClickAccepted != null){
			mouseManager.doMouseReleased(location);
			return true;
		}
		return false;
	}
	//////////////////////////////////////
	
	/** Returns the number of segments in the piecewise */
	public int getNumberSegments() 
	{
		return segments.size();
	}
	
	/** Returns false if this Piecewise is an illegal function (i.e. no
	overlapping Domains). Otherwise returns true. */
	public boolean checkFunction(double xval, boolean lastVal) 
	{
		if (lastVal) {
			return checkFunction(lastValidXval, xval) ;
		} else {
			return checkFunction(Double.NaN, xval);
		}
	}
	
	public boolean checkFunction(double lastXval, double xval) 
	{
		boolean retval = true;
		boolean isBad = false;
		
		// Checks the DomainManager at xval	    
		if (Double.isNaN(lastXval)) {
			isBad = dManager.isBad(xval);
		} else {
			// This changes a range to see if you've crossed a bad spot
			isBad = dManager.isBad(lastXval, xval); 
		}
		
		if (isBad) { 
			retval = false; //return a NaN for more than one value
			notifyMultipleDomains();
		} 
		
		if (retval) lastValidXval=xval;
		return retval;
	
	}
	
	/** This works like checkFunction, but it will only return false
	one time. This is to implement the dialog box function */
	
	public boolean checkFunctionOnce(double xval) 
	{
		boolean retval = true;
		if (inBadRange) {
			retval = checkFunction(xval, false);
			if (retval) {
			inBadRange = false;
			lastValidXval = xval;
			}
		} else {
			retval = checkFunction(xval, true);
			if (!retval) {
			inBadRange = true;
			} else {
			lastValidXval = xval;
			}
		}
		
		return retval;
	}
	
	/** Returns the value of this function for  this "x" value. This will
	return null if no segment applies. NOTE: I will rewrite this
	since double is inadaquate for multiple values and nulls.
	*/
	public double yval(double xval) 
	{
		double retval = Double.NaN;
		
		if (dManager.isBad(xval)) {
			retval = Double.NaN;
		} else {
			Enumeration segs = segments.elements(); 
			while ( ( Double.isNaN(retval)) && segs.hasMoreElements() ) {
			Segment aSegment = 
				(Segment)segs.nextElement();
			retval = aSegment.yval(xval);
			}
		}
		
		return retval;
	}

	/** Returns the first segment that is defined at xval.
		FIXME This should be made cleaner with yval etc.
	*/
	public Segment getSegment(double xval) 
	{
		Segment retval = null;
		Enumeration segs = segments.elements(); 
		while ( (retval == null) && segs.hasMoreElements() ) 
			{
				Segment aSegment = 
					(Segment)segs.nextElement();
				if (! Double.isNaN(aSegment.yval(xval)))
					retval = aSegment;
			}
		return retval;

	}

	
	// Ingrid TEST
	/** This is made so you can determine the last y value for domainEnd 
	*/
	public double yval(double xval, boolean checkDomain) 
	{
		double retval = yval(xval);
		if (!checkDomain && Double.isNaN(retval)){
			if (!segments.isEmpty()){
				LinearSegment aSegment = (LinearSegment)segments.lastElement();
				retval = aSegment.yval(xval, false);
			}
			else{
				retval = 0.0;				
			}
		}
		return retval;
	}
	
	/** Returns a String that describe this Piecewise */
	public String getString() 
	{
		return "Not implemented";
	}
	
	/** Returns the Domain */
	public SegmentDomain getDomain() 
	{
		return null;
	}
	
	/** Presets the Piecewise with the String */
	public void loadString(String instring) 
	{
	}
	
    /*
	 * Gets a string representation of this object. The format is
	 * determined by the subclass of Graphical 
	 */
    public String getLoadString()
	{
		return "";
	}

	/** 
	 * Returns a String version of the expression of this function 
	 * WARNING: NOT IMPLEMENTED YET!!
	*/
	public String getExpression(int expressionForm) 
	{
		return "Piecewise with "+getNumberSegments()+" segments";
	}
	
	/** Returns an Enumeration of Segments */
	public Enumeration getSegments() 
	{
		return segments.elements();
	}
	
	public double getDomainEnd() 
	{
		double retval = 0.0;
		boolean started = false;
		
		Enumeration segs = segments.elements();
		
		while ( segs.hasMoreElements() ) {
		
			double segend = 
			((Segment)segs.nextElement()).getDomain().getEnd();
			if (!started) {
			retval = segend;
			started = true;
			} else {
			if (retval < segend) retval = segend;
			}
		}
		return retval;
	}
		
	public double getDomainStart() 
	{
		double retval = 0.0;
		boolean started = false;
		
		Enumeration segs = segments.elements();
		
		while ( segs.hasMoreElements() ) {
		
			double segend = 
			((Segment)segs.nextElement()).getDomain().getStart();
			if (!started) {
			retval = segend;
			started = true;
			} else {
			if (retval > segend) retval = segend;
			}
		}
		return retval;
	}
	
	/** Returns the currently selected Segment of the Piecewise */
	public Segment getCurrentSegmentSelected() 
	{
		return (Segment)segments.getFirstSelectedObject();
	}
	
	/** Returns the index of currently selected Segment of the Piecewise */
	public int getCurrentSegmentSelectedIndex() 
	{
		return segments.getFirstSelectedIndex();
	}
	
	/*
	 * Creates a clone of this obect <br>
	*/
	public Graphable getCopy() 
	{
		Piecewise retval = new Piecewise();
		for (int i=0;i<segments.size();i++){
			retval.add((Segment)(((GraphFunction)segments.elementAt(i)).getCopy()));
		}
		return retval;
	}
	
	/*
	 * Sets the color of the graph 
	*/
	public void setColor(Color c)
	{
		super.setColor(c);
		//Changes the color of all segments
		int i;
		for (i=0;i<segments.size();i++){
			((GraphFunction)segments.elementAt(i)).setColor(getColor());
		}
	}
	
	// These methods implement the GraphableListener interface
	
	public void graphableChanged(EventObject e) 
	{
		notifyChange((Segment)e.getSource());
	}
	
	public void graphableRemoved(EventObject e) 
	{		
		Segment toRemove = (Segment)e.getSource();

		toRemove.removeGraphableListener(this);
		toRemove.removeSelectableListener(this);
		toRemove.removeDomainListener(this);

		notifyRemove(toRemove);
	}
	
	// These methods implement the SelectableListener interface
	
	public void graphableSelected(EventObject e) 
	{
		notifySelect((Segment)e.getSource());
	}
	
	public void graphableDeselected(EventObject e) 
	{
		notifyDeselect((Segment)e.getSource());
	}

	// These methods implement the DomainListener interface
	public void domainChanged(DomainEvent e) 
	{
		Enumeration doms = domainListeners.elements();
		while (doms.hasMoreElements()) {
			((DomainListener)doms.nextElement()).domainChanged(e);
		}
	}
	
	public void multipleDomains(DomainEvent e) {}
	public void noDomain(DomainEvent e) {}

	/**
	 * Indicates if the function will show a ghost when dragged
	*/
	public void setShowGhost(boolean showGhost)
	{
		super.setShowGhost(showGhost);
		//Changes show ghost of all segments
		int i;
		for (i=0;i<segments.size();i++){
			((GraphFunction)segments.elementAt(i)).setShowGhost(showGhost);
		}
	}

	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		super.setGraphArea(area);
		//Changes graph area of all segments
		int i;
		for (i=0;i<segments.size();i++){
			((GraphFunction)segments.elementAt(i)).setGraphArea(area);
		}
	}
	
} // public class Piecewise
