/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.4 $
 * $Date: 2004/07/14 17:12:13 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;

import org.concord.graph.engine.*;

/**
 * A quadratic function a^x2 + bx + c
 *
 * Date created: 04/07/2004
 *
 * @author Ingrid C Moncada<p>
 * @author Eric Brown-Munoz<p>
 *
 */

public class QuadraticFunction extends DefaultGraphFunction 
{
	//Expression form constants	
	public static final int FORM_POLYNOM = 10;
	public static final int FORM_ROOT = 11;
	public static final int FORM_VERTEX = 12;

    /* The three coefficients that define the Quadratic
       i.e. Ax^2+Bx+C
    */
    private double coeffA;
    private double coeffB;
    private double coeffC;
	
	private double vertexX;
	
	private double root1;
	private double root2;

	private int coeffPrecision = 10;
	private double coeffInterval = 0.1;

	//Temp
	/** A saved verson of the QuadraticFunction for use with drags */
	protected QuadraticFunction savedQG;

	//Note about transformations:
	//Reflection is done in methods reflectX() and reflectY()
	//We should do methods for translations and dilations 
	// for example translateX(float howMuch) , etc
	// translate methods should use setVertex() -> move the vertex

	//This is to draw the bezier curve
	protected Shape curve;
	protected boolean needCurveUpdate = true;

	private Point2D.Double clickLocation;

	/**
	 * Default constructor. 
	 * By default it creates
	 * y = x^2 black function, in the polynom form
	 */
	public QuadraticFunction()
	{
		//By default it's a y=x^2 parabola
		this(1.0,0.0,0.0);
	}
	
	/**
	 * Constructor that creates. 
	 * y = a x^2 + b x + c
	 * black function, in the polynom form
	 */
	public QuadraticFunction(double a, double b, double c)
	{
		this(a, b, c, Color.black, FORM_POLYNOM);
	}

	/**
	 * Constructor that creates. 
	 * y = a x^2 + b x + c
	 * with the specified color and expression form
	 */
	public QuadraticFunction(double a, double b, double c, Color color, int expressionForm)
	{
		super(color,expressionForm);
		setCoeffs(a, b, c);
		//Default drag mode is translate?
		dragMode = new DragMode(DragMode.TRANSLATE);
		clickLocation = new Point2D.Double();
	}

	/**
	 * Returns the a coefficient in the quadratic function ax^2 + bx + c
	 */
	public double getCoeffA()
	{
		return coeffA;
	}
	
	/**
	 * Returns the a coefficient in the quadratic function ax^2 + bx + c
	 */
	public double getCoeffB()
	{
		return coeffB;
	}
	
	/**
	 * Returns the a coefficient in the quadratic function ax^2 + bx + c
	 */
	public double getCoeffC()
	{
		return coeffC;
	}
	
	/**
	 * Changes the coefficients in the quadratic function ax^2 + bx + c
	 * This fires the graphable changed event
	 */
	public void setCoeffs(double a, double b, double c)
	{
		coeffA = a;
		coeffB = b;
		coeffC = c;
		calculateVertex();
		calculateRoots();
		needCurveUpdate = true;
		notifyChange();
	}
	
	/**
	 * Changes the a coefficient in the quadratic function ax^2 + bx + c
	 * This fires the graphable changed event
	 */
	public void setCoeffA(double a)
	{
		coeffA = a;
		calculateVertex();
		calculateRoots();
		needCurveUpdate = true;
		notifyChange();
	}
	
	/**
	 * Changes the b coefficient in the quadratic function ax^2 + bx + c
	 * This fires the graphable changed event
	 */
	public void setCoeffB(double b)
	{
		coeffB = b;
		calculateVertex();
		calculateRoots();
		needCurveUpdate = true;
		notifyChange();
	}

	/**
	 * Changes the c coefficient in the quadratic function ax^2 + bx + c
	 * This fires the graphable changed event
	 */
	public void setCoeffC(double c)
	{
		coeffC = c;
		//The vertex (x coordinate) doesnt depend on c
		//calculateVertex();
		calculateRoots();
		needCurveUpdate = true;
		notifyChange();
	}
	

    /*
	 * Reflects the function around the x axis ( - f(x) ) 
	 */
    public void reflectX() 
	{
		setCoeffs(-coeffA,-coeffB,-coeffC);
	}

    /*
	 * Reflects the function around the y axis ( f( -x ) ) 
	 */
    public void reflectY() 
	{
		setCoeffB(-coeffB);
	}

	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
    /** Forces classes to redraw any internal image buffer (if implemented).
	 * 	This makes sure the background is updated if there is a fundamental
	 * 	change that is not checked by this class. 
	 */
    public void update()
	{
		needCurveUpdate = true;
	}
	
    /** Draws this object on Graphics g */
    public void draw(Graphics2D g)
	{
		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		
		//super.draw(g);
		
		//Determine if it is quadratic or not
		
		if (curve==null){
			//curve = new QuadCurve2D.Double();
			needCurveUpdate = true;
		}
		
//For now: it always updates the curve. 
//This is because the axis scale could have been changed
		if (needCurveUpdate){
			updateCurve(g);
			needCurveUpdate = false;
		}

		if (isSelected()){
			g.setStroke(getSelectedLineStroke());
		}
		else{
			g.setStroke(getLineStroke());
		}
		g.setColor(getColor());
		
		if (curve != null){
			g.draw(curve);
		}
		
		g.setColor(oldColor);
		g.setStroke(oldStroke);
	}
	
    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
	// NOT IMPLEMENTED YET
    public boolean isPointInProximity(Point location)
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		Point2D coordScale = cs.getScale();

		boolean retval;
		Point2D.Double tLocation = new Point2D.Double();

		double deltaX = 3.0 / coordScale.getX();
		double deltaY = 3.0 / coordScale.getY();

		cs.transformToDisplay(location, tLocation);

		double leftSideYval = yval(tLocation.x - deltaX);
		double rightSideYval = yval (tLocation.x + deltaX);

		/* Check to see if the curve passes through a square
		   with side length of 7 pixels  centered on click location */

		if (((leftSideYval  < (tLocation.y - deltaY)) && 
			 (rightSideYval < (tLocation.y - deltaY))) ||
			((leftSideYval  > (tLocation.y + deltaY)) && 
			 (rightSideYval > (tLocation.y + deltaY)))) {
			
			retval = false;
			// System.err.println("Click false");
		} 
		else {
			
			retval = true;
			// System.err.println("Click true");
		}
		
		return retval;
	}

    /** Checks for a mouse click. If the location is owned by this object and
	the object wants to handle it, return true. Else return false */
    public boolean mousePressed(Point location)
	{
		//super.mousePressed(location);

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();
		
		cs.transformToDisplay(location, clickLocation); // store the click

		savedQG = (QuadraticFunction)getCopy(); // Store a copy of self for use with drags
		
		return true;
	}

    /** Checks for a mouse drag. */
    public boolean mouseDragged(Point location)
	//If we want to snap the drag factor to precision of 0.1:
	//dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		boolean retval = true;
		double dragFactor = 0.0; // Amount that Curve was dragged

		Point2D.Double tLocation = new Point2D.Double();
		cs.transformToWorld(location,tLocation); //mouse click in world coords

		if (dragMode.getValue() == DragMode.TRANSLATE_X) {
			dragFactor = tLocation.x - clickLocation.x;
			//Snap the drag factor to precision of 0.1:
			dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
			//
			setVertexX(savedQG.getVertexX() + dragFactor );
		} 
		else if (dragMode.getValue() == DragMode.TRANSLATE_Y) {
			dragFactor = tLocation.y - clickLocation.y;
			//Snap the drag factor to precision of 0.1:
			dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
			//
			setVertexY(savedQG.getVertexY() + tLocation.y - clickLocation.y);
		} 
		if (dragMode.getValue() == DragMode.TRANSLATE) {
			dragFactor = tLocation.x - clickLocation.x;
			//Snap the drag factor to precision of 0.1:
			dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
			//
			setVertexX(savedQG.getVertexX() + dragFactor );
			dragFactor = tLocation.y - clickLocation.y;
			//Snap the drag factor to precision of 0.1:
			dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
			//
			setVertexY(savedQG.getVertexY() + tLocation.y - clickLocation.y);
		}
		else if (dragMode.getValue() == DragMode.DILATE_X) {

			//Only do the dilation if the mouse is at least two pixels away from the y axis
			//This avoids making the dragFactor Infinite and screwing up the parabola
			
			Point2D origin = cs.getOriginOffsetDisplay();
			if (Math.abs(origin.getX() - location.getX()) > 1.0){
			
				dragFactor = clickLocation.x / tLocation.x;
				//Snap the drag factor to precision of 0.1:
				dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
				//
	
				setCoeffs(savedQG.getCoeffA() * dragFactor * dragFactor,
					savedQG.getCoeffB() * dragFactor,
					getCoeffC());
			}

		} 
		else if (dragMode.getValue() == DragMode.DILATE_Y) {
			dragFactor = tLocation.y / clickLocation.y ;
			//Snap the drag factor to precision of 0.1:
			dragFactor = MathUtil.roundTenth(dragFactor, coeffPrecision);
			//
			setCoeffs(savedQG.getCoeffA() * dragFactor,
				savedQG.getCoeffB() * dragFactor,
				savedQG.getCoeffC() * dragFactor);
		}
			
		return retval;
	}

    /** Checks for a mouse release. */
    public boolean mouseReleased(Point location) 
	{
		//super.mouseReleased(location);
		
		return true;
	}

    private void updateCurve(Graphics2D g)
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		//The three points of the bezier curve (in world coordinates!)
		double x0,y0,x1,y1,x2,y2;
		double a,b,c;

		a = coeffA;
		b = coeffB;
		c = coeffC;
		
		//Determine the points x0 and x1
		Rectangle clip = g.getClipBounds();
		Point2D startPoint = new Point2D.Double(clip.getX(),0);
		Point2D endPoint = new Point2D.Double(clip.getX() + clip.getWidth(),0);

		Point2D.Double pStart = new Point2D.Double();
		Point2D.Double pEnd = new Point2D.Double();
		cs.transformToWorld(startPoint,pStart);
		cs.transformToWorld(endPoint,pEnd);
		//for now
		
		x0 = pStart.getX();
		x1 = pEnd.getX();
		
		if (!MathUtil.equalsDouble(a,0.0)){
		//parabola!
			//Calculate y0 and y1
			y0 = yval(x0);
			y1 = yval(x1);
			
			//NO ---x2 is in the middle of x0 and x1
			//NO x2 = (x1 + x0)/2;
			
			//Determine x2
			//It is the point where the two tangents to x0 and x1 cross
			x2 = ((2 * a*( (x0*x0)-(x1*x1) )) + b * (x0-x1) - (y0-y1) )/ (2*a* (x0-x1) );
			//Calculate y2
			y2 = (a * x1 * x0) + ((b * x0)/2) + y1 - (a * x1 * x1) - ((b * x1)/2);
	
//			System.out.println("("+x0+","+y0+")");
//			System.out.println("("+x1+","+y1+")");
//			System.out.println("("+x2+","+y2+")");
	
			//This should be the same
			//y2 = (2 * a * x1 * x2) + (b * x2) + y1 - (2 * a * x1 * x1) - (b * x1);
			//System.out.println("("+x2+","+y2+")");
	
			//The points are in world coordinates!

			curve = createCurve(x0,y0,x1,y1,x2,y2);
		}
		else{
		//Draw a line instead
			y0 = (b * x0) + c;
			y1 = (b * x1) + c;
			
			Point2D p0 = new Point2D.Double(x0,y0);
			Point2D p1 = new Point2D.Double(x1,y1);

			Point2D pp0 = new Point2D.Double();
			Point2D pp1 = new Point2D.Double();

			cs.transformToDisplay(p0, pp0);
			cs.transformToDisplay(p1, pp1);
			
			curve = new Line2D.Double(pp0,pp1);
		}
	}

	public Shape createCurve(double x0, double y0, double x1, double y1, double x2, double y2)
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		Point2D p0 = new Point2D.Double(x0,y0);
		Point2D p1 = new Point2D.Double(x1,y1);
		Point2D p2 = new Point2D.Double(x2,y2);
		
		Point2D pp0 = new Point2D.Double();
		Point2D pp1 = new Point2D.Double();
		Point2D pp2 = new Point2D.Double();

		cs.transformToDisplay(p0, pp0);
		cs.transformToDisplay(p1, pp1);
		cs.transformToDisplay(p2, pp2);

		//Added checking for big integers (>1000000)
		//when antialiasing is on it causes an exception: 
		//sun.dc.pr.PRException: endPath: bad path 
		//(java bug http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4696289)
		if (pp0.getY() >= 1000000 || 
			pp1.getY() >= 1000000 || 
			pp2.getY() >= 1000000){
				
			return null;
		}
		//
		
		return new QuadCurve2D.Double(	(int)pp0.getX(), (int)pp0.getY(),
										(int)pp2.getX(), (int)pp2.getY(),
										(int)pp1.getX(), (int)pp1.getY() );
	}

	/** Returns a copy of itself */
    public Graphable getCopy() 
	{
		QuadraticFunction q = new QuadraticFunction(coeffA,coeffB,coeffC,getColor(),getExpressionForm());
		q.setFunctionName(getFunctionName());
		q.setFunctionOrdinal(getFunctionOrdinal());
//FIXME
//		q.setDragMode(this.getDragMode());
		return q;
	}
	//////////////////////////////////////

    /** Returns the value of the function represented by the Graphical
	at 'xval' */
    public double yval( double xval)
	{
		return (coeffA * xval * xval) + (coeffB * xval) + coeffC;
	}

    /*
	 * Creates a Graphical object using the string. The format of the
	 * string is determined by the subclass of Graphical. 
	 */
    public void loadString(String instring)
	{
	}

    /*
	 * Gets a string representation of this object. The format is
	 * determined by the subclass of Graphical 
	 */
    public String getLoadString()
	{
		return "";
	}
	
    /** 
	 * Returns a string representation of this graphable
	 * depending on the form specified
	 */
	public String getExpression(int form)
	{
		if (form == FORM_POLYNOM){
			return "y = " + 
				MathUtil.roundTenth(coeffA,getCoeffPrecision())  + " x^2 + "+
				MathUtil.roundTenth(coeffB,getCoeffPrecision()) + " x + "+
				MathUtil.roundTenth(coeffC,getCoeffPrecision());
		}
		else if (form == FORM_VERTEX){
			if (hasVertex()){
				return "y = " + 
					MathUtil.roundTenth(coeffA,getCoeffPrecision())  + " (x - "+
					MathUtil.roundTenth(getVertexX(),getCoeffPrecision()) + " )^2 + "+
					MathUtil.roundTenth(getVertexY(),getCoeffPrecision());
			}
			else{
				return "Vertex form not valid";
			}
		}
		else if (form == FORM_ROOT){
			if (hasRoots()){
				return "y = " + 
					MathUtil.roundTenth(coeffA,getCoeffPrecision())  + " (x - "+
					MathUtil.roundTenth(getRoot1(),getCoeffPrecision()) + " )(x - "+
					MathUtil.roundTenth(getRoot2(),getCoeffPrecision()) + " )";
			}
			else{
				return "Root form not valid";
			}
		}
		return "";
	}
	
	/**
	 * Sets the default form expression for this graphable
	*/
    public void setExpressionForm(int mode)
	{
		if (getExpressionForm() == mode) return;
		if (isExpressionFormSupported(mode)){
			super.setExpressionForm(mode);
		}
		notifyChange(false);
	}

	private boolean isExpressionFormSupported(int mode)
	{
		if (mode == FORM_POLYNOM || mode == FORM_ROOT || mode == FORM_VERTEX){
			return true;
		}
		return false;
	}

	public double getCoeffInterval()
	{
		return coeffInterval;
	}
	
	public void setCoeffInterval(double i)
	{
		if (i<0.0) return;
		coeffInterval = i;
		if (coeffInterval > 0 && coeffInterval < 1){
			coeffPrecision = Math.abs(MathUtil.log10(coeffInterval));
			coeffPrecision = (int)Math.pow(10,coeffPrecision);
		}
		else{
			coeffPrecision = 1;
		}
	}

	public int getCoeffPrecision()
	{
		return coeffPrecision;
	}

	public void setVertex(double h, double k)
	{
		//Special case, if a=0 && b=0, then remember the h
		if (MathUtil.equalsDouble(coeffA,0.0) && MathUtil.equalsDouble(coeffB,0.0)){
			vertexX = h;
		}
		adjustCoeffsVertexForm(coeffA, h, k);
	}
	
	public void setCoeffAVertex(double m)
	{
		adjustCoeffsVertexForm(m, getVertexX(), getVertexY());
	}

	// ax^2 + bx + c
	// When is in the form m(x-h)^2 + k means that (h,k) is the vertex
	// To transform to polynomial form:
	// a = m
	// b = - 2mh
	// c= mh^2 + k
	private void adjustCoeffsVertexForm(double m, double h, double k)
	{
		setCoeffs(m, (-2*m*h), ((m*h*h) + k));
	}
	
	public void setVertexX(double h)
	{
		setVertex(h, getVertexY());
	}
	
	public void setVertexY(double k)
	{
		setVertex(getVertexX(), k);
	}

	private void calculateVertex()
	// h = -b / 2a
	//Special case when a=0, then it remembers last vertex x (doesn't change it)
	{
		if (MathUtil.equalsDouble(coeffA,0.0)) return;
		vertexX = -coeffB / (2* coeffA);
	}
	
	public double getVertexX()
	// h = -b / 2a
	// but in this case, it returns vertexX that should be the same
	{
		return vertexX;
	}
	
	public double getVertexY()
	// k = (- b^2 / 4a) + c
	// but in this case, it returns yval(vertexX) that should be the same
	{
		return yval(vertexX);
	}

	public boolean hasVertex()
	//if a = 0 and b!=0 returns false
	//otherwise it returns true
	{
		if (MathUtil.equalsDouble(coeffA,0.0) && !MathUtil.equalsDouble(coeffB,0.0)) return false;
		return true;
	}

	public void setRoots(double r1, double r2)
	{
		adjustCoeffsRootForm(coeffA, r1, r2);
	}
	
	public void setCoeffARoot(double d)
	{
		adjustCoeffsRootForm(d, getRoot1(), getRoot2());
	}
	
	// ax^2 + bx + c
	// When is in the form d(x-r1)(x-r2) means that r1 and r2 are the roots
	// To transform to polynomial form:
	// a = d
	// b = - d (r1+r2)
	// c = d * r1 * r2
	private void adjustCoeffsRootForm(double d, double r1, double r2)
	{
		setCoeffs(d, (-d * (r1 + r2) ), (d * r1 * r2) );
	}
	
	public void setRoot1(double r1)
	{
		setRoots(r1, getRoot2());
	}
	
	public void setRoot2(double r2)
	{
		setRoots(getRoot1(), r2);
	}

	private void calculateRoots()
	// r1 = (-b - sqrt(b^2 -4ac) ) / 2a
	// r2 = (-b + sqrt(b^2 -4ac) ) / 2a
	//Special case when a=0, then it remembers last roots (doesn't change them)
	{
		double sqRoot;
		if (MathUtil.equalsDouble(coeffA,0.0)) return;
		sqRoot = Math.sqrt( (coeffB*coeffB) - (4*coeffA*coeffC) );
		if (MathUtil.equalsDouble(sqRoot,0.0)) sqRoot = 0.0;
		if (Double.isNaN(sqRoot)) {
			root1 = Double.NaN;
			root2 = Double.NaN;
			return;
		}
		if (MathUtil.equalsDouble(sqRoot,coeffB)) sqRoot = coeffB;
		root1 = (-coeffB - sqRoot ) / (2*coeffA);
		root2 = (-coeffB + sqRoot ) / (2*coeffA);
	}

	public double getRoot1()
	// r1 = (-b - sqrt(b^2 -4ac) ) / 2a
	//But in this case it returns root1, that should be the same
	{
		return root1;
	}
	
	public double getRoot2()
	// r2 = (-b + sqrt(b^2 -4ac) ) / 2a
	//But in this case it returns root2, that should be the same
	{
		return root2;
	}
	
	public boolean hasRoots()
	//For now it does calculations all the time! FIXME
	//Special case, if a=0, b=0, c=0, returns true
	//if a=0 returns false
	//if b^2 - 4ac is negative returns false
	//otherwise it returns true
	{
		double sqRoot;
		if (MathUtil.equalsDouble(coeffA,0.0) && 
			MathUtil.equalsDouble(coeffB,0.0) && 
			MathUtil.equalsDouble(coeffC,0.0)) return true;
		if (MathUtil.equalsDouble(coeffA,0.0)) return false;
		sqRoot = (coeffB*coeffB) - (4*coeffA*coeffC);
		if (sqRoot < 0.0) return false;
		return true;
	}
	
}


