/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.5 $
 * $Date: 2004/07/14 17:12:13 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.event.*;

import javax.swing.Timer;

import org.concord.graph.engine.*;

import org.concord.graphfunction.event.*;

/**
 * GraphFunctionGhost
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid Moncada<p>
*/

public class SingleGhost
	implements GraphFunctionGhost, GraphFunctionView, ActionListener
{
	protected GraphFunction function;
	protected GraphFunction original;

	protected Color ghostColor = null;

	protected Stroke ghostStroke;

	private GraphArea graphArea;

	protected int ghostDuration = 0;	//Time that the ghost will remain active after deactivated ( deactivate() )
										//If the ghost duration is -1, it will NOT disappear after deactivated
										//If the ghost duration is 0, it will disappear immediately when deactivated

	protected boolean active = false;
	
	//protected Vector listeners;
	
	protected Timer timer;
	
	public SingleGhost()
	{
		active = false;
	
		timer = new javax.swing.Timer(ghostDuration, this);
		timer.setRepeats(false);
		
		float dash[] = {10.0f};
		setStroke(new BasicStroke(3.0f, 
			BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 5.0f, dash, 0.0f));
	}

	public SingleGhost(GraphFunction f)
	{
		this();
		setGraphFunction(f);
	}

	public void setGraphFunction(GraphFunction f)
	{
		this.original = f;
		this.function = null;
	}
	
	/** Draws this object on Graphics g */
	public void draw(Graphics2D g)
	{
		if (!active) return;
		if (function == null) return;


		
		function.draw(g);
	}
	
	/**
	 * Activates (shows) the ghost
	 */
	public void activate()
	{
		active = true;
		function = (GraphFunction)original.getCopy();
		function.setGraphArea(graphArea);
		
		if (function!=null && ghostColor!=null){
			function.setColor(ghostColor);
		}

		updateStroke();
	}

	/**
	 * Activates (shows) the ghost for a specific duration
	 * After the time has elapsed, it will automatically deactivate
	 * @param duration	The duration in milliseconds. 
	 * 					If it is < 0, then it will not deactivate
	 */
	public void activate(int duration)
	{
		active = true;
		function = (GraphFunction)original.getCopy();
		function.setGraphArea(graphArea);

		if (function!=null && ghostColor!=null){
			function.setColor(ghostColor);
		}

		updateStroke();

		startTimer(duration);
	}

	/**
	 * Starts the timer to schedule the deactivation
	 */
	private void startTimer(int duration)
	{
		if (duration>=0){
			timer.setDelay(duration);
			timer.setInitialDelay(duration);
			timer.start();
		}
	}
	
	/**
	 * Deactivates (hides) the ghost
	 */
	public void deactivate()
	{
		timer.stop();
		if (ghostDuration>=0){
			startTimer(ghostDuration);
		}
		else{
			//If the duration is -1, it doesn't deactivate!
			//deactivateInternal();
		}
	}

	private void deactivateInternal()
	{
		active = false;
		timer.stop();
	}
	
	/**
	 * Sets the color of the ghost
	 * By default, the color is the same as the original function
	 * @param	color	The color of the ghost. If null, the ghost will have
	 *					the same color as the original function
	 */
	public void setColor(Color color)
	{
		ghostColor = color;
		if (function == null) return;
		
		if (ghostColor!=null){
			function.setColor(ghostColor);
		}
		else{
			function.setColor(original.getColor());
		}
	}

	public void setStroke(Stroke s)
	{
		ghostStroke = s;
		
		updateStroke();
	}
	
	public void updateStroke()
	{
		if (function == null || ghostStroke == null) return;
		
		function.setSelectedLineStroke(ghostStroke);
		function.setLineStroke(ghostStroke);
	}
	
	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		this.graphArea = area;
	}

	/** 
	 * Returns the graph area where the graphable will be drawn 
	 */
	public GraphArea getGraphArea()
	{
		return this.graphArea;
	}

	/**
	 * Sets the time that the ghost will remain active after deactivated ( deactivate() )
	 * @param durationAfterDeactivate	If the ghost duration is -1, it will 
	 *									NOT disappear after deactivated
	 * 									If the ghost duration is 0, it will disappear 
	 *									immediately when deactivated
	 */
	public void setDuration(int durationAfterDeactivate)
	{
		if (durationAfterDeactivate>=0){
			ghostDuration = durationAfterDeactivate;
		}
		else{
			ghostDuration = -1;
		}
	}

    public void actionPerformed(ActionEvent e) 
	{
		deactivateInternal();
		//Hack?
		if (original instanceof DefaultGraphFunction){
			((DefaultGraphFunction)original).notifyChange();
		}
	}
}
