/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.3 $
 * $Date: 2004/07/27 23:04:53 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.engine;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;
import org.concord.graphfunction.event.*;

/**
 * A class to manage and draw traces on Graphable objects. This
 * class can be registered in a Graphwindow. You need to add
 * a Value object and a Graphable object for this class to do
 * anything.
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid Moncada<p>
 *
 */
public class Trace 
	implements Graphable, ValueListener, GraphableListener, GraphFunctionView
{
	protected Value xvalue;
	
	protected GraphFunction subject;
	
	//protected GraphFunctionList functionList;
	protected SelectableList functionList;
	
	protected Vector listeners = new Vector();
	
	protected Stroke traceStroke = null;
	protected Color traceColor = Color.black;
	
	// To draw coordinates (x,y);
	private boolean coordinates = true;
	private boolean showProjection = false;
	
	protected Font coordFont = new Font("SansSerif",Font.PLAIN,12);
	
	/** Is the trace Visible? */
	private boolean visible = true;

	private int numDecimalsX = 1;
	private int numDecimalsY = 1;

	protected GraphArea graphArea;

	//Temp
	protected Point2D.Double tracePoint;
	
	/** Constructs a Trace Object */
	public Trace() 
	{
	}
	
	/** Constructs a Trace Object */
	public Trace(Value xvalue) 
	{
		this();
		setXValue(xvalue);
	}

	/** Constructs a Trace Object initialized with the provided
	Graphable and Value objects */
	public Trace(GraphFunction subject, Value xvalue) 
	{
		this();
		setGraphFunction(subject);
		setXValue(xvalue);
	}
	
	/** Constructs a Trace Object initialized with the provided
	Graphable and Value objects */
	public Trace(SelectableList list, Value xvalue) 
	{
		this();
		setGraphFunctionList(list);
		setXValue(xvalue);
	}

	/** Sets the Value object that will manage the x value. This sets a
		listener in the Value to handle value changes (and redraw).
	*/
	public void setXValue(Value xvalue) 
	{
		this.xvalue = xvalue;
		xvalue.addValueListener(this);
	}
	
	/** Sets the Graphable object that will manage the x value. This sets
	a listener to handle Graphable Changes (and redraw).
	*/
	public void setGraphFunction(GraphFunction subject) 
	{
		if (this.subject!=null){
			this.subject.removeGraphableListener(this);
		}
		this.subject = subject;
		
		if (this.subject!=null){
			this.subject.addGraphableListener(this);
		}
	}
	
	/** Sets the Graphable object that will manage the x value. This sets
	a listener to handle Graphable Changes (and redraw).
	*/
	public void setGraphFunctionList(SelectableList list)
	{
		functionList = list;
	}

	//////////////////////////////////////
	// Methods from Graphable interface
	//////////////////////////////////////
    /** Forces classes to redraw any internal image buffer (if implemented).
	 * 	This makes sure the background is updated if there is a fundamental
	 * 	change that is not checked by this class. 
	 */
    public void update()
	{
	}

	/** Draws a trace on each appropriate Graphable on Graphics context g,
	with positions transformed by Axis axis */
	public void draw(Graphics2D g) 
	{
		Stroke savedStroke = null;
		
		Color savedColor = g.getColor();
		savedStroke = g.getStroke();
		
		g.setColor(traceColor);
		
		if (traceStroke != null) {
			g.setStroke(traceStroke);
		}
		
		// check if trace is visible and graphable and value are set
		if ((!visible) || (subject == null && functionList==null) || (xvalue == null)) {
			return;
		}
		
		if (functionList != null){
			Enumeration graphables;

			// Loop through the non-selected elements first
			graphables = functionList.elements();
			while (graphables.hasMoreElements()) {
				GraphFunction graph = (GraphFunction)graphables.nextElement();
				//Only draw the non-selected graphs
				if (!graph.isSelected()){
					drawGraph(g, graph);
				}
			} // while graphables.hasMoreElements()

			// Then loop through the selected elements
			graphables = functionList.getSelectedObjects().elements();
			while (graphables.hasMoreElements ()) {
				GraphFunction graph = (GraphFunction)graphables.nextElement();
				drawGraph(g, graph);
			} // while graphables.hasMoreElements()
		}
		
		
		//In case this is a trace for only one subject
		if (subject != null){
			drawGraph(g,subject);
		}
		
		g.setColor(savedColor);
		g.setStroke(savedStroke);
	
	} // draw()
		
	/* Draw one Graphable. Called by draw() method. */
	protected void drawGraph(Graphics2D g, GraphFunction graph) 
	{

		CoordinateSystem cs = getGraphArea().getCoordinateSystem();

		tracePoint = null;
		
		if ( graph.isVisible() ) {
			// A spot for the tracePoint in screen coords
			tracePoint = new Point2D.Double();
			
			double value = xvalue.getValue();
			double yval = graph.yval(value);
				
			if (! Double.isNaN(yval)) {
				cs.transformToDisplay ( new Point2D.Double( value, yval),
										tracePoint);
	
				g.drawOval((int)(tracePoint.x)-5, 
					   (int)(tracePoint.y)-5,
					   10,10);
				g.drawLine((int)(tracePoint.x)+3,(int)(tracePoint.y)-3,
						   (int)tracePoint.x, (int)tracePoint.y);
			
				if (coordinates) {
			
					Font oldFont = g.getFont();
					int xPos, yPos;
					
					g.setFont(coordFont);
			
					//Draw the trace of selected graphs in BOLD
					if (graph.isSelected()){
						g.setFont(g.getFont().deriveFont(Font.BOLD, g.getFont().getSize()));
					}
					
					String expression;
					//Show x with max 1 decimal, and y with max 2 decimals
					expression = "(" + MathUtil.getStringRound(value, numDecimalsX) + "," +
						MathUtil.getStringRound(yval, numDecimalsY) + ")";
					
					xPos = (int)(tracePoint.x) + 15;
					yPos = (int)(tracePoint.y) + 5;
					
					//Draw a white box under the coordinates
					FontMetrics tempFontM = g.getFontMetrics();
					int boxW = tempFontM.stringWidth(expression);
					int boxH = tempFontM.getHeight();
					g.setColor(Color.white);
					g.fillRect(xPos,yPos-(boxH*3)/4,boxW,boxH);
					//
					
					g.setColor(traceColor);
					
					g.drawString(expression, xPos, yPos);
					
					g.setFont(oldFont);
				}
		
			} // !Double.isNaN(yval)			
		} // if (graph.getVisible)
	} // drawGraph()      

	/**
	 * Sets if the trace is visible or not
	*/
	public void setVisible(boolean visible) 
	{
		this.visible = visible;
	}
	
	/**
	 * Returns if the trace is visible or not
	*/
	public boolean isVisible() 
	{
		return visible;
	}

	/** Removes the object from any enclosing view. */
	public void remove()
	{
		notifyRemove();
	}
	
	public void addGraphableListener(GraphableListener listener) 
	{
		if (! listeners.contains(listener)) {
			listeners.add(listener);
		}
	}
	
	public void removeGraphableListener(GraphableListener listener) 
	{
		if (listeners.contains(listener)) {
			listeners.remove(listener);
		}
	}
	
	public void notifyChange() 
	{
		Enumeration lists = listeners.elements();
		
		EventObject e = new EventObject(this);
		
		while (lists.hasMoreElements()) {
			((GraphableListener)lists.nextElement()).graphableChanged(e);
		}
	}

	protected void notifyRemove() 
	{
		EventObject e = 
			new EventObject(this);
		Enumeration lists = listeners.elements();
		while (lists.hasMoreElements()) {
			GraphableListener gl = (GraphableListener)lists.nextElement();
			gl.graphableRemoved(e);
		}
	}

	/** Returns a copy of itself */
	//NOT IMPLEMENTED YET
	public Graphable getCopy()
	{
		return new Trace();
	}
	//////////////////////////////////////	
	
	/**
	 * Sets if the coordinates are visible or not
	*/
	public void setShowCoordinates(boolean b) 
	{
		this.coordinates = b;
	}
	
	/**
	 * Returns if the coordinates are visible or not
	*/
	public boolean getShowCoordinates() 
	{
		return this.coordinates;
	}
	
	public int getNumDecimalsXCoordinate()
	{
		return numDecimalsX;
	}
	
	public int getNumDecimalsYCoordinate()
	{
		return numDecimalsY;
	}
	
	public void setNumDecimalsXCoordinate(int n)
	{
		if (n>=0){
			numDecimalsX = n;
		}
	}
	
	public void setNumDecimalsYCoordinate(int n)
	{
		if (n>=0){
			numDecimalsY = n;
		}
	}
	
	/* Listeners for ValueListener and GraphableListener interfaces */
	
	public void valueChanged(ValueEvent e)
	{
		notifyChange();
	}
	
	public void graphableChanged(EventObject e)
	{
		notifyChange();
	}
	public void graphableRemoved(EventObject e)
	{
	}
	
	/** 
	 * Sets the graph area where the graphable will be drawn 
	 */
	public void setGraphArea(GraphArea area)
	{
		this.graphArea = area;
	}

	/** 
	 * Returns the graph area where the graphable will be drawn 
	 */
	public GraphArea getGraphArea()
	{
		return this.graphArea;
	}
	
}
