/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.2 $
 * $Date: 2004/07/14 17:12:13 $
 * $Author: ebrown-munoz $
 *
 * Copyright 
*/
package org.concord.graphfunction.engine;

import java.awt.*;

import org.concord.graph.engine.*;

/**
 * Displays and manages a Transformed graphable. This was
 * designed for use with the Transform tool. 
 *
 * Date created: 04/07/2004
 *
 * @author Eric Brown-Munoz<p>
 *
 */

public class TransformedQuadraticFunction extends QuadraticFunction 
	implements MouseMotionReceiver
{
	private QuadraticFunction original;
	
	private Stroke highlightedLineStroke;
	
	private boolean mouseInside = false;

	public TransformedQuadraticFunction() 
	{
		super();
	}
	
	/*
	 * Constructs and Initializes a Transformed Graphable with
	 *	the same starting function as the original 
	 */
	public TransformedQuadraticFunction(QuadraticFunction original) 
	{
		this.original = original;
		this.setExpressionForm(original.getExpressionForm());
		
		resetFromOriginal();
	}

	/*
	 * Resets the transformed quadratic function to be equal to the original
	 */
	public void resetFromOriginal()
	{
		setCoeffs(original.getCoeffA(), original.getCoeffB(),
				  original.getCoeffC());
	}
	
	public QuadraticFunction getOriginal()
	{
		return original;
	}

	/**
	 * Sets the drag mode of this function
	*/
	public void setDragMode(DragMode dragMode)
	{
		super.setDragMode(dragMode);
		changeCursor();
		this.notifyChange();
	}
	
	public double getAmountDragged() 
	// This method assumes there is only ONE transformation! (dragMode)
	//
	// The original function is 	ax^2 + bx + c
	// The transformed function is 	Ax^2 + Bx + C
	{
		//Calculate this value depending on the drag mode
		double a,b,c;
		double d = 0.0;
		int mode;
		a = original.getCoeffA();
		b = original.getCoeffB();
		c = original.getCoeffC();
		
		mode = getDragMode().getValue();
		
		if (mode == DragMode.TRANSLATE_X){
			if (!MathUtil.equalsDouble(a,0)){
				//d = (B-b)/2a
				d = (getCoeffB() - b) / (2*a);
			}
			else if (!MathUtil.equalsDouble(b,0)){
				//d = (C-c)/b;
				d = (getCoeffC() - c) / b;
			}
		}
		else if (mode == DragMode.TRANSLATE_Y){

			d = getCoeffC() - c;
		}
		else if (mode == DragMode.DILATE_X){
			
			if (!MathUtil.equalsDouble(b,0)){
				//d = B/b
				d = getCoeffB() / b;
			}
			else if (!MathUtil.equalsDouble(a,0)){
				//d = sqrt(A/a)
				d = Math.sqrt(getCoeffA()/a);
			}
		}
		else if (mode == DragMode.DILATE_Y){
			
			if (!MathUtil.equalsDouble(a,0)){
				d = getCoeffA()/a;
			}
			else if (!MathUtil.equalsDouble(b,0)){
				d = getCoeffB()/b;
			}
			else if (!MathUtil.equalsDouble(c,0)){
				d = getCoeffC()/c;
			}
		}
		
		return d;
		//return dragFactor;
	}
	
	public boolean mousePressed(Point location)
	{
		if (super.mousePressed(location)){
			changeCursor();
			return true;
		}
		else{
			return false;
		}
	}
	
	public boolean mouseReleased(Point location)
	{
		if (super.mouseReleased(location)){
			changeCursor();
			return true;
		}
		else{
			return false;
		}
	}

	public boolean mouseEntered(Point location)
	{
		int mode;
		
		mouseInside = false;

		mode = getDragMode().getValue();
		if (mode == DragMode.NONE || mode == DragMode.REFLECT_X || mode == DragMode.REFLECT_Y){
			return false;
		}

		if (highlightedLineStroke != null){
			mouseInside = true;
			notifyChange(false);
			return true;
		}

		return false;
	}

	public boolean mouseExited(Point location)
	{
		int mode;
		
		mouseInside = false;

		mode = getDragMode().getValue();
		if (mode == DragMode.NONE || mode == DragMode.REFLECT_X || mode == DragMode.REFLECT_Y){
			return false;
		}

		if (highlightedLineStroke != null){
			notifyChange(false);
			return true;
		}

		return false;
	}
	
	public boolean mouseMoved(Point location)
	{
		return false;
	}

	public boolean isMouseReceiving()
	{
		return mouseInside;
	}

    protected void changeCursor()
	{
		int mode;
		mode = getDragMode().getValue();
		
		if (mode == DragMode.NONE){
			setCursor(null);
		}
		else if (!isMouseControlled()){
			if (mode == DragMode.REFLECT_X || mode == DragMode.REFLECT_Y){
				setCursor(null);
			}
			else{
				setCursor(org.concord.graphfunction.ui.ResourceLoader.getCursorImage("images/hand.gif","drag cursor",15,8));
			}
		}
		else if (mode == DragMode.TRANSLATE_X || mode == DragMode.DILATE_X){
			setCursor(org.concord.graphfunction.ui.ResourceLoader.getCursorImage("images/hand_leftright.gif","horizontal drag cursor",15,8));
		}
		else if (mode == DragMode.TRANSLATE_Y || mode == DragMode.DILATE_Y){
			setCursor(org.concord.graphfunction.ui.ResourceLoader.getCursorImage("images/hand_topbottom.gif","vertical drag cursor",15,8));
		}
		else{
			setCursor(null);
		}
	}
	
	public Stroke getHighlightedLineStroke()
	{
		return highlightedLineStroke;
	}
	
	public void setHighlightedLineStroke(Stroke stroke)
	{
		highlightedLineStroke = stroke;
	}
	
	public Stroke getSelectedLineStroke()
	{
		if (mouseInside && highlightedLineStroke!=null){
			return highlightedLineStroke;
		}
		else {
			return super.getSelectedLineStroke();
		}
	}

	public Stroke getLineStroke()
	{
		if (mouseInside && highlightedLineStroke!=null){
			return highlightedLineStroke;
		}
		else {
			return super.getLineStroke();
		}
	}
	
}
