/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/* Author: Eric Brown-Munoz
   $Revision: 1.1.2.2 $
*/

package org.concord.graphfunction.ui;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;

import org.concord.graph.engine.*;
import org.concord.graph.event.*;
import org.concord.graphfunction.engine.*;
import org.concord.graphfunction.event.*;

/** A AreaModel that represents a quadratic function as a Rectangle. This
	takes a AreaQuad object which is a subclass of QuadGraphable (and can
	be displayed with GraphWindow). */

public class RectangleAreaModel extends JPanel implements ImageObserver,
														  MouseListener,
														  MouseMotionListener,
														  GraphableListener
{

	static double SIN45 = 0.7071;

	Value x;

	AreaQuad curve = null;

	double scale = 30.0;
	int originx = 200;
	int originy = 200;

	Color squaredColor = Color.blue;
	Color xColor = Color.green;
	Color scalarColor = Color.orange;
	Color lineColor = Color.black;
	Color handleColor = Color.red;
	Color labelColor = Color.blue;

	Point clicked = null;
	Point startDrag = null;

	Font labelFont = new Font("Serif",Font.BOLD, 14);

	boolean grid = false;

	int mouseMode;
	static int POINT = 0;
	static int OBJECT = 1;

	String root1String = "(x - r1)";
	String root2String = "(x - r2)";
	String constantString = "b";

	BufferedImage stripes;
	TexturePaint stripePaint;

	/** Constructs and initializes a RectangleArea Model/ */
	public RectangleAreaModel() {


		x = new TenthValue();

		x.addValueListener (new ValueListener () {
				public void valueChanged(ValueEvent e) {
					RectangleAreaModel.this.repaint();
				}
			});

		x.setValue(2.0);

		addMouseListener(this);
		addMouseMotionListener(this);

		stripes = createStripedImage();
		stripePaint =
			new	TexturePaint(stripes, 
							 new Rectangle2D.Double(0.0,0.0,
													(double)stripes.getWidth(this),
													(double)stripes.getHeight(this)));
	}

	/** Sets the AreaQuad being modeled to 'curve'. */
	public void setCurve (AreaQuad curve) {

		if (this.curve != null) {
			this.curve.removeGraphableListener(this);
		}
		root1String = "(x -" + curve.getRoot1() + ")";
		root2String = "(x -" + curve.getRoot2() + ")";
		constantString = Integer.toString(curve.getConstant());
		this.curve = curve;
		curve.addGraphableListener(this);
	}

	/** Returns the AreaQuad being modeled. */
	public AreaQuad getCurve() {
		return curve;
	}

	public boolean getGrid () {
		return grid;
	}

	public void setGrid( boolean gridValue ) {
		grid = gridValue;
		repaint();
	}

	public void setScale (double scale) {
		this.scale = scale;
		repaint();
	}

	public double getScale () {
		return scale;
	}

	/* Return the value object used by this model */
	public Value getValueObject() {
		return x;
	}

	public Point2D.Double getTopLeft () {

		Point2D.Double retval = null;

		double xValue = x.getValue();

		if (xValue < 0 ) {
			double xWidth = xValue * scale;
			retval =  new Point2D.Double(originx + xWidth,
										 originy + xWidth);
		} else {
			return new Point2D.Double(originx, originy);
		}

		return retval;
	}

	public double getRectangleWidth () {
		return (Math.abs(x.getValue()) + Math.abs((double)curve.getRoot1())) *scale ;
	} 

	public double getRectangleHeight () {
		return (Math.abs(x.getValue()) + Math.abs((double)curve.getRoot2())) *scale ;
	} 

	public void drawRectangle (Graphics2D g2, double posx, double posy) {

		double xval = x.getValue();
		int root1 = (int)curve.getRoot1();
		int root2 = (int)curve.getRoot2();
		int multiple = curve.getMultiple();
		int constant = curve.getConstant();

		int absRoot1 = Math.abs(root1);
		int absRoot2 = Math.abs(root2);
		double absXval = Math.abs(xval);

		double conx = 0.0;
		double cony = 0.0;
		int constRows =0;
		int constCols =0;

		Color savedColor = g2.getColor();

		/* Draw a square for x^2 */
		Rectangle2D.Double squaredRect = 
			new Rectangle2D.Double((double)posx, (double)posy,
								   Math.abs(xval)*scale,  // FIXME 
								   Math.abs(xval) * scale);

		g2.setColor(squaredColor);
		g2.fill(squaredRect);
		g2.setColor(lineColor);
		g2.draw(squaredRect);
		
		/* Draw the vertical rectangles for X.*/
		for (int ix = 0; ix < absRoot1 ; ix ++) {
			Rectangle2D.Double xRect =
				new Rectangle2D.Double( posx + (absXval + ix) * scale, 
										posy ,
										scale, absXval * scale);
			g2.setColor(xColor);
			g2.fill(xRect);

			if (( -root1 * xval) < 0 ) {
				g2.setPaint(stripePaint);
				g2.fill(xRect);
				g2.setPaint(null);
			}
				
			g2.setColor(lineColor);
			g2.draw(xRect);

		}

		/* Draw the horizontal x rectangles */
		for (int iy = 0; iy < absRoot2 ; iy ++) {
			Rectangle2D.Double xRect =
				new Rectangle2D.Double( posx ,
										posy +(absXval + iy) * scale, 
										absXval * scale, scale);
			g2.setColor(xColor);
			g2.fill(xRect);

			if (( -root2 * xval)  < 0 ) {
				g2.setPaint(stripePaint);
				g2.fill(xRect);
				g2.setPaint(null);
			}

			g2.setColor(lineColor);
			g2.draw(xRect);
		}

		/* Draw grid lines */
		if (grid) {

			double endx = posx + (absXval + absRoot1)*scale;
			double endy = posy + (absXval + absRoot2)*scale;

			g2.setColor(Color.gray);
			for ( double ilx = scale ; ilx < (absXval * scale); ilx += scale) {
				if (xval > 0.0) {
					g2.draw(new Line2D.Double( posx , posy+ilx, 
											   endx,
											   posy+ilx));
					g2.draw(new Line2D.Double( posx + ilx , posy , 
											   posx + ilx, 
											   endy));
				} else {

					double tempy = posy - xval*scale - ilx;
					double tempx = posx - xval*scale - ilx; 

					g2.draw(new Line2D.Double( posx , tempy,
											   endx,  tempy));
					g2.draw(new Line2D.Double( tempx, posy, 
											   tempx, endy));
				} // if xval> 0
			} // for
		}// if (grid)

		/* Draw the unit squares */
		g2.setColor(scalarColor);

		Rectangle2D.Double unitRect = 
			new Rectangle2D.Double( posx + absXval*scale, 
								   posy + absXval*scale,
								   absRoot1 * scale, absRoot2 * scale);

		g2.fill(unitRect);
		g2.setColor(lineColor);
		g2.draw(unitRect);

		if (( root1 * root2) < 0 ) {
			g2.setPaint(stripePaint);
			g2.fill(unitRect);
			g2.setPaint(null);
		}

		g2.setColor(lineColor);

		for (int iux = 0; iux <= absRoot1 ; iux ++ ) {
			g2.draw( new Line2D.Double( posx +( absXval + iux) *scale, 
										posy + absXval * scale,
										posx + ( absXval + iux) *scale, 
										posy + (absXval + absRoot2) * scale));
		}

		for (int iuy = 0; iuy <= absRoot2  ; iuy ++ ) {
			g2.draw( new Line2D.Double( posx + absXval * scale , 
										posy + ( absXval + iuy) *scale,
										posx + (absXval + absRoot1) * scale,
										posy + ( absXval + iuy) *scale)); 

		}


		/* Draw the Constant squares */
		if (constant != 0) {
			conx = posx + getRectangleWidth();
			cony = posy + getRectangleHeight();

			int absConstant = Math.abs(constant);
			constCols = (int)(Math.sqrt((double)absConstant));
			constRows = (int)(absConstant/constCols);
			int remainder = absConstant - constRows * constCols;

			Rectangle2D.Double con1 = 
				new Rectangle2D.Double(conx, cony, constCols * scale, constRows * scale);
			Rectangle2D.Double con2 =
				new Rectangle2D.Double(conx, cony + constRows * scale,
									   remainder * scale, scale);
			g2.setColor(scalarColor);
			g2.fill(con1);
			if (remainder != 0) g2.fill(con2);

			if (constant < 0 ) {
				g2.setPaint(stripePaint);
				g2.fill(con1);
				if (remainder != 0) g2.fill(con2);
				g2.setPaint(null);
			}



			g2.setColor(lineColor);

			/* Vertical lines */
			for (int icx = 0; icx <= constCols; icx ++) {
				if ((remainder != 0) && (icx <= remainder)) {
					g2.draw(new Line2D.Double( conx + icx*scale, cony,
											   conx + icx * scale,
											   cony + (constRows + 1) * scale));
				} else {
					g2.draw(new Line2D.Double( conx + icx*scale, cony,
											   conx + icx * scale,
											   cony + constRows * scale));
				}
			}

			/* Horizontal lines */
			for (int icy = 0; icy <= constRows; icy ++) {
				g2.draw(new Line2D.Double( conx, cony + icy*scale,
										   conx + constCols * scale,
										   cony + icy*scale
										   ));
			}										   
			if (remainder != 0) 
				g2.draw(new Line2D.Double( conx,
										   cony + (constRows + 1) * scale,
										   conx + remainder * scale,
										   cony + (constRows + 1) * scale));
				
		}

			

		/* Draw Labels */
		g2.setFont(labelFont);
		g2.setColor(labelColor);

		FontMetrics fm = g2.getFontMetrics();
		Rectangle2D sb1 = fm.getStringBounds(root1String, g2);
		Rectangle2D sb2 = fm.getStringBounds(root2String, g2);

		double l1posx = posx + ( getRectangleWidth() - sb1.getWidth())/ 2.0 - 3.0;
		double l1posy = posy - 4.0;

		g2.drawString(root1String,(float)l1posx, (float)l1posy);

		AffineTransform savedtransform = g2.getTransform();

		g2.rotate(-Math.PI/2.0);
		g2.translate( - posy - (getRectangleHeight()+ sb2.getWidth())/2.0 ,
					  posx - 4.0);
		g2.drawString(root2String, 0.0f, 0.0f);

		g2.setTransform(savedtransform);
		
		if (constant != 0) 
			g2.drawString(constantString,(float)(conx + constCols * scale + 5),
						  (float)cony + 15.0f);


		g2.setColor(savedColor);
	}

	public void paintComponent (Graphics g) {

		if (curve == null) {
			return;
		}

		// coords of top left corner of rectangle;
		double topleftx = getTopLeft().x; 
		double toplefty = getTopLeft().y;

		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;

		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							RenderingHints.VALUE_ANTIALIAS_ON);

		double xval = (double)x.getValue();
		double absXval = Math.abs(xval);

		AffineTransform savedTransform = g2.getTransform();
		Color savedColor = g2.getColor();

		Rectangle2D.Double squaredRect = null;

		g2.setColor(squaredColor);

		/* Draw the Squared Rectangle */
		drawRectangle(g2, topleftx, toplefty);



		/* Draw the handle */

		Ellipse2D.Double handleEllipse = 
			new Ellipse2D.Double( originx + xval * scale - 4, 
								  originy + xval * scale - 4,
								  9, 9);
		g2.setColor(handleColor);
		g2.draw(handleEllipse);
						 
		g2.setColor(savedColor);
		g2.setTransform(savedTransform);

		// Paint the x-squared piece
	   
	}

	BufferedImage createStripedImage () {
		BufferedImage pattern = new BufferedImage(100,200, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = (Graphics2D)pattern.getGraphics();

		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							RenderingHints.VALUE_ANTIALIAS_ON);

		g2.setStroke (new BasicStroke(5));
		g2.setColor(Color.white);

		double ypos = -80;

		while (ypos < 200) {
			g2.draw(new Line2D.Double(0.0, ypos, 100.0, 
									  ypos + 100.00 * SIN45));

			ypos += 10.0;
		}

		return pattern;
		
	}

	public static void main(String args[]) {

		final JFrame frame = new JFrame();
		final JPanel fa = new RectangleAreaModel();
		frame.getContentPane().add(fa);
		frame.setSize(600,600);
		frame.show();
		
	}	

	public boolean imageUpdate(Image img, int infoflags, 
							   int x, int y, int width, int height) {
		return false;
	}

	public void mouseDragged (MouseEvent e) {

		if (mouseMode == POINT) {
			double xtemp = ( e.getX() - originx ); 
			double ytemp = ( e.getY() - originy ); 

			x.setValue(  SIN45 * SIN45 * ( + xtemp +  ytemp ) / scale  );
		} else if (mouseMode == OBJECT) {
			originx = startDrag.x + (e.getX() - clicked.x);
			originy = startDrag.y + (e.getY() - clicked.y);
		}
		repaint();
	}

	public void mouseMoved (MouseEvent e) {}

	public void mouseClicked(MouseEvent e) {}

	public void mouseEntered(MouseEvent e) {}

	public void mouseExited(MouseEvent e) {}

	public void mousePressed(MouseEvent e) {

		double xpos = (double)x.getValue() * scale;
		
		double distance = Math.sqrt( Math.pow((double)e.getX() - 
											  (xpos + originx),2.0) +
									 Math.pow((double)e.getY() - 
											  (xpos+originy),2.0) );

		
		clicked = new Point(e.getX(), e.getY());
		startDrag = new Point (originx, originy);

		if (distance < 9) {
			mouseMode = POINT;
		} else {
			mouseMode = OBJECT;
		}
							  

	}

	public void mouseReleased(MouseEvent e) {
		clicked = null;
	}

	public void graphableChanged (java.util.EventObject e ) {
		root1String = "(x -" + curve.getRoot1() + ")";
		root2String = "(x -" + curve.getRoot2() + ")";
		constantString = Integer.toString(curve.getConstant());
		repaint();
	}

	public void graphableRemoved (java.util.EventObject e ) {
		curve = null;
		repaint();
	}
		

}
				
