/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.3 $
 * $Date: 2004/07/30 18:10:02 $
 * $Author: jqian $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.ui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.Vector;

import org.concord.graphfunction.engine.*;

/**
 * An abstract class that can be used as a label to display double values
 *
 * @author Eric Brown-Munoz<p>
 * @author Ingrid Moncada<p>
*/

public abstract class ValueBox extends JLabel 
	implements PropertyBox, ItemSelectable
{
	/**
	 * Only notifies selections made by clicking with the mouse
	*/
	
	protected Object source;
	private boolean selected;
	private boolean editable = true;
	
	Vector selectionListeners;
	
	protected Color selectedColor = Color.cyan;
	
	private double increment = 0.1;
	private int defaultNumDecimalsPrecision = 1;
	private int numDecimalsPrecision = -1;	//Not set
	
	private Border border;
	
	public ValueBox()
	{
		this(null);
	}
	
	public ValueBox(Object source) 
	{
		setSource(source);
		reset();
	
		addMouseListener(new MouseAdapter () 
		{
			public void mouseClicked (MouseEvent e) 
			{
				if (!editable) return;
								
				if (selected && hasFocus()){
					deselect();
				}
				else if (!selected){
					select();
					requestFocus();
				}
				else if (!hasFocus()){
					requestFocus();
				}
				
				notifySelection();
			}
		});
		
		setOpaque(false);
		
		selectionListeners = new Vector();
		
		adjustSize();
		
		setHorizontalAlignment(RIGHT);
		
		this.addKeyListener(keyListener);
		
		//border = BorderFactory.createEtchedBorder();
		border = BorderFactory.createLineBorder(new Color(235,235,235));
		this.setBorder(border);
	}
	
	public void setSource(Object source)
	{
		this.source = source;
	}
	
	public abstract double getValue();
	protected abstract void setValue(double value);
	
	public void setEditable(boolean b)
	{
		editable = b;
		if (!b){
			hideBorder();
		}
		else{
			showBorder();
		}
	}
	
	public void hideBorder()
	{
		this.setBorder(null);
	}
	
	public void showBorder()
	{
		this.setBorder(border);
	}
	
	public boolean isEditable()
	{
		return editable;
	}
	
	public void reset()
	{
        double value;
		int numD;
		String strText = "";
		if (source != null){
			//strText = MathUtils.roundTenth(this.getValue(), precision);
	
			numD = getNumDecimalsValue();
            value = this.getValue();
            if (Double.isInfinite(value))
            {
                if (value == Double.POSITIVE_INFINITY)
                {
                    strText = "\u221e";
                }
                else
                {
                    strText = "\u2212\u221e";
                }
            }
            else
            {
			    strText = MathUtil.getStringRound(value, numD);
			    strText = strText+" ";
            }
	
		}
		setText(strText);
		setToolTipText(strText);
	}
	
	/**
	 * Rounds the value according with the decimals value
	 */
	public void roundWithDecimals()
	{
		int numD;
		double value;
		
		numD = getNumDecimalsValue();
		value = this.getValue();
		//System.out.println("old value:"+value+" decimals:"+numD);
		value = MathUtil.roundTenth(value, Math.pow(10,numD));
		//System.out.println("new value:"+value);
		this.setValue(value);
	}
	
	/**
	 * Rounds the value according with the increment value
	 */
	public void roundWithIncrementInterval()
	{
		int numD;
		double value;
		
		numD = defaultNumDecimalsPrecision;
		value = this.getValue();
		//System.out.println("old value:"+value+" defaultNumDecimalsPrecision:"+defaultNumDecimalsPrecision);
		value = MathUtil.roundTenth(value, Math.pow(10,numD));
		//System.out.println("new value:"+value);
		this.setValue(value);
	}

	public void setSelectedColor(Color c)
	{
		selectedColor = c;
	}
	
	public Color getSelectedColor()
	{
		return selectedColor;
	}

	public void setIncrementInterval(double increment)
	{
		this.increment = increment;
	
		if (increment > 0 && increment < 1){
			defaultNumDecimalsPrecision = Math.abs(MathUtil.log10(increment));
			//precision = (int)Math.pow(10,precision);
		}
		else{
			defaultNumDecimalsPrecision = 0;
		}
	}
	
	public void setNumDecimalsValue(int n)
	{
		if (n>=0){
			numDecimalsPrecision = n;
			adjustSize();
		}
	}
	
	public int getNumDecimalsValue()
	{
		if (numDecimalsPrecision != -1){
			return numDecimalsPrecision;
		}
		else{
			return defaultNumDecimalsPrecision;
		}
	}

	public double getIncrementInterval()
	{
		return this.increment;
	}
	
	public void increment() 
	{
		if (source != null){
			setValue(getValue() + increment);
		}
	}
	
	public void decrement() 
	{
		if (source != null){
			setValue(getValue() - increment);
		}
	}
	
	public void select() 
	{
		if (!editable) return;
		selected = true;
		setOpaque(true);
		setBackground(selectedColor);
		hideBorder();
	}
	
	public void deselect() 
	{
		if (!editable) return;
		selected = false;
		setOpaque(false);
		if (getParent()!=null){
			setBackground(getParent().getBackground());
		}
		showBorder();
	}
	
	protected void adjustSize()
	//Default width was 43
	{
		int width = 35;

		width = width + (getNumDecimalsValue()*10);
		
		setPreferredSize(new Dimension(width,27));
		setMinimumSize(new Dimension(width,27));
		setMaximumSize(new Dimension(width,27));
	}
	
	/////////////////////////////////////////////////////
	//Methods from ItemSelectable
	public void addItemListener(ItemListener l)
	{
		if (!selectionListeners.contains(l)){
			selectionListeners.addElement(l);
		}
	}
	
	public Object[] getSelectedObjects()
	{
		return null;
	}
	
	public void removeItemListener(ItemListener l)
	{
		selectionListeners.remove(l);
	}
	/////////////////////////////////////////////////////
	
	private void notifySelection()
	{
		int i;
		ItemEvent e;
		if (selected){
			e = new ItemEvent(this, 0, this, ItemEvent.SELECTED);
		}
		else{
			e = new ItemEvent(this, 0, this, ItemEvent.DESELECTED);
		}
		
		for (i=0;i<selectionListeners.size();i++){
			ItemListener l = (ItemListener)selectionListeners.elementAt(i);
			l.itemStateChanged(e);
		}
	}
	
	//Handle arrow keys to increment and decrement values
	KeyListener keyListener = new KeyAdapter(){
		public void keyPressed(KeyEvent e)
		{
			if (!isEditable() || !selected) return;
			if (e.getID() != KeyEvent.KEY_PRESSED) return;
			if (e.getKeyCode() == KeyEvent.VK_UP){
				increment();
			}
			else if (e.getKeyCode() == KeyEvent.VK_DOWN){
				decrement();
			}
			else if (e.getKeyCode() == KeyEvent.VK_RIGHT){
				increment();
			}
			else if (e.getKeyCode() == KeyEvent.VK_LEFT){
				decrement();
			}
		}
	};
	
}




