/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.2 $
 * $Date: 2004/07/14 17:23:31 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.ui.modelview;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import org.concord.graph.event.*;
import org.concord.graphfunction.engine.*;
import org.concord.graphfunction.event.*;
import org.concord.graphfunction.ui.*;

/**
 * AreaModelView.
 * This is a view of a graph function
 * It represents the graph function drawing its value as a filled area
 * made by boxes that represent a value of 1 unit.
 * By default, red boxes represent variable values, and green boxes
 * represent constants.
 * 
 * @author Ingrid Moncada<p>
*/

public class AreaModelView extends JPanel 
	implements GraphFunctionView, 
			   ValueListener,
			   SelectableListListener
{	
	//Only represents one function
	GraphFunction function;

	// ... selected from this List
	GraphFunctionList glist;
	
	//Vector of area boxes
	Vector areaBoxes;
	
	//Variable to listen to 
	Value variableValue;
	
	public static final int NONE_ONCLICK = 0;
	public static final int DELETE_ONCLICK = 1;

	private int clickBehavior = NONE_ONCLICK;
	
	int drawMode = LinearAreaBox.DRAW_SINGLEBOXES;
	
	int maxUnitSizePx = 40;
	
	int unitSizePx;
	int totalBoxSizePx;
	
	boolean autoResizeHorizontal = true; 
	boolean autoResizeVertical = true; 
	
	boolean showTotalNumbers = false;
	
	int currentHeight = 0;
	int currentWidth = 0;

	//Hack to only take care of changes of slope of yIntercept
	double currentM = Double.NaN;
	double currentB = Double.NaN;
	
	String title = "Virtual Algebra Tiles Model";
	/**
	 * Constructs an empty area model view
	*/
	public AreaModelView()
	{
		areaBoxes = new Vector();
		this.setLayout(null);
	}
	
	/**
	 * Constructs an area model view for a GraphFunction
	*/
	public AreaModelView(GraphFunction g)
	{
		this();
		setGraphFunction(g);
	}

	/**
	 * Sets the draw mode for this view.<p>
	 * The possible values are specified in the LinearAreaBox class:
	 * DRAW_BAR, DRAW_BOXES, DRAW_SINGLEBOXES
	 * @see org.concord.mgraph.ui.LinearAreaBox#setDrawMode
	*/
	public void setDrawMode(int drawMode)
	{
		int i;

		this.drawMode = drawMode;
		if (areaBoxes.size() > 0){
			for (i=0; i<areaBoxes.size(); i++){
				((LinearAreaBox)areaBoxes.elementAt(i)).setDrawMode(drawMode);
			}
			resizeComponent();
			repaint();
		}
		
	}

	/**
	 * Returns the draw mode for this view.<p>
	 * @see org.concord.mgraph.ui.LinearAreaBox#getDrawMode
	*/
	public int getDrawMode()
	{
		return this.drawMode;
	}

	/**
	 * Sets the GraphFunction that this view will represent. 
	 * It accepts any GraphFunction, but currently only linear objects will be displayed.
	*/
	public void setGraphFunction(GraphFunction g)
	{

		if (function!=null){
			function.removeGraphableListener(graphableHandler);
		}
		
		if (g instanceof FunctionEquation){
			function = ((FunctionEquation)g).getFunction();
		}
		else{
			function = g;
		}
		
		if (function!=null){
			function.addGraphableListener(graphableHandler);
		}

		initView();
		resizeComponent();
		repaint();
	}

	/***/
	private GraphableListener graphableHandler = new GraphableAdapter() 
	{
		/* Listener to the changes in the GraphFunction.
		Repaints the view when it changes.
		*/
	    public void graphableChanged(EventObject e) {
			//Hack to only take care of changes of slope of yIntercept
			if (isChangeRelevant((GraphFunction)e.getSource())){
				initView();
				resizeComponent();
				repaint();
			}
	    }
	};

	public void setGraphFunctionList(GraphFunctionList gl)
	{
		if (glist!=null){
			glist.removeGraphableListListener(this);
		}
		glist = gl;
		if (glist!=null){
			glist.addGraphableListListener(this);
			GraphFunction current = (GraphFunction)gl.getFirstSelectedObject();
			if (current != null) {
				setGraphFunction(current);
			}
		}
	}
	
	/**
	 * Sets the Value that this view will listen to as the variable
	*/
	public void setVariableValue(Value varValue)
	{
		if (variableValue != null) {
			variableValue.removeValueListener(this);
		}

		variableValue = varValue;

		if (variableValue != null) {
			variableValue.addValueListener(this);
		}
		
		setVariableValueBoxes();
		resizeComponent();
		repaint();
	}	

	/**
	 * Listener for the changes in the variable value
	*/
    public void valueChanged (ValueEvent e)
	{
		resizeComponent();
		repaint();
	}
	
	/** 
	 * Sets if the view will resize automatically or not,
	 * vertically or horizontally.
	*/
	public void setAutoResize(boolean horizontal, boolean vertical)
	{
		autoResizeHorizontal = horizontal;
		autoResizeVertical = vertical;
		resizeComponent();
	}
	
	/**
	 * Sets if the component will display the total numbers 
	*/
	public void setShowTotalNumbers(boolean b)
	{
		this.showTotalNumbers = b;
	}

	/**
	 * Returns if the view resizes automatically in horizontal direction  
	*/
	public boolean isAutoResizeHorizontal()
	{
		return autoResizeHorizontal;
	}
	
	/**
	 * Returns if the view resizes automatically in vertical direction  
	*/
	public boolean isAutoResizeVertical()
	{
		return autoResizeVertical;
	}

	/**
	 * Sets the maximum size in pixels of the boxes in the view.
	 * Even if auto resize is on, the boxes will not be bigger than the size specified.
	 * The default value is 40 pixels.
	 * To specify that there is no limit, set it to 0.
	*/	
	public void setMaxUnitSizePx(int sizePixels)
	{
		if (sizePixels >= 0){
			maxUnitSizePx = sizePixels;
		}
	}

	private void removeAllBoxes()
	{
		//Removes listeners from boxes
		int i;
		for (i=0; i<areaBoxes.size();i++){
			LinearAreaBox box = (LinearAreaBox)areaBoxes.elementAt(i);
			box.removeMouseListener(mouseListener);
		}
		areaBoxes.removeAllElements();
		
		this.removeAll();
	}
	
	private void addBox(LinearAreaBox box)
	{
		box.setDrawMode(drawMode);
		areaBoxes.addElement(box);
		box.addMouseListener(mouseListener);
		
		box.setSize(maxUnitSizePx+1, this.getHeight());
		this.add(box);
	}

	private boolean isChangeRelevant(GraphFunction g)
	{
		if (Double.isNaN(currentM) || Double.isNaN(currentB)) return true;
		if (g instanceof LinearFunction){
			LinearFunction LinearFunction = (LinearFunction)function;
			double m, b;
			m = LinearFunction.getSlope();
			b = LinearFunction.getYIntercept();
			if (currentM != m || currentB != b){
				return true;
			}
		}
		return false;
	}
	
	/** 
	 * Inititalizes the view according to the new GraphFunction
	*/
	private void initView()
	{
		removeAllBoxes();
		
		if (function==null) return;
		
		if (function instanceof LinearFunction){
			LinearFunction LinearFunction = (LinearFunction)function;
			
			double m, b;
			int intM, intB;
			double fracM, fracB;
			double coeff;
			int i;

			m = LinearFunction.getSlope();
			b = LinearFunction.getYIntercept();
			
			currentM = m;
			currentB = b;
			
			intM = (int)Math.floor(Math.abs(m));
			fracM = Math.abs(m) - intM;
			intB = (int)Math.floor(Math.abs(b));
			fracB = Math.abs(b) - intB;
			
			LinearAreaBox box;

			//Create one area box for each coeff of x
			coeff = m/Math.abs(m);
			for (i=0;i<intM;i++){
				box = new LinearAreaBox(coeff,variableValue);
				addBox(box);
			}
			//Create a partial area box for the decimal part of m
			if (fracM > 0.1){
				box = new LinearAreaBox(fracM*coeff,variableValue);
				addBox(box);
			}
			
			//Create one area box for each b
			coeff = b/Math.abs(b);
			for (i=0;i<intB;i++){
				box = new LinearAreaBox(coeff);
				addBox(box);
			}
			//Create a partial area box for the decimal part of b
			if (fracB > 0.1){
				box = new LinearAreaBox(fracB*coeff);
				addBox(box);
			}

		}
		else{
		//For now, only linear functions supported!
			System.err.println("Area Model View: GraphFunction is not linear!");
		}
		
	}

	/**
	 * Sets the variable in all the area boxes
	*/
	private void setVariableValueBoxes()
	{
		int i;
		for (i=0; i<areaBoxes.size();i++){
			LinearAreaBox box = (LinearAreaBox)areaBoxes.elementAt(i);
			box.setVariableValue(variableValue);
		}
	}
	
	public void resizeComponent()
	{
		currentHeight = this.getHeight();
		currentWidth = this.getWidth();
		
		if (function == null || variableValue == null) return;
		
		unitSizePx = maxUnitSizePx;

		//Calculate box size if autoresize is on
		if (autoResizeHorizontal && areaBoxes.size()>0){
			totalBoxSizePx = currentWidth/areaBoxes.size();
			unitSizePx = (totalBoxSizePx * 10) / 11;
		}
		totalBoxSizePx = unitSizePx + (unitSizePx/10);
		if (autoResizeVertical && variableValue.getValue()!=0 && 
				(int)Math.floor(Math.abs(((LinearFunction)function).getSlope())) > 0) {
					
			int totalBoxSizePx2 = (int)(currentHeight/ Math.ceil(Math.abs(variableValue.getValue())));
			if (totalBoxSizePx2<totalBoxSizePx){
				totalBoxSizePx = totalBoxSizePx2;
				if (drawMode == LinearAreaBox.DRAW_SINGLEBOXES){
					unitSizePx = (totalBoxSizePx * 10) / 11;
				}
				else{
					unitSizePx = totalBoxSizePx;
				}
			}
		}
		if (unitSizePx>maxUnitSizePx){
			unitSizePx = maxUnitSizePx;
		}
		totalBoxSizePx = unitSizePx + (unitSizePx/10);
		
		//Resize and relocate boxes
		int i;
		for (i=0; i<areaBoxes.size();i++){
			LinearAreaBox box = (LinearAreaBox)areaBoxes.elementAt(i);
			
			//box.paint(g, totalBoxSizePx * i, this.getHeight(), unitSizePx);
			
			box.setSize(unitSizePx+1,currentHeight);
			box.setLocation(totalBoxSizePx * i, 0);
			box.setUnitSize(unitSizePx);
		}
	}
	
	/** 
	 * Returns the max number of boxes that will fit in the screen horizontally
	*/
	public int getMaxBoxesHorizontal()
	{
		//With autoresize on, the boxes can be up to 1 pixels each
		if (this.autoResizeHorizontal) return this.getWidth();
		
		if (currentHeight!=this.getHeight() || currentWidth!=this.getWidth()){
			this.resizeComponent();
		}
		if (totalBoxSizePx==0) return 0;
		return (int)Math.floor(this.getWidth()/totalBoxSizePx);
	}
	
	/** 
	 * Returns the max number of boxes that will fit in the screen vertically
	*/
	public int getMaxBoxesVertical()
	{
		//With autoresize on, the boxes can be up to 1 pixels each
		if (this.autoResizeVertical) return this.getHeight();
		
		if (currentHeight!=this.getHeight() || currentWidth!=this.getWidth()){
			this.resizeComponent();
		}
		if (drawMode == LinearAreaBox.DRAW_SINGLEBOXES){
			if (totalBoxSizePx==0) return 0;
			return (int)Math.floor(this.getHeight()/totalBoxSizePx);
		}
		else{
			if (unitSizePx==0) return 0;
			return (int)Math.floor(this.getHeight()/unitSizePx);
		}
	}

	/** 
	 * Returns how many boxes would a GraphFunction have in this view
	*/
	public static int getNumBoxesHorizontal(GraphFunction g)
	{
		if (g==null) return 0;
		if (g instanceof LinearFunction){
			LinearFunction LinearFunction = (LinearFunction)g;
			
			return (int)(Math.ceil(Math.abs(LinearFunction.getSlope())) + Math.ceil(Math.abs(LinearFunction.getYIntercept())));
		}
		return 0;		
	}
	
	/** 
	 * Returns how many boxes would a GraphFunction have in this view
	*/
	public static int getNumBoxesVertical(GraphFunction g, Value val)
	{
		if (g==null || val==null) return 0;
		if (g instanceof LinearFunction){
			
			return (int)(Math.ceil(Math.abs(val.getValue())));
		}
		return 0;		
	}

	public void paintComponent(Graphics g)
	{
		if (currentHeight!=this.getHeight() || currentWidth!=this.getWidth()){
			this.resizeComponent();
		}
		super.paintComponent(g);
		///////////////////////////////////////////////////////////////////////////
		//Temporary-> print information about the GraphFunction
		//g.drawString("Area Model", 10, 10);		
		//if (function instanceof LinearFunction){
		//	LinearFunction linearFunction = (LinearFunction)function;
		//	g.drawString(linearFunction.getExpression(), 10, 30);
		//}
		//if (variableValue!=null){
		//	g.drawString("X="+variableValue.getValue(), 10, 50);
		//}
		///////////////////////////////////////////////////////////////////////////
		Font oldFont = g.getFont();
		Color oldColor = g.getColor();
		
		g.setFont(new Font("Dialog",Font.BOLD,14));
		if (function!=null){
			g.setColor(function.getColor());
		}
		else{
			g.setColor(Color.black);
		}
		g.drawString(title, this.getWidth()- 230,25);
		
		g.setFont(oldFont);
		g.setColor(oldColor);
	}
	
	/**
	 * Sets what to do when clicking on boxes
	*/
	public void setClickBehavior(int mode)
	{
		if (mode == DELETE_ONCLICK || mode == NONE_ONCLICK){
			this.clickBehavior = mode;
			if (this.clickBehavior == DELETE_ONCLICK){
				ResourceLoader.setCursorImage(this, "erasercursor.gif", "erasercursor", 0, 29);
			}
			else if (this.clickBehavior == NONE_ONCLICK){
				this.setCursor(null);
			}
		}
	}

	/**
	 * Listener for the mouse click and drag
	*/
    private MouseListener mouseListener = new MouseAdapter () {

		/* Listener to the mouse
		*/
		public void mousePressed(MouseEvent e)
		{
			if (clickBehavior == NONE_ONCLICK) return;
			
			LinearAreaBox box;
			box = (LinearAreaBox)e.getSource();
			
			if (clickBehavior == DELETE_ONCLICK){
				//Remove the box
				if (function instanceof LinearFunction){
					LinearFunction LinearFunction = (LinearFunction)function;
					if (box.isVariable()){
						LinearFunction.setSlope(LinearFunction.getSlope()-box.getCoefficient());
					}
					else{
						LinearFunction.setYIntercept(LinearFunction.getYIntercept()-box.getCoefficient());
					}
				}//if (function instanceof LinearFunction)
			}
		}

		public void mouseClicked(MouseEvent e)
		{
		}
 
	};

	public void listGraphableAdded(EventObject e)
	{
		repaint();
	}
	public void listGraphableChanged(EventObject e) 
	{
		repaint();
	}
	public void listGraphableRemoved(EventObject e) 
	{
		repaint();
	}
	public void listGraphableSelected(EventObject e) 
	{
		setGraphFunction((GraphFunction)e.getSource());
		repaint();

	}
	public void listGraphableDeselected(EventObject e) 
	{
	}


} // class AreaModelView













