/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.1 $
 * $Date: 2004/05/24 20:55:23 $
 * $Author: imoncada $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.ui.modelview;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

import org.concord.graphfunction.engine.*;

/**
 * LinearAreaBox.
 * Box that represents a value as a filled area. <br>
 * It can represent a constant value, or it can listen to a variable value. 
 * It can draw the value as a vertical bar in the screen, or as boxes piled up vertically.
 * 
 * @author Ingrid Moncada<p>
*/

public class LinearAreaBox extends JComponent
{
	private Value variableValue;	//Value to represent
	private double coefficient;
	
	private Color posIntegerColor;
	private Color posFractionColor;
	private Color negIntegerColor;
	private Color negFractionColor;
	private Color posDeltaColor;
	private Color negDeltaColor;
	
	final double minFracValue = 0.05; 
	
	public static final int DRAW_BAR = 0;
	public static final int DRAW_BOXES = 1;
	public static final int DRAW_SINGLEBOXES = 2;
	
	public static final Color DEFAULT_COLOR_CONSTANT = new Color(0,180,0);
	public static final Color DEFAULT_COLOR_VARIABLE = new Color(255,0,0);

	private int drawMode;
	
	private int unitSize=1;
	
	private int boxHeight=0;

	/**
	 * Constructs an area box that represents a constant value<br>
	 * By default, the colors for constants are GREEN for positive integer values,
	 * light green for decimal portions and dark green for negative values.
	 * @param value 	constant value to represent
	*/
	public LinearAreaBox(double value)
	{
		variableValue = null;
		coefficient = value;
		//Defaul colors for constants:		(green)
		setColors(	DEFAULT_COLOR_CONSTANT,
					new Color(0,210,0),
					new Color(0,210,0),
					new Color(0,50,0)   ,
					new Color(0,80,0),
					new Color(0,80,0)   );
	}
	
	/**
	 * Constructs an area box that represents a variable value<br>
	 * By default, the colors for variables are RED for positive integer values,
	 * light red (pink) for decimal portions and dark red for negative values.
	 * @param coeff 	coefficient of the variable to represent
	 * @param varValue 	variable to represent
	*/
	public LinearAreaBox(double coeff, Value varValue)
	{
		variableValue = varValue;
		coefficient = coeff;
		//Defaul colors for variables:		(red)
		setColors(	DEFAULT_COLOR_VARIABLE,
					new Color(255,100,100),
					new Color(255,150,150),
					new Color(100,0,0)   ,
					new Color(150,0,0),
					new Color(200,0,0)   );
	}

	/**
	 * Sets the dependent variable
	*/
	public void setVariableValue(Value varValue)
	{
		variableValue = varValue;
	}
	
	/**
	 * Returns if the component depends on the variable or not
	*/
	public boolean isVariable()
	{
		if (variableValue==null) return false;
		return true;
	}

	/**
	 * Returns if the component is a constant or not
	*/
	public boolean isConstant()
	{
		return !isVariable();
	}

	/**
	 * Returns the coefficent 
	*/
	public double getCoefficient()
	{
		return coefficient;
	}
	
	/**
	 * Sets all the colors as the color specified
	*/
	public void setColor(Color c)
	{
		setColors(c,c,c,c,c,c);
	}
	
	/**
	 * Set the different colors of the box
	 * @param posInt 	Color used to draw the positive integer portion of the box
	 * @param posFrac 	Color used to draw the positive decimal portion of the box
	 * @param negInt 	Color used to draw the negative integer portion of the box
	 * @param negFrac 	Color used to draw the negative decimal portion of the box
	*/
	public void setColors(Color posInt, Color posFrac, Color posDelta, Color negInt, Color negFrac, Color negDelta)
	{
		setPositiveIntegerColor(posInt); 
		setPositiveFractionColor(posFrac);
		setPositiveDeltaColor(posDelta);
		setNegativeIntegerColor(negInt); 
		setNegativeFractionColor(negFrac);
		setNegativeDeltaColor(negDelta);
	}

	/**
	 * Set the color used to fill the positive integer portion of the box
	*/
	public void setPositiveIntegerColor(Color c)
	{
		if (c==null) return;
		posIntegerColor = c; 
	}

	/**
	 * Set the color used to fill the positive decimal portion of the box
	*/
	public void setPositiveFractionColor(Color c)
	{
		if (c==null) return;
		posFractionColor = c;
	}

	/**
	 * Not used for now
	*/
	private void setPositiveDeltaColor(Color c)
	{
		if (c==null) return;
		posDeltaColor = c;
	}
		
	/**
	 * Set the color used to fill the negative integer portion of the box
	*/
	public void setNegativeIntegerColor(Color c)
	{
		if (c==null) return;
		negIntegerColor = c; 
	}

	/**
	 * Set the color used to fill the negative decimal portion of the box
	*/
	public void setNegativeFractionColor(Color c)
	{
		if (c==null) return;
		negFractionColor = c;
	}

	/**
	 * Not used for now
	*/
	private void setNegativeDeltaColor(Color c)
	{
		if (c==null) return;
		negDeltaColor = c;
	}
	
	/**
	 * Returns the draw mode for this view.<p>
	 * The possible values are: <br>
	 * DRAW_BAR.			Draws the values as vertical bars<br>
	 * DRAW_BOXES.			Draws the values as boxes stuffed vertically (one on top of the other)<br>
	 * DRAW_SINGLEBOXES. 	Draws the values as boxes stuffed vertically with a space in between them<br>
	*/	
	public int getDrawMode()
	{
		return drawMode;
	}

	/**
	 * Sets the draw mode for this view.
	 * @param mode	draw mode for the boxes
	 * The possible values are: <br>
	 * DRAW_BAR.			Draws the values as vertical bars<br>
	 * DRAW_BOXES.			Draws the values as boxes stuffed vertically (one on top of the other)<br>
	 * DRAW_SINGLEBOXES. 	Draws the values as boxes stuffed vertically with a space in between them<br>
	*/	
	public void setDrawMode(int mode)
	{
		if (mode == DRAW_BAR ||	mode == DRAW_BOXES || mode == DRAW_SINGLEBOXES){
			drawMode = mode;
		}
	}
	
	/**
	 * Sets the unit size
	 * @param unitSize	unit size in pixels
	*/	
	public void setUnitSize(int unitSize)
	{
		this.unitSize = unitSize;
		resizeComponent();
	}

	public void paintComponent(Graphics g)
	{
//		System.out.println("painting component...("+this.getX()+","+this.getY()+") ("+this.getWidth()+" "+this.getHeight());
		paint(g,0,this.getHeight(),this.unitSize);
	}
	
	public void paint(Graphics g, int posX_BL, int posY_BL, int unitSize)
	{
		double value;
		int intValue;
		double fracValue;
		int yPos;
		double coeff;

		if (variableValue==null){
			value = 1;
		}
		else{
			value = variableValue.getValue();
		}
		//Round the value to pay attention to decimals 2 order (0.01)
		value = (double)(Math.round(value*100))/100;
		
		intValue = (int)Math.floor(Math.abs(value));
		fracValue = Math.abs(value) - intValue;
		coeff = Math.abs(coefficient);
		
		Color oldColor = g.getColor();
		//Paint a box 
		//width  = coefficient
		//height = value
		int unitSizePx = unitSize;		
		int width, height;

		//The position (posX_BL, posY_BL) is the left lower corner

		if (drawMode == DRAW_BAR){
		
			//Draw the whole border in black
			width = unitSizePx;
			height = (int)Math.round(Math.abs(value) * unitSizePx);
			yPos = posY_BL - height;
			g.setColor(Color.black);
			g.drawRect(posX_BL, yPos-1, width, height);

			//Integer portion
			if (Math.abs(intValue) > 0){
				//Draw a bar for the integer portion
				width = (int)Math.round(coeff * unitSizePx);
				height = intValue * unitSizePx;
				yPos = posY_BL - height;
				if (value*coefficient > 0){
					g.setColor(posIntegerColor);
				}
				else{
					g.setColor(negIntegerColor);
				}
				
				g.fillRect(posX_BL+1, yPos, width-1, height-1);
			}
	
			//Fraction portion
			if (fracValue >= minFracValue){
				//Draw a bar for the fraction portion
				width = (int)Math.round(coeff * unitSizePx);
				height = (int)Math.round(fracValue * unitSizePx);		
				if (Math.abs(intValue) > 0){
					yPos = posY_BL - (intValue * unitSizePx) - height;
					height = height + 1;
				}
				else{
					yPos = posY_BL - height;
				}
				if (value*coefficient > 0){
					g.setColor(posFractionColor);
				}
				else{
					g.setColor(negFractionColor);
				}
				g.fillRect(posX_BL+1, yPos, width-1, height-1);
			}
			
		}//drawMode == DRAW_BAR
		else if (drawMode == DRAW_BOXES){
		
			//Draw the whole border in black
			width = unitSizePx;
			height = intValue * unitSizePx;
			if (fracValue >= minFracValue){
				height = height + unitSizePx;
			}
			yPos = posY_BL - height;
			g.setColor(Color.black);
			g.drawRect(posX_BL, yPos-1, width, height);

			//Integer portion
			if (Math.abs(intValue) > 0){
				//Draw a bar for the integer portion
				width = (int)Math.round(coeff * unitSizePx);
				height = intValue * unitSizePx;
				yPos = posY_BL - height;
				if (value*coefficient > 0){
					g.setColor(posIntegerColor);
				}
				else{
					g.setColor(negIntegerColor);
				}
				
				g.fillRect(posX_BL+1, yPos, width-1, height-1);
				
				//Draw a line for each unit
				int i;
				g.setColor(Color.black);
				for (i=0;i<intValue;i++){
					g.drawLine(posX_BL+1, yPos+i*(unitSizePx) -1, posX_BL+unitSizePx, yPos+i*(unitSizePx) -1);
				}
			}
	
			//Fraction portion
			if (fracValue >= minFracValue){
				//Draw a bar for the fraction portion
				width = (int)Math.round(coeff * unitSizePx);
				height = (int)Math.round(fracValue * unitSizePx);		
				if (Math.abs(intValue) > 0){
					yPos = posY_BL - (intValue * unitSizePx) - height;
				}
				else{
					yPos = posY_BL - height;
				}
				if (value*coefficient > 0){
					g.setColor(posFractionColor);
				}
				else{
					g.setColor(negFractionColor);
				}
				g.fillRect(posX_BL+1, yPos, width-1, height-1);
			}

		}//drawMode == DRAW_BOXES
		else if (drawMode == DRAW_SINGLEBOXES){
			
			int separationPx = unitSizePx/10;

			//Integer portion
			if (Math.abs(intValue) > 0){
				//Draw a box for each unit
				width = (int)Math.round(coeff * unitSizePx);
				height = unitSizePx;
				yPos = posY_BL - (intValue * unitSizePx) - ((intValue-1) * separationPx);
				
				//Draw a box for each unit
				int i;
				for (i=0;i<intValue;i++){
					//Draw the border
					g.setColor(Color.black);
					g.drawRect(posX_BL, yPos-1, unitSizePx, height);
					
					//Fill the box
					if (value*coefficient > 0){
						g.setColor(posIntegerColor);
					}
					else{
						g.setColor(negIntegerColor);
					}
					g.fillRect(posX_BL+1, yPos, width-1, height-1);
					
					yPos = yPos + unitSizePx + separationPx;
				}
			}
	
			//Fraction portion
			if (fracValue >= minFracValue){

				//Draw a box for the fraction portion

				//Draw the border
				width = (int)Math.round(coeff * unitSizePx);
				height = unitSizePx;		
				if (Math.abs(intValue) > 0){
					yPos = posY_BL - (intValue * unitSizePx) - (intValue * separationPx) - height;
				}
				else{
					yPos = posY_BL - height - 1;
				}
				g.setColor(Color.black);
				g.drawRect(posX_BL, yPos-1, unitSizePx, height);
				
				//Fill the box
				width = (int)Math.round(coeff * unitSizePx);
				height = (int)Math.round(fracValue * unitSizePx);		
				if (Math.abs(intValue) > 0){
					yPos = posY_BL - (intValue * unitSizePx) - (intValue * separationPx) - height;
				}
				else{
					yPos = posY_BL - height - 1;
				}
				if (value*coefficient > 0){
					g.setColor(posFractionColor);
				}
				else{
					g.setColor(negFractionColor);
				}
				g.fillRect(posX_BL+1, yPos, width-1, height-1);
			}

		}//drawMode == DRAW_SINGLEBOXES
		
		g.setColor(oldColor);
	}

	private void resizeComponent()
	{
		double value;
		int intValue;
		double fracValue;

		if (variableValue==null){
			value = 1;
		}
		else{
			value = variableValue.getValue();
		}
		
		//Round the value to pay attention to decimals 2 order (0.01)
		value = (double)(Math.round(value*100))/100;
		
		intValue = (int)Math.floor(Math.abs(value));
		fracValue = Math.abs(value) - intValue;
		
		int unitSizePx = unitSize;		
		int width, height=0;

		width = unitSizePx;

		if (drawMode == DRAW_BAR){
		
			height = (int)Math.round(Math.abs(value) * unitSizePx);
			
		}//drawMode == DRAW_BAR
		else if (drawMode == DRAW_BOXES){
		
			height = intValue * unitSizePx;
			if (fracValue >= minFracValue){
				height = height + unitSizePx;
			}
			
		}//drawMode == DRAW_BOXES
		else if (drawMode == DRAW_SINGLEBOXES){
			
			int separationPx = unitSizePx/10;
			height = (intValue * unitSizePx) + ((intValue-1) * separationPx);
		
			if (fracValue >= minFracValue){
				height = height + unitSizePx + separationPx;
			}

		}//drawMode == DRAW_SINGLEBOXES

		//Set its own size!
		boxHeight = height;
		//this.setSize(width+1, height+1);
		//this.setLocation(this.getX(), this.getParent().getHeight()- height );
//		System.out.print("--- resizing from "+this.getY());
//		System.out.println("--- resizing to "+this.getParent().getHeight()+" "+height+" "+(this.getParent().getHeight()- height));
		
	}
	
	protected void processMouseEvent(MouseEvent e)
	{
		//only fire mouse events if the area clicked is inside the filled box!
		if (e.getY() > this.getHeight() - boxHeight){
			super.processMouseEvent(e);
		}
	}

}












