/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/**
 * Last modification information:
 * $Revision: 1.4 $
 * $Date: 2004/08/11 15:58:40 $
 * $Author: ebrown-munoz $
 *
 * Copyright 2004 The Concord Consortium
*/
package org.concord.graphfunction.ui.symbolicview;

import java.awt.*;
import java.text.*;
import java.util.*;

import javax.swing.*;

import org.concord.graph.event.*;
import org.concord.graphfunction.engine.*;
import org.concord.graphfunction.event.*;
import org.concord.graphfunction.ui.modelview.*;

/**
 * LinearEvalLabel.
 * Label that shows a Linear Function 
 * evaluated at a certain variable value <br>
 * 
 * @author Ingrid Moncada<p>
*/

public class LinearEvalLabel extends JLabel
	implements GraphFunctionView, 
			   ValueListener,
			   SelectableListListener
{

	//Only represents one function
	LinearFunction function;

	// ... maintained by a graphFunctionList
	GraphFunctionList glist;
	
	//Variable to listen to 
	Value variableValue;
		
	Format coeffFormatObj = new DecimalFormat("#0.0#");
	Format valuesFormatObj = new DecimalFormat("#0.0#");

    Font labelFont = new Font("SansSerif",Font.BOLD,16);
	
	/**
	 * Constructs an empty eval label
	*/
	public LinearEvalLabel()
	{
		super();
		setFont(labelFont);
		setOpaque(false);
	}
	
	/**
	 * Constructs an area model view for a linear function
	*/
	public LinearEvalLabel(LinearFunction g)
	{
		this();
		setGraphFunction(g);
	}
	
	/**
	 * Sets the GraphFunction that this view will represent. 
	*/
	public void setGraphFunction(GraphFunction g)
	{
		//Has to be linear
		if (g instanceof LinearFunction){
			setGraphFunction((LinearFunction)g);
		}
		else if (g instanceof LinearEquation){
			setGraphFunction(((LinearEquation)g).getLinear());
		}
	}

	public void setGraphFunctionList(GraphFunctionList gl)
	{
		if (glist!=null){
			glist.removeGraphableListListener(this);
		}
		glist = gl;
		if (glist != null){
			glist.addGraphableListListener(this);
			Object current = gl.getFirstSelectedObject();
			if (current instanceof GraphFunction) {
				setGraphFunction((GraphFunction)current);
			}
		}
	}
	
	/**
	 * Sets the GraphFunction that this view will represent. 
	*/
	public void setGraphFunction(LinearFunction g)
	{
		if (function!=null){
			function.removeGraphableListener(graphableHandler);
		}
		function = g;
		if (function!=null){
			function.addGraphableListener(graphableHandler);
		}

		update();
	}

	/***/
	private GraphableListener graphableHandler = new GraphableAdapter () 
	{
		/* Listener to the changes in the graphable.
		Repaints the view when it changes.
		*/
	    public void graphableChanged(EventObject e) 
		{
			update();
	    }
	};
	
	/**
	 * Sets the Value that this view will listen to as the variable
	*/
	public void setVariableValue(Value varValue)
	{
		if (variableValue != null) {
			variableValue.removeValueListener(this);
		}

		variableValue = varValue;

		if (variableValue != null) {
			variableValue.addValueListener(this);
		}
	}	

	public void update() {

		if ( (variableValue != null ) && (function != null)) {
			double slope = function.getSlope();
			double yIntercept = function.getYIntercept();
			double value = variableValue.getValue();
			double result = slope*value + yIntercept;
			String slopeString = 
				Double.toString(Math.floor( slope * 10.0 + 0.5) / 10.0);
			String yInterceptString = 
				Double.toString(Math.floor( yIntercept * 10.0 + 0.5) / 10.0);
			String valueString = 
				Double.toString(Math.floor( value * 10.0 + 0.5) / 10.0);
			String resultString = 
				Double.toString(Math.floor( result * 10.0 + 0.5) / 10.0);
			setText("<html> <i>" + resultString + "</i> = <font color=#ff0000>" + 
					slopeString + "</font><i> (" + 
					valueString + ")</i> + <font color=#00ff00>" + yInterceptString + 
					"</font>");
			repaint();
		}
	}

	/**
	 * Listener for the changes in the variable value
	*/
    public void valueChanged (ValueEvent e)
	{
		update();
	}

	public void listGraphableAdded(EventObject e)
	{
		update();
	}
	public void listGraphableChanged(EventObject e) 
	{
		update();
	}
	public void listGraphableRemoved(EventObject e) 
	{
		update();
	}
	public void listGraphableSelected(EventObject e) 
	{

		setGraphFunction((GraphFunction)e.getSource());

		update();

	}
	public void listGraphableDeselected(EventObject e) 
	{
	}

	
} // class LinearEvalLabel











