/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/* Author: Eric Brown-Munoz 
   $Revision: 1.6 $
*/


/** A Simple Model to use with the Function Analyzer. It is quick
	elegant and efficient. It has the added advantage that it
	is easy to maintain.
*/
package org.concord.rttapplets.fanalyzer;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.swing.*;

import org.concord.graph.engine.*;
import org.concord.graphfunction.engine.*;
import org.concord.graph.event.*;
import org.concord.graphfunction.event.*;

public class SimpleFAModel 
	extends JPanel
	implements ValueListener,
			   GraphableListener,
			   SelectableListListener,
			   GraphFunctionView,
			   MouseListener
{

	private static double SIN45 = 1.0;

	public static int DELETE_ONCLICK = 1;
	public static int NOTHING_ONCLICK = 0;

	private int clickMode = NOTHING_ONCLICK;

	Value variable;
	LinearEquation function;
	GraphFunctionList glist;


	// Attributes
	int sideLength = 25;
	int spacing = 5;
	Color slopeColor = new Color(255,0,0);
	Color slopeFractionColor = new Color(255,100,100);
	Color interceptColor = new Color(0,255,0);
	Color interceptFractionColor = new Color(100,255,100);
	Color lineColor = Color.black;

	public SimpleFAModel () {
		this(null, null);
	}

	public SimpleFAModel (Value var, LinearFunction funct) {
		super();
		addMouseListener(this);
		setVariable(var);
		setGraphFunction(funct);
	}

	public void setVariable(Value var) {

		if (variable != null) {
			variable.removeValueListener(this);
		}
		variable = var;
		if (variable != null ) {
			variable.addValueListener(this);
		}
	} 

	public Value getVariable() {
		return variable;
	}

	public void setGraphFunction(GraphFunction gf) {

		if (function != null) {
			function.removeGraphableListener(this);
		}

		// FIXME - Check for Class?
		function = (LinearEquation)gf;

		if (gf != null)
			gf.addGraphableListener(this);
	}
			

	public void setGraphFunctionList(GraphFunctionList gl)
	{
		if (glist!=null){
			glist.removeGraphableListListener(this);
		}
		glist = gl;
		if (glist!=null){
			glist.addGraphableListListener(this);
			Object current = gl.getFirstSelectedObject();
			if (current instanceof GraphFunction) {
				setGraphFunction((GraphFunction)current);
			}
		}
	}

	public void setClickBehavior (int mode) {
		clickMode = mode;
		
	}

	public void paintComponent (Graphics g) {
		super.paintComponent(g);

		if (function != null)
			g.setColor(function.getColor());

		// Change to a 16 point font
		g.setFont(g.getFont().deriveFont(12.0f));

		g.drawString("Modeling with Tiles", getWidth() - 180, 20); 
		g.drawString("Area = Value of the function.", getWidth() - 180, 35); 


		if ( (function == null ) || (variable == null)) return;

		Graphics2D g2 = (Graphics2D)g;

		double slope = function.getSlope();
		double yintercept = function.getYIntercept();
		double value = variable.getValue();

		Paint savedPaint = g2.getPaint();
		Color savedColor = g2.getColor();

		boolean isSlopeStriped = false;
		boolean isInterceptStriped = false;

		int xcursor = spacing;
		int ycursor = getHeight();

		int barwidth;


		Rectangle2D.Double shape = new Rectangle2D.Double();
		Line2D.Double tempLine = new Line2D.Double();


		if ( (slope * value) < 0.0)
			isSlopeStriped = true;

		if ( yintercept < 0.0)
			isInterceptStriped = true;

		value = Math.abs(value); 
		yintercept = Math.abs(yintercept);
		slope = Math.abs(slope);

		int valueInt = (int)value; // FIXME make these look pretty
		double valueFractional = value - (double)valueInt;
		
		// ---- Draw the slope boxes
		for (double tslope = slope; tslope > 0.0; 
			 tslope = tslope - 1.0) 
			{
				if (tslope >= 1.0) {
					barwidth = sideLength;
				} else {
					barwidth = (int)(sideLength * tslope);
				}
				
				g2.setColor(slopeColor);
	
				shape.setRect(xcursor, ycursor - sideLength * valueInt,
								  barwidth, sideLength * valueInt);
				g2.fill(shape);

				g2.setColor(slopeFractionColor);
				shape.setRect(xcursor, ycursor - sideLength * value,
								  barwidth, sideLength * valueFractional);
				g2.fill(shape);

				if (isSlopeStriped) {
					shape.setRect(xcursor, ycursor - sideLength * value,
								  barwidth, sideLength * value);
					
					drawStripes(g2,shape);
				}

				// Draw the lines (outlines)
				double boxYposition = ycursor - sideLength - Math.ceil(value);
				double boxheight = sideLength * Math.ceil(value) ;

				g2.setColor(lineColor);
				shape.setRect(xcursor, ycursor-sideLength * Math.ceil(value),
							  sideLength, boxheight -1.0);
				g2.draw(shape);
				
				for ( int tline = 1 ; tline < (int)Math.ceil(value);
					  tline ++)
					{
						tempLine.setLine( xcursor, ycursor - tline*sideLength,
										  xcursor + sideLength, 
										  ycursor - tline*sideLength);
						g2.draw(tempLine);
					}


				xcursor = xcursor + sideLength + spacing;

			} // for tslope

		// ---- Draw Intercept boxes
		for (double tintercept = Math.floor(yintercept); tintercept > 0;
			 tintercept = tintercept - 1.0)
			{
				// We can handle fractional parts in the future-- right here
				shape.setRect( xcursor, ycursor - sideLength, 
							   sideLength, sideLength -1.0);

				g2.setColor(interceptColor);
				g2.fill(shape);

				if (isInterceptStriped) {
					drawStripes(g2,shape);
				}

				g2.setColor(lineColor);
				g2.draw(shape);

				xcursor = xcursor + sideLength + spacing;

			}

	}


	void drawStripes (Graphics2D g2, Rectangle2D stripeShape) {
		Stroke saveStroke = g2.getStroke();

		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
							RenderingHints.VALUE_ANTIALIAS_ON);

		g2.setStroke (new BasicStroke(5));
		g2.setColor(Color.white);

		double startx = stripeShape.getX();
		double starty = stripeShape.getY();
		double width = stripeShape.getWidth();
		double height = stripeShape.getHeight();

		double bottom = starty + height;
		double right = startx + width;

		double ypos = starty;

		while (ypos <  bottom ) {
			if ( ( ypos + width * SIN45 ) < bottom) {
				
				g2.draw(new Line2D.Double(startx, ypos, startx + width, 
										  ypos + width * SIN45));
			} else { // hits the bottom
				double xEnd = (bottom - ypos) / SIN45;
				g2.draw(new Line2D.Double(startx, ypos, startx + xEnd, 
										  bottom));
			}

			ypos += 10.0;
		}

		double xpos = startx;

		while (xpos <= right ) { 
			if (( xpos + height * SIN45 ) > right ) { // hits the side 
				g2.draw(new Line2D.Double ( xpos, starty, 
											startx + width,
											starty + (right - xpos * SIN45)));
						} else {
							g2.draw(new Line2D.Double( xpos, starty, 
													   xpos + height * SIN45, 
													   bottom));
						}// if (xpos + height ... > right)

			xpos += 10.0;

		} // while 
		g2.setStroke(saveStroke);

	} // drawStripes()




	// ---- Implement GraphableListListener 
	public void listGraphableAdded(EventObject e)
	{
		repaint();
	}
	public void listGraphableChanged(EventObject e) 
	{
		repaint();
	}
	public void listGraphableRemoved(EventObject e) 
	{
		repaint();
	}
	public void listGraphableSelected(EventObject e) 
	{

		setGraphFunction((GraphFunction)e.getSource());

		repaint();

	}
	public void listGraphableDeselected(EventObject e) {}

	// ---- Implement GraphableListener

	public void graphableChanged(EventObject e) {

	}
	
	public void graphableRemoved(EventObject e) {
		function = null;
	}

	// ---- Implement ValueListener

	public void valueChanged(ValueEvent e) {
		repaint();
	}

	// ---- Implement MouseMotionListener
	public void mouseClicked(MouseEvent e) {

		if (clickMode == DELETE_ONCLICK ) {

			int x = e.getX();
			int y = e.getY();
			int slope = (int)Math.ceil(Math.abs(function.getSlope()));
			int value = (int)Math.ceil(Math.abs(variable.getValue()));
			int intercept = (int)Math.ceil(Math.abs(function.getYIntercept()));
			int wavelength = spacing + sideLength; // length of one cycle.
			int height = getHeight();

			// Check if the click is inside the slopeboxes
			if ( ( x <= ( wavelength * slope) ) && 
				 ( e.getY() > ( height - value * sideLength) ) &&
				 ( x % wavelength ) > spacing)
				{
					double oldslope = function.getSlope();
					double newslope = 
						(oldslope > 0.0) ? oldslope -1 : oldslope +1;

					function.setSlope(newslope);
				} else if (( x >= (wavelength * slope)) &&
						   ( x <= (wavelength * slope + wavelength * intercept)) &&
						   ( y >= (height - sideLength)) &&
						   ( (x % wavelength ) > spacing)) {
					double oldint = function.getYIntercept();
					double newint =
						(oldint > 0.0) ? oldint - 1 : oldint + 1;
					
					function.setYIntercept(newint);
				}
		} // if clickMode
	}

	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e) {}

}
