/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/* Author: Eric Brown-Munoz

   A Graphable object for use with the CC Graph
   package. Based on the QGrapher Applet.
*/

/** The QualGraphable is a continuous piecewise Graph.
	It consists of "handles" that are moveable connected
	by "segments" that are either straight lines or
	quadratic curves. 
*/

package org.concord.rttapplets.qgrapher;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import org.concord.graph.engine.*;
import org.concord.graphfunction.engine.*;

public class QualGraphable extends DefaultGraphFunction 
						   implements MouseControllable
{

    /** The list of segments that make up the Piecewise function. */ 
    protected Vector segments;

    /** The list of handles that make up the Piecewise function. */
    protected Vector handles;

    // A temp veriable to help with the search for yval
    private int CurrentSegment;

	// The Handle that was clicked (and presumably will be dragged)
	private Handle proximateHandle = null; 
	private Handle clickedHandle = null; 
	private Point clickedPosition = null;
	private Point2D.Double worldLocation = new Point2D.Double();

	// The range that the clicked handle can be dragged to
	private int dragMin = Integer.MIN_VALUE;
	private int dragMax = Integer.MAX_VALUE;

    // The String used to initialize this Object. Remembered for reset method.
    private String initString;

	protected boolean isVisible = true;
	protected Color color = Color.blue;
	protected double adjust = 2.0;
	
	public QualGraphable() {
		segments = new Vector();
		handles = new Vector();
	}

	public Graphable getCopy() {
		return null;
	}

	public void draw( Graphics2D g) {
		Enumeration handlesEnum = handles.elements();
		Enumeration segmentsEnum = segments.elements();

		while (segmentsEnum.hasMoreElements()) {
			QualSegment segment = (QualSegment)segmentsEnum.nextElement();
			segment.draw(g);
		}

		while (handlesEnum.hasMoreElements()) {
			Handle handle = (Handle)handlesEnum.nextElement();
			handle.draw(g);
		}
	}

	/** Creates a new handle at point (x,y) */
	public Handle newHandle(double x, double y) {
		Handle handle = new Handle(this, new Point2D.Double(x,y));
		handles.add(handle);

		return handle;
	}

	public void setColor (Color c) {
		color = c;
	}

	public Color getColor () {
		return color;
	}

	/** Returns the last value of X that is in the domain
	 */
	public double getEnd () { 
		Handle endH = lastHandle();
		if (endH != null) 
			return endH.getX();
		else
			return Double.NaN;
	
	}

	/** Creates a new Segment starting a last Handle and extending
		to point (x,y) */
	public QualSegment newSegment (double x, double y , int opens ) {
		Handle first;
		Handle last;
		QualSegment segment;

		if (handles.size() == 0) {
			first = newHandle (0.0,0.0);
		} else {
			first = lastHandle();
		}

		if ( x > first.position.x) {
			last = newHandle(x, y);

			segment = new QualSegment(this, first, last, opens);
			segments.add(segment);

			notifyChange();

			return segment;
		} else {
			return null;
		}

	}

	public void removeLastSegment() {
		if (segments.size() > 0 ) {
			QualSegment lastSegment = (QualSegment)segments.lastElement();
		
			Handle lastStartHandle = lastSegment.start;
			Handle lastEndHandle   = lastSegment.end;

			segments.remove(lastSegment);
			handles.remove(lastEndHandle);
			
			if (segments.size() == 0) {
				handles.remove(lastStartHandle);
			}

			notifyChange();
		}
	}

	public Handle lastHandle() {
		if (handles.size() > 0 ) {
			return (Handle)handles.lastElement();
		}
		return null;
	}

	public CoordinateSystem getCoordinateSystem() {
		return getGraphArea().getCoordinateSystem();
	}

	public synchronized boolean isPointInProximity(Point location) {
		boolean retval = false;
		Enumeration handlesEnum = handles.elements();

		while (handlesEnum.hasMoreElements() && !retval) {
			Handle handle = (Handle)handlesEnum.nextElement();
			if (handle.isPointInProximity(location)) {
				retval = true;
				proximateHandle = handle;
			}
		} // while

		return retval;
	}

	/** Return the y value of the graphable at x for the QualGraphable
	 */

	public double yval (double xval) {
		double retval = Double.NaN;
		Enumeration segmentEnum = segments.elements();
		QualSegment eachSeg;
		QualSegment theSegment = null;
		double startx;
		double endx;

		while ((theSegment == null ) && segmentEnum.hasMoreElements()) {
			eachSeg = (QualSegment)segmentEnum.nextElement();
			startx = eachSeg.getStart().getX();
			endx = eachSeg.getEnd().getX();
			if ( ( startx <= xval) && ( endx >= xval) ) 
				theSegment = eachSeg;
		}

		if (theSegment != null) {
			retval = theSegment.yval(xval);
		}

		return retval;
	}


	// Methods for the MouseControllable Interface

	/** Now does nothing since isPointinProximity already confirmed
		the mouse press. */
	public boolean mousePressed(Point location) {
		int index;
		Point2D.Double previousLocation = new Point2D.Double();
		Point2D.Double nextLocation = new Point2D.Double();
		CoordinateSystem cs = getCoordinateSystem();

		clickedHandle = proximateHandle;

		// Set dragMin and dragMax. This ensure that segments dont cross
		index=handles.indexOf(clickedHandle);

		if (handles.size() <= 1) {
			dragMin = Integer.MIN_VALUE;
			dragMax = Integer.MAX_VALUE;
		} else {

			if (index > 0) {
				cs.transformToDisplay(((Handle)
									   handles.get(index-1)).getPosition(),
									  previousLocation);
				dragMin=(int)previousLocation.x + 1;
			} else {
				dragMin = Integer.MIN_VALUE;
			}

			if (index < (handles.size()-1)) {
				cs.transformToDisplay(((Handle)
									   handles.get(index+1)).getPosition(),
									  nextLocation);
				dragMax=(int)nextLocation.x - 1;
			} else {
				dragMax = Integer.MAX_VALUE;
			}
		}


		return true;
	}
	
	/** Handles the mouse drag event. 
	 * The location is given in display coordinates. 
	 */

	public boolean mouseDragged(Point location) {
		Point tlocation = (Point)location.clone();

		if (tlocation.x < dragMin)
			tlocation.x = dragMin;

		if (tlocation.x > dragMax)
			tlocation.x = dragMax;

		getCoordinateSystem().transformToWorld(tlocation, worldLocation);

		clickedHandle.setPosition(worldLocation);
		notifyChange();

		return true;
		
	}
	
	/** Handles the mouse released event. 
	 * The location is given in display coordinates. 
	 */
	public boolean mouseReleased(Point location) {
		clickedHandle = null;
		return true;
	}
	
	/**  
	 * Returns if the object is currently being
	 * controlled by the mouse (dragged)
	 */
	public boolean isMouseControlled() {
		return false;
	}

	public String getExpression (int express) {
		return "";
	}

	public void update() {
	}


}


