/* Copyright (c) 2004 The Concord Consortium, Inc., 
   10 Concord Crossing, Concord, MA 01741.

   This library is free software; you can redistribute it
   and/or modify it under the terms of the GNU Lesser General
   Public License as published by the Free Software 
   Foundation; version 2.1 of the License.

   This library is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the 
   implied warranty of MERCHANTABILITY or FITNESS FOR A 
   PARTICULAR PURPOSE. See the GNU Lesser General Public 
   License for more details.

   The GNU Lesser General Public License can be found at 
   http://opensource.org/licenses/lgpl-license.php; for a 
   paper copy, write to the Free Software Foundation, Inc., 
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


/* Author: Eric Brown-Munoz

   A Graphable object for use with the CC Graph
   package. Based on the QGrapher Applet.
*/

/** The QualGraphable is a continuous piecewise Graph.
	It consists of "handles" that are moveable connected
	by "segments" that are either straight lines or
	quadratic curves. 
*/

package org.concord.rttapplets.qgrapher;

import java.awt.*;
import java.awt.geom.*;
import java.util.Vector;

import org.concord.graph.engine.*;

public class QualSegment 
{

	public final static int LINE = 0 ;
	public final static int RIGHT = 1;
	public final static int LEFT = 2;

	QualGraphable parent;
	Handle start;
	Handle end;

	Color segmentColor = Color.blue;
	BasicStroke stroke = new BasicStroke(4.0f);

	public Shape curve;

	private int type = LINE;

	protected Point2D.Double transStart = new Point2D.Double();
	protected Point2D.Double transEnd = new Point2D.Double();
	protected Point2D.Double control = new Point2D.Double();
	protected Point2D.Double transControl = new Point2D.Double();


	public QualSegment(QualGraphable parent, Handle start, Handle end, 
					   int type) {
		this.start = start;
		this.end = end;
		this.parent = parent;
		this.type = type;
	}

	public QualSegment(QualGraphable parent, Handle start, Handle end) {
		this(parent,start,end,LINE);
	}

	public Handle getStart() {
		return start;
	}

	public Handle getEnd() {
		return end;
	}

	public synchronized void  draw (Graphics2D g) {

		Color savedColor = g.getColor();
		Stroke savedStroke = g.getStroke();
		Point2D startPoint = start.getPosition();
		Point2D endPoint = end.getPosition();
		double width;
		double height;
		double slope; // a scratch variable to determine control point

		g.setColor(segmentColor);
		g.setStroke(stroke);


		CoordinateSystem cs = parent.getCoordinateSystem();

		cs.transformToDisplay(startPoint,transStart);
		cs.transformToDisplay(endPoint,transEnd);
		
		if (type == LINE) {
			curve = new Line2D.Double(transStart, transEnd);
		} else {
			width = endPoint.getX() - startPoint.getX();
			height = endPoint.getY() - startPoint.getY();
			slope = height/width * 2.0;

			if (type == RIGHT) {
				control.setLocation ( startPoint.getX() + height/slope , 
									  endPoint.getY());

				cs.transformToDisplay ( control,transControl);
			} else if (type == LEFT){
				control.setLocation ( endPoint.getX() - height/slope , 
									  startPoint.getY());

				cs.transformToDisplay ( control,transControl);
			}

			curve = new QuadCurve2D.Double(transStart.x, transStart.y, 
										   transControl.x, transControl.y,
										   transEnd.x  , transEnd.y);

		} // ! (type == LINE)

		g.draw(curve);

		g.setColor(savedColor);
		g.setStroke(savedStroke);


	}


	public double yval (double xval) {
		double retval = Double.NaN;

		Point2D startPoint = start.getPosition();
		Point2D endPoint = end.getPosition();
		double width;
		double height;
		double slope;
		double ixval;

		width = endPoint.getX() - startPoint.getX();
		height = endPoint.getY() - startPoint.getY();

		if (type == LINE) {
			slope = height/ width;
			retval = startPoint.getY() + slope * (xval - startPoint.getX());

		} else if (type == RIGHT) {
			ixval = endPoint.getX() - xval;
			retval =  startPoint.getY() - ixval * ixval * 
				height / (width * width) + height;
		} else if (type == LEFT) {
			ixval = xval - startPoint.getX();
			retval = startPoint.getY() + ixval * ixval * 
				height / (width * width);
		}

		return retval;
	}


			

	public void setParent(QualGraphable parent) {
		this.parent = parent;
	}

	public QualGraphable getParent() {
		return parent;
	}



}
